/*
 * file: Bootstrapper.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */
/**********************************************************************************************************************
 * @author Brook Novak
 * 
 * Assumes: 
 *         - Core.js is prepended(or loaded).
 *         - An array named "baseScripts" exists with all relative paths of scripts to load first in the given order.
 *         - An array named "dependantScripts" exists with all relative paths of scripts to load after the base scripts.
 *         - A string named "thisFilename" exists, set to the filename of this file, which is contained in the libraries TDL.
 *         - That this script is encapsulated in an anonymous function (To avoid polluting global scope)
 **********************************************************************************************************************/

// Mimic the addHandler method in the events module to make available before the events module script is even available
var eventHandlers = [];
de.events.addHandler = function(eventSource, eventName, handler) {   // will be overridden
    eventHandlers.push(arguments);
}

// Override the onready function (declared in Core)- so that all callbacks are invoked once
// the window.onload occurs AND all scripts are loaded
var onReadyCallbacks = [],
    origReady = de.onready;
de.onready = function(handler) {
    onReadyCallbacks.push(handler); 
}

/**
 * @namespace
 * The bootstrap namespace.
 */
bootstrap = {
    
    /**
     * A list of all laoded module names
     * @type [String]
     */
    loadedModules : [],
    
    /**
     * For all scripts in the API you must invoke this function to register the script when the script is first executed.
     * 
     * @param {String} name  The name of module that is loaded. 
     *                       The Module name is the relative path without the file extension (.js), 
     *                       and directory separators are replaced with period-symbols. 
     */
    provides : function(name) {
        
        // debug.println("Loaded: " + name);
        
        this.loadedModules.push(name);
        
        // First Phase: base-script loading
        if (baseScripts.length > 0) {
            
            // A base script is loaded - remove it
            baseScripts.shift();
            
            if (baseScripts.length > 0) { // More base scripts to load
                
                // Load the next base script
                loadScript(baseURL + baseScripts[0]);
                
            } else { // Finished download base scripts
                
                // Pull in remaining API source files asynchronously
                for (var i in dependantScripts) {
                    loadScript(baseURL + dependantScripts[i]);
                } 
            }
            
        } else { // Second phase: dependant-script loading
                    
            // A dependant script is loaded.
            
            // Get relative path for registered module
            var registeredPath = name.replace(/\./g,"/").toLowerCase() + ".js";
        
            // Remove include path for the registered module
             for (var i = 0; i < dependantScripts.length; i++) {
                if (dependantScripts[i].toLowerCase() == registeredPath) {
                    dependantScripts.splice(i, 1);
                    break;
                } 
            }
            
            if (dependantScripts.length == 0) { // All scripts loaded?
        
                // Schedule finishing-up code on another event to let the last script load
                setTimeout(function() {
                    
                    // Add all registered events...
                    for (var i in eventHandlers) {
                        var args = eventHandlers[i];
                        
                        var eventSource = (args.length > 0) ? args[0] : null,
                            eventName = (args.length > 1) ? args[1] : null,
                            handler = (args.length > 2) ? args[2] : null;
                
                        _addHandler(eventSource, eventName, handler);
                    }
                    
                    // Add all on-ready funcs
                    for (var i in onReadyCallbacks) {
                       origReady(onReadyCallbacks[i]); // Fires instantly is onload already occured
                    }
                
                    de.onready = origReady;
                    
                }, 1);
        
            }
        }
        
    }
    
};

// Determine the base URL of the library.
var baseURL;
var libcoreRegExp = new RegExp("(^|[\\/\\\\])" + thisFilename.replace(/\./g,"\\.") + "(\\?|$)");
var scriptElements = document.getElementsByTagName("script");

for (var i in scriptElements) {
    var src = scriptElements[i].src;
    if (src && src.match(libcoreRegExp)) {

        baseURL = src;
        i = src.lastIndexOf("/");
        if (i == -1) i = src.lastIndexOf("\\");
        baseURL = (i == -1) ? "" : src.substring(0, i + 1);

        break;
    }
    
}

if (!baseURL) throw new Error("Unable to import library: could not discover DEdit libhome");

// Pull in API base source files first, once this is done the rest of the api source will be pulled in
loadScript(baseURL + baseScripts[0]);

function loadScript(scriptSrc){

    // gets document head element  
    var docHead = document.getElementsByTagName('head')[0];
    
    if (docHead) {
        // creates a new script tag        
        var scriptEl = document.createElement('script');
        
        // adds src and type attribute to script tag  
        scriptEl.setAttribute('src', scriptSrc);
        scriptEl.setAttribute('type', 'text/javascript');

        // append the script tag to document head element          
        docHead.appendChild(scriptEl);
    }
}

