/*
 * file: ContainerNormalization.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */
bootstrap.provides("ContainerNormalization");

/*
 * TODO: MAKE HTML VALID BY PLACING INLINE ANCESTORS WITHIN THE BLOCK-LEVEL CONTAINERS
 * Inlines cannot contain blocks
 * 
 * 
 * ARGGG... OK JUST RUBBISH THIS.. For all invalid inlines, remove them from the document.. that will naturally
 * "normalize" the range. Cannot just move inlines into containers because containers may contain block decendants.
 * 
 * This will simplify the code a lot and keep the HTML tidy.
 * 
 * Should design algrorithms to assume valid HTML, ad if invalid then loss of formatting as a result of throwing away
 * invalid tags doesnt matter.
 */

/**
 * Creates containers, adjusts structure within the given range, such that
 * it garauntees that all block-level elements in the range do not share any ancestor inline elements
 * that occur between the containers and the common-block-level-ancestor,
 * with other block-level elements. This property is useful for itemizing container in
 * a given range.
 * <br>
 * This operation is undoable, and is intended for indentation and itemizing ranges.
 * 
 * @param {Node} startNode	        The starting dom node of the range to normalize.
 *
 * @param {Node} endNode	        The ending dom node of the range to normalize. 
 *                                  Can be the same as start node or it must occur
 *                                  after the start dom (in-order left-to-right traversal)
 *                          
 * @param {Node} containerTemplate  (Optional) An element which supports inline elements to be
 *                                  used for inline-groups which need containers. Defaults to
 *                                  paragraph.
 * 
 * @return {[Node]}                 An array containing a list of all the top-level containers in the given range,
 *                                  in order if traversing the dom tree in-order. Can be empty.
 * 
 */
function _getNormalizedContainerRange (startNode, endNode, containerTemplate) {
    
    var template = containerTemplate || $createElement("p");
    
    // Determine the master container for the normalization range
    var masterContainer = _findAncestor(_getCommonAncestor(startNode, endNode, true), 
                                      docBody, 
                                      _isBlockLevel, 
                                      true) 
                                      || docBody;
                                      
    // Deepen start/end range
    while (startNode.firstChild) {
        startNode = startNode.firstChild;
    }
    while (endNode.lastChild) {
        endNode = endNode.lastChild;
    }

    // Check if the master container allows the container template
    if (masterContainer != docBody && !_isValidRelationship(template, masterContainer)) {
        
        // Special case - list items
        if (_nodeName(masterContainer) == "ul" || _nodeName(masterContainer) == "ol") {
            startNode = _findAncestor(startNode, masterContainer);
            endNode = _findAncestor(endNode, masterContainer);
            var containers = [];
            while (startNode) {
                if (_nodeName(startNode) == "li")
                    containers.push(startNode);
                startNode = startNode == endNode ? null : startNode.nextSibling;
            }
            return containers;
            
        }
        
        // If the master container doesn't allow sub-containers, then the normalized range
        // becomes a single container... the master container.
        return [masterContainer]
    }
    
    // Extend range to point to begin and start at top-level sub-containers if they exist,
    // or to the end of inline groups
    var extendedRangeNode, 
        protectedContainers = [],
        reinsertProcContainers = [],
        extendedStart, 
        extendedEnd;

    // Extend start range
    _visitAllNodes(masterContainer, startNode, false, extendRange);
    
    // Remove any protect containers
    if (protectedContainers.length > 0) {
        for (var i in protectedContainers) {
            if (protectedContainers[i].parentNode) {
                protectedContainers[i].parentNode.removeChild(protectedContainers[i]);
                reinsertProcContainers.push(protectedContainers[i]);
            }
        }

    }
    
    extendedStart = extendedRangeNode;
    if (!extendedStart) {
        // Set start through to end of initial inline group
        extendedStart = masterContainer.firstChild;
        while (extendedStart.firstChild) {
            extendedStart = extendedStart.firstChild;
        }
    }
    
    // Extend end range
    extendedRangeNode = null;
    protectedContainers = [];
    _visitAllNodes(masterContainer, endNode, true, extendRange);
    
    // Remove any protect containers
    if (protectedContainers.length > 0) {
        for (var i in protectedContainers) {
            if (protectedContainers[i].parentNode) {
                protectedContainers[i].parentNode.removeChild(protectedContainers[i]);
                reinsertProcContainers.push(protectedContainers[i]);
            }
        }
    }
    
    extendedEnd = extendedRangeNode;
    if (!extendedEnd) {
        // Set start through to end of post inline group
        extendedEnd = masterContainer.lastChild;
        while (extendedEnd.lastChild) {
            extendedEnd = extendedEnd.lastChild;
        }
    }
    
    // Move protected nodes out of the extended range if there are any
    for (var i in reinsertProcContainers) {
        docBody.appendChild(reinsertProcContainers[i]);
    }
    
    // The range only containers protected nodes
    if (!extendedEnd || ! extendedStart) 
        return [];
    
    // Perform normalization operations (two phases)
    separateBlockPaths(extendedStart, extendedEnd, masterContainer);
    encapsulateIGroups(extendedStart, extendedEnd, masterContainer, template);
        
    // Get top-level containers within the original range
    var tdc = _findAncestor(startNode, masterContainer);
    endNode = _findAncestor(endNode, masterContainer);

    var range = [];
        
    do {
        var node = tdc;
        while (node) {
            if (_isBlockLevel(node)) {
                range.push(node);
                break;
            }
            node = node.firstChild;
        }
        tdc = tdc == endNode ? null : tdc.nextSibling;
    } while (tdc);
    
    return range;
    
    /**
     * Helper function for extending the range backward/forward.
     * Sets the extendedRangeNode local iff a top-level block container is found within the master container
     * 
     * @param {Node} domNode Provided by de.visit function
     */
    function extendRange(domNode) {
        if (domNode == masterContainer) 
            return;
            
        // Find top-level block element from the master-container - inclusive of self
        var node = domNode, blNode = null;
        while (node != masterContainer) {
            if (_isBlockLevel(node)) 
                blNode = node;

            // If this sub-tree is protected then note the protected node root                    
            if (node.parentNode == masterContainer && de.doc.getProtectedNodeContainer(node) == node) {
                protectedContainers.push(node);
                blNode = null; // Ignore this since this subtree will be repositioned in the document
            }
                
            node = node.parentNode;
        }
        
        if (blNode) {
            // If this node has a top-level block element from the master-container
            // then the extended start has been located
            extendedRangeNode = blNode;
            return false;
        }
    }
        
    
    /**
     * Ensures that all first occuring block-level elements from the master container own their own path
     * up to the master container.
     * 
     * @param {Node} startNode        The start node in the range. This must not have a block-level element
     *                                in the path from itself up to the master container (exclusive).
     *                 
     * @param {Node} endNode          The end node in the range.
     * 
     * @param {Node} masterContainer  The master container for all sub containers
     * 
     */
    function separateBlockPaths(startNode, endNode, masterContainer) {
        
        var scanPoint = startNode, 
            seenEndPoint = false;
        
        // Scan through range for all top-level block elements from the master container
        while (scanPoint && !seenEndPoint) {
            
            _visitAllNodes(masterContainer, scanPoint, true, function(domNode) {
                
                scanPoint = null;
                seenEndPoint |= domNode == endNode;
                
                if (domNode == masterContainer) 
                    return;
                
                // If this dom node is a block element then ensure that it owns its own path
                // up to the master container.
                if (_isBlockLevel(domNode)) {
                    
                    // Note: this block level element is top-level from the master container... i.e. there is no
                    // block-level elements between this node up to the master container
                    
                    var pivotNode = domNode;
                    while (pivotNode.parentNode != masterContainer) {
                        
                        // Separate previous siblings into duplicate inline node
                        if (pivotNode.previousSibling) {
                            
                            // Clone shared inline element
                            var clone = pivotNode.parentNode.cloneNode(false);
                             
                             // Migrate previous siblings into duped inline element
                            while (pivotNode.previousSibling) {
                                    
                                var psib = pivotNode.previousSibling;
                                
                                _execOp(_Operation.REMOVE_NODE, psib);
                                _execOp(_Operation.INSERT_NODE, psib, clone, 0);
                                
                            }
                            
                            // Insert the cloned inline element back into the document
                            _execOp(_Operation.INSERT_NODE, clone, pivotNode.parentNode.parentNode, _indexInParent(pivotNode.parentNode));
       
                        }
    
                        // Separate next siblings into duplicate inline node
                        if (pivotNode.nextSibling) {
                            
                            // Clone shared inline element
                            var clone = pivotNode.parentNode.cloneNode(false)
                             
                             // Migrate next siblings into duped inline element
                            while (pivotNode.nextSibling) {
                                var nsib = pivotNode.nextSibling;
                                _execOp(_Operation.REMOVE_NODE, nsib);
                                _execOp(_Operation.INSERT_NODE, nsib, clone);
                            }
                            
                            // Insert the cloned inline element back into the document
                            _execOp(_Operation.INSERT_NODE, clone, pivotNode.parentNode.parentNode, _indexInParent(pivotNode.parentNode) + 1);
   
                        }
                        
                        pivotNode = pivotNode.parentNode;
                        
                    } // End separating path to master container
                    
                    // This dom node now owns its own path to the master container.
                    // Setup to scan down next path
                    scanPoint = pivotNode.nextSibling; // NB: Pivot is an immediate child of master container
                    
                    // Check if the end node lies within this block-level element
                    _visitAllNodes(domNode, domNode, true, function(subDomNode) {
                        if (subDomNode == endNode) 
                            seenEndPoint = true;
                        return !seenEndPoint;
                    });
                    
                    return false;
                }
                
                return !seenEndPoint;
                
            });
            
        } // End Loop: separating paths to block-level elements in given range
        
    } // End function separateBlockPaths
    
    
    /**
     * For all sub-trees within the given range of the master container (i.e. Immediate children of the master continer), 
     * all sub-trees containing only inline elements/text nodes are migrated into a containerTemplate clone 
     * within the master container.
     * 
     * 
     * @param {Node} startNode           The start node to encapsulate from. This should not be within an inline sub-tree
     *                                   which has a previous sub-tree which contains only inline elements.
     * 
     * @param {Node} endNode             The end node to encapsulate to. This should not be within an inline sub-tree
     *                                   which has a following sub-tree which contains only inline elements.
     * 
     * @param {Node} masterContainer     The master container
     * 
     * @param {Node} containerTemplate   The container to clone for migrating inline groups into
     * 
     */
    function encapsulateIGroups(startNode, endNode, masterContainer, containerTemplate) {
        
        // Adjust start and end nodes to their root nodes within the master container
        startNode = _findAncestor(startNode, masterContainer);
        endNode = _findAncestor(endNode, masterContainer);
        
        var subTreeNode = startNode, 
            inlineGroupStart = null;
            
         // Visit all sub-trees in the master container within the range 
         // (i.e. immediate children in the master contianer).
        while (subTreeNode) {
            
            var domNode = subTreeNode, 
                containsBlock = false;
                
            while (domNode) {
                if (_isBlockLevel(domNode)) {
                    containsBlock = true;
                    break;
                }
                // Just check down left-most-path since all block level elements in
                // the given range have their own path to the master container
                domNode = domNode.firstChild;    
            } 
            
            if (containsBlock) {
                
                if (inlineGroupStart) {
                    encapsulate(inlineGroupStart, subTreeNode);
                    inlineGroupStart = null;
                }
    
            } else {
                // Mark start of inline group - avoid including whitespace nodes which should not be encapsulated.
                // For example, whitespace in between list items
                if (!inlineGroupStart) {
                    if (!(subTreeNode.nodeType == Node.TEXT_NODE && !_doesTextSupportNonWS(subTreeNode)))
                        inlineGroupStart = subTreeNode;
                }
            }
    
            subTreeNode = subTreeNode == endNode ? null : subTreeNode.nextSibling;
        }
         
        if (inlineGroupStart) 
            encapsulate(inlineGroupStart, null);
        
        /**
         * Encapsulates a run of inline siblings within the master container. 
         * 
         * @param {Node} start The sibling to start encapsulating from.
         * 
         * @param {Node} endEx The exclusive end sibling. 
         *                     Can be null for encapsulating all siblings from start node.
         */
        function encapsulate(start, endEx) {
            
            var igContainer = containerTemplate.cloneNode(false),
                igSubTree = start;
                
            // Insert the new inline-group container into the master container
            debug.assert(_isValidRelationship(igContainer, masterContainer));
            _execOp(_Operation.INSERT_NODE, igContainer, masterContainer, _indexInParent(start));
            
            // Migrate inline run into the new container
            while (igSubTree != endEx) {
                var nextSib = igSubTree.nextSibling;
                 _execOp(_Operation.REMOVE_NODE, igSubTree);
                 _execOp(_Operation.INSERT_NODE, igSubTree, igContainer);
                igSubTree = nextSib;
            }
        }
        
    } // End function encapsulateIGroups
    

} // End getNormalizedContainerRange

    
    

    
