/*
 * file: Cursor.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */
bootstrap.provides("Cursor");

(function(){
    
    // TODO: REFACTOR de.cursor ro de.Cursor ... do when refactor de prefix to chosen name

    /* A cursor descriptor that represents the current cursor that is showing. Null if no cursor showing */
    var currentCursorDesc,
        
        /* The visual representation for the cursor on the actual web page */
        cursorDiv,
        
        /* A Constant defining the time in ms between cursor blinks */
        CURSOR_BLINK_MS_TIME = 460,
        
        /* The cursor blinker timeout ID */
        cursorBlinkTOId,
        
        /* The x position which the cursor should realign to when moving up/down the content */
        cursorXAlign,
        
        /* Used for measurement purposes */
        measureSpanEl,
        measureSpanTextNode,
        measurePreTextNode,
        measurePostTextNode,
        measureFullText,
        
        // ------------- Define lookup maps according to the DirectEdit DOM-based Web Editor Specification 1.0 -------------
        
        /*
         * Refer to specification 2.2.2
         * Elements which the cursor can appear directly before
         */
        beforeElements = $createLookupMap("img,table,input,select,button,textarea,object"),
        
        /*
         * Refer to specification 2.2.3
         * Elements which the cursor can appear directly after
         */
        afterElements = $createLookupMap("img,table,input,select,button,textarea,object,br");
    

    $enqueueInit("Cursor", function() {

        // Create elements for measurement purposes
        measureSpanEl = $createElement("span");
        measureSpanTextNode = document.createTextNode("");
        measureSpanEl.appendChild(measureSpanTextNode);
        measurePreTextNode = document.createTextNode("");
        
        // Create cursor div and add it to the document
        cursorDiv = $createElement("div");
        _setClassName(cursorDiv, _PROTECTED_CLASS + " sw-cursor"); // Avoid the cursor from being edited
        docBody.appendChild(cursorDiv);
        
        // Set cursor background and z-index to defaults if css sheets don't supply them
        var cursorStyle = "",
            cssVal = _getComputedStyle(cursorDiv, 'z-index');
            
        if (!cssVal || cssVal == "0" || cssVal == "auto");
            cursorStyle += "z-index:100";
        
		_setFullStyle(cursorDiv, "position:absolute; width:2px;visibility:hidden;" + cursorStyle);
        
        // Register to events
        _addHandler(window, "resize", onWindowResized);
        _addHandler(document, "keystroke", onKeyStroke);

        // Make as subject
        _model(de.cursor);
        
        // Keep cursor in view after actions are executed
        de.UndoMan.addObserver({
            onAfterExec : scrollToCursor,
            onAfterUndo : scrollToCursor,
            onAfterRedo : scrollToCursor
        });
        
    }, "UndoMan");

    /**
     * @namespace The cursor namespace packages cursor specific operations.
     */
    de.cursor = {
    
        /**
         * @class 
         * Provides flag constants for describing the cursor relation to the dom node.
         */
        PlacementFlag : {
    
            /**
             * Read Only: Refer to specification 2.2.1. AKA Text nodes (containing least one renderable symbol)
             * @type Number
             */
            INSIDE : 1, // @REPLACE de.cursor.PlacementFlag.INSIDE 1
            
            /**
             * Read Only: Refer to specification 2.2.2
             * @type Number
             */
            BEFORE : 2, // @REPLACE de.cursor.PlacementFlag.BEFORE 2
            
            /**
             * Read Only: Refer to specification 2.2.3
             * @type Number
             */
            AFTER : 4 // @REPLACE de.cursor.PlacementFlag.AFTER 4
        },
        
        /**
         * @param {Node} domNode a dom node to test
         * @return {Boolean} True iff the dom node can support a cursor placed by it.
         */
        doesNodeSupportCursor : function(domNode) {
            return !de.doc.isProtectedNode(domNode) && de.doc.isNodeEditable(domNode);
        },
        
        /**
         * Sets the new cursor. The position is updated immediatly.
         * Set to null to hide/destroy the cursor.
         *
         * If the cursor is not in an editable area, then it will be set to null
         *
         * @param {de.cursor.CursorDescriptor} cursorDesc The new cursor. Null for no cursor.
         *
         * @return {Boolean} True if there is a cursor after the operation. False if the
         * 			operation resulted in no cursor.
         */
        setCursor: function(cursorDesc){
            
            // Dissallow cursor placement at protected nodes
            if (cursorDesc && !this.doesNodeSupportCursor(cursorDesc.domNode))
                cursorDesc = null;

            // Set the new cursor info
            currentCursorDesc = cursorDesc;

            // Stop and hide the cursor blink
            cursorBlink(false);

            // Update cursor GUI
            if (currentCursorDesc) {
                
                // Update cursor position
                cursorDiv.style.left = (currentCursorDesc.docLeft +
                    (currentCursorDesc.isRightOf ? currentCursorDesc.width : 0)) + "px";
                cursorDiv.style.top = currentCursorDesc.docTop + "px";
                cursorDiv.style.height = currentCursorDesc.height + "px";

                // Determine color for the cursor.
                var color = _getComputedStyle(currentCursorDesc.domNode, "color");
                cursorDiv.style.backgroundColor = color ? color : "black";
                
                // Begin cursor blink
                cursorBlink(true);
            }
            
            // Always reset cursorXAlign
            cursorXAlign = null;

            // Notify observers of cursor change
            this.fireEvent("CursorChanged", this.getCurrentCursorDesc());
            
            return currentCursorDesc != null;
            
        },
        
            
        /**
         * Discovers the closest matching cursor descriptor for a given position.
         * 
         * @param {Number} targetX The X coordinate in the window from where to find the closest
         * 							cursor position.
         *
         * @param {Number} targetY The Y coordinate in the window from where to find the closest
         * 							cursor position.
         * 
         * @param {Node} targetNode (optional) The node at the given position. If not provided this will be determined for you.
         *
         * @return {de.cursor.CursorDescriptor}	A CursorInfo object containing the closest cursor position
         * 			to the given target coordinates. Null if there are no valid nearby
         * 			cursor positions.
         */
        getCursorDescAtXY: getCursorDescAtXY,
        
        /**
         * @return {de.cursor.CursorDescriptor} The clone of the current cursor info object. 
         * 										Null if there is none.
         */
        getCurrentCursorDesc: function(){
            return currentCursorDesc ? _clone(currentCursorDesc) : null;
        },

        /**
         * Updates the cursor GUI.
         */
        refreshCursor: function(){
            if (!currentCursorDesc) 				
            	return;
            
            // IE Has this annoying event-threading model where when querrying some
            // dom objects' properties IE instantly raises an event on the same trace of execution!
            // This causes a nasty bug when using any of the cursor alrgorithms, this function
            // is 'randomly' invoked during critical sections which depend on the measuring nodes -
            // since a resize event is invoked due to querrying various spatial information on dom objects,
            // such as getting the document.scrollLeft for rereiving a elements absolute position in the window.
            // Work around:
            if (measurePostTextNode) return ;

            // Get new position
            setSpatialMembers(currentCursorDesc);
            
            // Update the gui
            cursorDiv.style.left = (currentCursorDesc.docLeft +
                (currentCursorDesc.isRightOf ? currentCursorDesc.width : 0)) + "px";
            cursorDiv.style.top = currentCursorDesc.docTop + "px";
            cursorDiv.style.height = currentCursorDesc.height + "px";
            
        },
        
        /**
         * Scrolls the document to view the current cursor
         */
        scrollToCursor : scrollToCursor,
        
        /**
         * @return {Boolean} True if the cursor is shown. False if no cursor exists.
         */
        exists: function(){
            return currentCursorDesc != null;
        },
        
        /**
         * @param {Node} node A DOM node to test
         * @return {Boolean} True if node is the cursor blinker element.
         */
        isCursorEle : function (node) {
            return node == cursorDiv;
        },
        
        /**
         * A wrapper function for _getRenderedNodeAtXY - excludes
         * @param {Number} x The x pixel coordinate relative to the window.
         * @param {Number} y The x pixel coordinate relative to the window.
         * @return {Node}    The node at the given window position - gauranteed not to be the cursor blinker.
         */
        getNonCursorNodeAtXY : function(x,y) {
            var dval = cursorDiv.style.display;
            cursorDiv.style.display = "none"; // Ensure not visible
            var targetNode = _getRenderedNodeAtXY(x, y);
            cursorDiv.style.display = dval; // Restore original value
            return targetNode;
        },
        
        /**
         * Gets the next cursor descriptor before or after a given cursor descriptor. I.E. The next physical
         * move of the cursor.
         * 
         * @param {de.cursor.CursorDescriptor}  cursorDesc The cursor descriptor to reference from.
         *                                      it does not have to be a valid cursor descriptor, that is, a
         *                                      node which the cursor cannot be place in/next to.
         * 
         * @param {Boolean} left                True to get cursor desc left of reference point, false for right.
         * 
         * @return {de.cursor.CursorDescriptor} The next cursor move from the reference point. Null if there is none.
         *                                      Note that it might be outside of an editable section.
         * 
         */
        getNextCursorMovement : getNextCursorMovement,

        /**
         * Note: only uses y and height members
         * 
         * @param {de.cursor.CursorDescriptor} desc1 A cursor to compare
         *
         * @param {de.cursor.CursorDescriptor} desc2 A cursor to compare
         *
         * @return {Boolean} True iff desc1 and desc2 are on the same line.
         */
        isOnSameLine: isOnSameLine,
        
        /**
         * Builds a cursor descriptor (provides the spatial information).
         *
         * @param {Node} domNode The dom node
         *
         * @param {Number} relIndex The index within the dom node (only applies for text nodes)
         *
         * @param {Boolean} isRightOf True if the cursor is to the right of index position.
         * 
         * @return {de.cursor.CursorDescriptor} The cursor descriptor. Null If a cursor cannot be placed at the given position.
         *                                      Note: if the dom not is a text node but the index is not rendered then null will be returned.
         * 
         */
        createCursorDesc: function(domNode, relIndex, isRightOf){
            
            var placement = getPlacementFlags(domNode);
            
            if (placement == 0 || (placement != de.cursor.PlacementFlag.INSIDE &&
                ((isRightOf && !(placement & de.cursor.PlacementFlag.AFTER))
                || (!isRightOf && !(placement & de.cursor.PlacementFlag.BEFORE)))))
                return null;
            
            var desc = {
                domNode: domNode,
                relIndex: relIndex,
                isRightOf: isRightOf,
                placement: placement
            };
            
            setSpatialMembers(desc);
            
            // If the text node / index is not rendered then return null
            if (placement == de.cursor.PlacementFlag.INSIDE && (desc.width == 0 || desc.height == 0))
                return null;
            
            return desc;
            
        },
        
        /**
         * Note: even if de.cursor.PlacementFlag.INSIDE is returned, the text node may
         * not be able to support a cursor being placed inside if it contains nothing but
         * non-renderable symbols.
         * 
         * @param {Node} node A dom node to get the placement flags for
         * 
         * @return {Number} A bitwise or combination of de.cursor.PlacementFlag's. 
         * 					Zero if it is not a candidate for supporting a cursor placement.
         */
         getPlacementFlags : getPlacementFlags,
         
         /**
          * Gets the nearest cursor descriptor to the given position information.
          * 
          * @param {Node} domNode A dom node in the document
          * 
          * @param {Number} relIndex A relative index in the dom node
          * 
          * @param {Boolean} isRightOf True if the cursor is placed to the right of the node/index.
          *                            False for left placement.
          * 
          * @param {Boolean} searchLeft True to search for nearest cursor to the left of the given position.
          *                             False to search right.
          * 
          * @return {de.cursor.CursorDescriptor} A cursor descriptor nearest to the given position. Null if there is none.
          */
         getNearestCursorDesc: function(domNode, relIndex, isRightOf, searchLeft){
             
			 if (_nodeName(domNode) == "br") 
			 	isRightOf = true;
			 
            var cDesc = de.cursor.createCursorDesc(domNode, relIndex, isRightOf); // Gives null if invalid request
            
            // Special case: line breaks
            if (cDesc && _nodeName(domNode) == "br") {
                
                // Get the cursor pos to the left and right of this line break
                var leftCDesc = getNextCursorMovement(cDesc, true),
                    rightCDesc = getNextCursorMovement(cDesc, false);
                   
                // If this line break is on its own line then return it
                if ((!leftCDesc || !isOnSameLine(leftCDesc, cDesc)) && (!rightCDesc || !isOnSameLine(rightCDesc, cDesc))) 
                    return cDesc;
                
                // Otherwise return the cursor pos to the left/right of the break - if it exists
                return (searchLeft ? leftCDesc : rightCDesc) || cDesc;
            }
            
            return cDesc || getNextCursorMovement({
                     domNode: domNode,
                     isRightOf: isRightOf,
                     relIndex: relIndex,
                     y : _getPositionInWindow(domNode).y}, searchLeft);

         } // End getNearestCursorDesc function
    
    }; // End Cursor Namespace
    
    /* -------------------------------------------------------------------------------------- */
    // Events
    /* -------------------------------------------------------------------------------------- */
    
    function onKeyStroke(e, normalizedKey) {
        
        if (!currentCursorDesc || e.ctrlKey || e.metaKey || e.altKey) return;
        
        var moveAction = 0;
        
        switch(normalizedKey) {
            case "Left": // Arrow left
            case "Right": // Arrow right
            
                moveAction = 1;
                
                // Attempt to move the cursor left/right
                var neighbour = getNextCursorMovement(currentCursorDesc, normalizedKey == "Left");
                
                // If the neighbour is null, then reached end point
                if (!neighbour) neighbour = currentCursorDesc;
                
                var oldCDesc = currentCursorDesc;
               
                if (!de.cursor.setCursor(neighbour) && oldCDesc) 
                    // If failed to set the cursor (probably due to it being outside
                    // of editable section) then set back to the old cursor position
                    de.cursor.setCursor(oldCDesc)
    
                break;
                
            case "Up": // Arrow up
            case "Down": // Arrow down
            
                moveAction = 1;
            
                var isUpward = normalizedKey == "Up";
                
                var docScrollPos = _getDocumentScrollPos();
  
                var xAlign = cursorXAlign ? cursorXAlign : (currentCursorDesc.docLeft - docScrollPos.left) +
                    (currentCursorDesc.isRightOf ? currentCursorDesc.width : 0);     
                    
                var searchSpace = getUpDownSearchSpace();
                var cDesc;
                
                if (searchSpace) {
                    // Perform specialized/narrowed dual binary search to discover cursor position directly above/below
                    cDesc = searchBestCursorPos(xAlign, currentCursorDesc.docTop - docScrollPos.top, searchSpace, currentCursorDesc, isUpward);
                    restoreMeasuringNodes();
                }

                if (!cDesc) cDesc = currentCursorDesc;
               
                var oldCDesc = currentCursorDesc;
               
                if (!de.cursor.setCursor(cDesc) && oldCDesc) 
                    // If failed to set the cursor (probably due to it being outside
                    // of editable section) then set back to the old cursor position
                    de.cursor.setCursor(oldCDesc)

                // Remember for next up/down movement
                cursorXAlign = xAlign;

                break;
                
        } // End case
        
        if (moveAction) { 
            scrollToCursor(); // Auto scroll
            return false; // Consume key event
        }
        
        return true;
       
        // Inner support functions to follow
    
        /**
         * An inner supporting function.
         * 
         * @return {Object} The search space for a dual binary search
         */
        function getUpDownSearchSpace(){
        
            // Build array of all nodes to search inside of targetEl
            var nodesToSearch = [];
            var pendingBANodes = [];
            var nextLine;
    
            _visitAllNodes(docBody, currentCursorDesc.domNode, !isUpward, function(domNode) {
            
                // Add in any pending before/after nodes
                appendPendingBAnodes(domNode);
                
                var placementFlags = getPlacementFlags(domNode);
                
                if (placementFlags == 0)                 
                    return true; // Cursor cannot be placed in this node
                    
                // Check to see if found a new line
                var pos, height;
                if (domNode.nodeType == Node.TEXT_NODE) {
                    pos = _getPositionInWindow(domNode.parentNode);
                    height = domNode.parentNode.offsetHeight;
                } else if (_nodeName(domNode) == "br") {
                    pos = measureLineBreak(domNode)
                    height = pos.height;
                } else {
                    pos = _getPositionInWindow(domNode);
                    height = domNode.offsetHeight;
                }
                
                if (!isOnSameLine(nextLine ? nextLine : currentCursorDesc, {y : pos.y, height : height})) {
                    // If already found the next line then the searchspace can end here
                    if (nextLine) return false;
                    
                    // If this is the first node that is in the next line, then remeber line information
                    nextLine = {y : pos.y, height : height};
                }
    
                if (placementFlags == de.cursor.PlacementFlag.INSIDE) { // AKA a text node
                    nodesToSearch.push({
                        domNode: domNode,
                        placement: de.cursor.PlacementFlag.INSIDE
                    });
    
                } else { // Before and/or After
                
                    if (placementFlags & de.cursor.PlacementFlag.BEFORE) {
                        if (isUpward) {
                            pendingBANodes.push(domNode);
                        } else {
                            nodesToSearch.push({
                                domNode: domNode,
                                placement: de.cursor.PlacementFlag.BEFORE // Only store before flag
                            });
                        }
                    }
                    
    	            if (placementFlags & de.cursor.PlacementFlag.AFTER) {
                        if (isUpward) {
                            nodesToSearch.push({
                                domNode: domNode,
                                placement: de.cursor.PlacementFlag.AFTER // Only store before flag
                            });
                        } else {
                            pendingBANodes.push(domNode);
                        }
                    }
                    
                }
                
            });
            
            // Was the next line even found?
            if (!nextLine) return null;
    
            // Append any pending before/after nodes
            appendPendingBAnodes(null);
            
            // Make sure search space is top-down
            if (isUpward) nodesToSearch.reverse();
            
            // Suppliment search space with index information
            var totalPlacementLength = 0;
            for (var i in nodesToSearch) {
                var node = nodesToSearch[i];
                var len = _nodeLength(node.domNode, 1);
                node.startIndex = totalPlacementLength;
    	        node.endIndex = node.startIndex + (len - 1);
                node.length = len;
                totalPlacementLength += len;
            }
    
            return {
                nodes: nodesToSearch,
                totalLength: totalPlacementLength
            };
            
            /**
             * Adds pending before/after nodes to the nodesToSearch
             * @param {Node} domNode the current dom node or null
             */
            function appendPendingBAnodes(domNode){
            
                while (pendingBANodes.length > 0) {
                
                    if (!domNode || !_isAncestor(pendingBANodes[pendingBANodes.length - 1], domNode)) {
                        nodesToSearch.push({
                            domNode: pendingBANodes.pop(),
                            placement: isUpward ? de.cursor.PlacementFlag.BEFORE : de.cursor.PlacementFlag.AFTER
                        });
                        
                    } else 
                        break;
                }
                
            } // End inner appendPendingBAnodes
            
        } // End inner getUpDownSearchSpace

    
    } // End onKeyStroke
    
    /**
     * If the window resized, the cursor must be re-positioned.
     * 
     * @param {Event} e
     */
    function onWindowResized(e) {
        de.cursor.refreshCursor();
    }
    
    
    /* -------------------------------------------------------------------------------------- */
    // Support functions for cursor GUI
    /* -------------------------------------------------------------------------------------- */

    /**
     * @param {Boolean} visible True to toggle cursor div to visible, false to hidden.
     */
    function setCursorVisible(visible) {
        cursorDiv.style.visibility = visible ? "visible" : "hidden";
    }
    /**
     * @return {Boolean} True if the cursor div is visisble.
     */
    function isCursorVisible() {
        return cursorDiv.style.visibility == "visible";
    }
    
    /**
     * Continuously toggles the cursor div's visibility over time
     * @param {Boolean} on True to turn it on, false to turn it off and hide it.
     */
    function cursorBlink(on) {
        
        if (typeof on != "undefined")
            cursorBlink.on = on;
            
        if (cursorBlink.on) {
            setCursorVisible(!isCursorVisible());
            cursorBlinkTOId = setTimeout(cursorBlink, CURSOR_BLINK_MS_TIME, 1); // NB: IE does not pass arguments
        } else if (cursorBlinkTOId) {
            clearTimeout(cursorBlinkTOId);
            cursorBlinkTOId = null;
            setCursorVisible(false);
        }
    }
    
    
    /** 
     * Scrolls the document to the current cursor's position
     */
    function scrollToCursor() {
        
        if (!currentCursorDesc) return;

        var viewPortSize = _getViewPortSize(),
            dx = 0, 
            dy = 0;
        
        // Get dy
        if ((currentCursorDesc.y + currentCursorDesc.height) >= viewPortSize.height) {
            dy = (currentCursorDesc.y + currentCursorDesc.height) - viewPortSize.height;
        } else if (currentCursorDesc.y < 0) {
            dy = currentCursorDesc.y;
        }
        
        // get dx
        var xpos = currentCursorDesc.x + (currentCursorDesc.isRightOf ? currentCursorDesc.width : 0) + parseInt(cursorDiv.style.width);
        if (xpos >= viewPortSize.width) {
            dx = xpos - viewPortSize.width;
        } else if (currentCursorDesc.x < 0) {
            dx = currentCursorDesc.x;
        }

        if (dx || dy) window.scrollBy(dx, dy);
       
    }
    
    /* -------------------------------------------------------------------------------------- */
    // Support functions for locating cursor information
    /* -------------------------------------------------------------------------------------- */

    // See de.cursor.getPlacementFlags doc
    function getPlacementFlags(node) {

        // Always place cursor after the packages.
        // NOTE: Testing this first to avoid placing cursor in placeholders within the packaged nodes.
        var pcon = de.doc.getPackageContainer(node);
        if (pcon) {
            var pflags = 0;
            if (pcon == node)
                pflags = de.cursor.PlacementFlag.BEFORE | de.cursor.PlacementFlag.AFTER;
            return pflags;
        } 
        
        // Cursors can be placed before placeholders, but not inside of them
        if (de.doc.isESPlaceHolder(node, false)) 
            return (de.doc.isESPlaceHolder(node, true)) ? de.cursor.PlacementFlag.BEFORE : 0;
        
        if (de.doc.isMNPlaceHolder(node, false)) 
            return (de.doc.isMNPlaceHolder(node, true)) ? de.cursor.PlacementFlag.BEFORE : 0;

        if (node.nodeType == Node.TEXT_NODE) {
            
            if (_doesTextSupportNonWS(node))
                return de.cursor.PlacementFlag.INSIDE;
                
            return 0;
        }
        
        var flags = 0;
        
        // TODO: REFACTOR/DOCUMENT
        // Apply user flag funciton if specified
        if (de.cursor.usrGetPlacementFlags) {
            flags = de.cursor.usrGetPlacementFlags(node);
            if (flags === $undefined)
                flags = 0;
            else return flags;
        }
        
        if (beforeElements[_nodeName(node)]) 
            flags = de.cursor.PlacementFlag.BEFORE;
            
        if (afterElements[_nodeName(node)]) 
            flags |= de.cursor.PlacementFlag.AFTER;
            
        return flags;
    }

    /**
     * Determines whether cinf1 is closer to target than cinf2. Uses adx/ady information
     *
     * @param {de.cursor.CursorDescriptor} desc1 A cursor to compare
     * @param {de.cursor.CursorDescriptor} desc2 A cursor to compare
     * @param {Number} targetY The target Y (used for boundry cases)
     *
     * @return True if desc1 is closer than desc2.
     */
    function isCloserToTarget(desc1, desc2, targetY){
        
        // If they are on the same line...
        if (isOnSameLine(desc1, desc2)) {
            // Then compare there absolute delta x's
            return desc1.adx < desc2.adx;
        } else if (desc1.ady == desc2.ady) {
            // Else if not on the same line AND they have the same absolute delta y's
            // then choose the closest to target Y from in the middle of their height
            return Math.abs((desc1.y + (desc1.height / 2)) - targetY) < Math.abs((desc2.y + (desc2.height / 2)) - targetY);
        }
        
        // They do not accur on the same line, and have diffent ady's... so pick closest to target Y
        return desc1.ady < desc2.ady;
    }
    
    
    /**
     * Determines whether desc1 is closer to (x, y) than desc2.
     *
     * @param {de.cursor.CursorDescriptor} desc1 A cursor to compare
     * @param {de.cursor.CursorDescriptor} desc2 A cursor to compare
     * @param {Number} tx The target x-coord
     * @param {Number} ty The target y-coor
     *
     * @return True if desc1 is closer than desc2.
     */
    function isCloserToXY(desc1, desc2, x, y){
    
        // If they are on the same line then compare there absolute delta x's
        if (isOnSameLine(desc1, desc2)) 
            return Math.abs(desc1.x - x) < Math.abs(desc2.x - x);
  
        var ady1 = Math.min(Math.abs(desc1.y - y), Math.abs((desc1.y + desc1.height) - y));
        var ady2 = Math.min(Math.abs(desc2.y - y), Math.abs((desc2.y + desc2.height) - y));
        
        // Else if not on the same line AND they have the same absolute delta y's
        // then choose the closest to target Y from in the middle of their height
        if (ady1 == ady2) 
            return Math.abs((desc1.y + (desc1.height / 2)) - y) < Math.abs((desc2.y + (desc2.height / 2)) - y);

        // They do not accur on the same line, and have diffent ady's... so pick closest to target Y
        return ady1 < ady2;
    }
    
    // See de.cursor.isOnSameLine jsdoc
    function isOnSameLine(desc1, desc2){
        return (desc1.y >= desc2.y && desc1.y < (desc2.y + desc2.height)) ||
            (desc2.y >= desc1.y && desc2.y < (desc1.y + desc1.height));
    }

    /**
     * If there are any elements/nodes in the document that were used for measurement
     * purposes (via setupMeasuringNodes).
     */
    function restoreMeasuringNodes() {
        if (measurePostTextNode) {
            measurePostTextNode.parentNode.removeChild(measureSpanEl);
            measurePostTextNode.parentNode.removeChild(measurePreTextNode);
            measurePostTextNode.nodeValue = measureFullText;
            measurePostTextNode = null;
        }
    }
    
    /**
     * Lazily sets up measuring elements/nodes in document.
     * Be sure to call restoreMeasuringNodes if you want the nodes to be removed
     * (i.e. The DOM document to return to it's original state)
     * 
     * @param {Node} textNode The text-node for which measurements are to take place.
     */
    function setupMeasuringNodes(textNode){
    
        if (measurePostTextNode == textNode) 
            return;
        if (measurePostTextNode) 
            restoreMeasuringNodes();
        
        measurePostTextNode = textNode;
        measureFullText = textNode.nodeValue;
        
        // Split the text node into 3 nodes (including self)
        textNode.parentNode.insertBefore(measureSpanEl, textNode);
        textNode.parentNode.insertBefore(measurePreTextNode, measureSpanEl);
    }
    
    /**
     * Requires that setupMeasuringNodes has been invoked.
     * Isolates measure nodes so that the measureSpanEl encapsulates a single charactor
     * at a given index.
     * 
     * @param {Number} index The index of the charactor to isolate.
     * @return True if the isolated charactor is a renderable symbol.
     */
    function measureCharactor(index) {
        measurePreTextNode.nodeValue = measureFullText.substr(0, index);
        measureSpanTextNode.nodeValue = measureFullText.charAt(index);
        measurePostTextNode.nodeValue = measureFullText.substr(index + 1);
        return measureSpanEl.offsetHeight != 0 && measureSpanEl.offsetWidth != 0;
    }
    
    /**
     * Measures a line break element. 
     * Restores any measuring nodes before and after operation.
     * 
     * @param {Node} lb A line break element to measure.
     * @return {Object} A tuple containing the position of the
     *                  line break in the window, and its height.
     */
    function measureLineBreak(lb) {
        
        restoreMeasuringNodes();
        
        // Flag this to signify that meaure nodes are occupied.
        measurePostTextNode = {};
        
        measureSpanTextNode.nodeValue = _NBSP;
        _insertAfter(measureSpanEl, lb);
        
        var spatInf = _getPositionInWindow(measureSpanEl);
        spatInf.height = measureSpanEl.offsetHeight;
        
        measureSpanEl.parentNode.removeChild(measureSpanEl);
        
        // Reset flag
        measurePostTextNode = null;
        
        return spatInf;
    }
    
    /**
     * Remeasures a cInfo's position (window and document) at its given
     * textnode / charactor, index - and updates the info's properties
     * accordingly.
     * 
     * @param {de.cursor.CursorDescriptor} cursorDesc
     */
    function setSpatialMembers(cursorDesc) {
        
        if (!cursorDesc) return;
        
        if (_nodeName(cursorDesc.domNode) == "br") {
            var inf = measureLineBreak(cursorDesc.domNode);
            cursorDesc.x = inf.x;
            cursorDesc.y = inf.y;
            cursorDesc.height = inf.height;
            cursorDesc.width = 0;
            
        } else {
        
            var spatEl;
            
            // Determine which element to get the spatial info from
            if (cursorDesc.placement == de.cursor.PlacementFlag.INSIDE) {
            
                setupMeasuringNodes(cursorDesc.domNode);
                
                measureCharactor(cursorDesc.relIndex);
                
                spatEl = measureSpanEl;
                
            } else 
                spatEl = cursorDesc.domNode;
            
            var pos = _getPositionInWindow(spatEl);
            cursorDesc.x = pos.x;
            cursorDesc.y = pos.y;
            cursorDesc.width = spatEl.offsetWidth;
            cursorDesc.height = spatEl.offsetHeight;
        
        }
        
        var docScrollPos = _getDocumentScrollPos();
        cursorDesc.docLeft = docScrollPos.left + cursorDesc.x;
        cursorDesc.docTop = docScrollPos.top + cursorDesc.y;
        
        restoreMeasuringNodes();
    }
    
    // see de.cursor.getNextCursorMovement
    function getNextCursorMovement(srcCDesc, left) {
        
        var startNode = srcCDesc.domNode, 
            startIndex = srcCDesc.relIndex,
            startIsRightOf = srcCDesc.isRightOf, 
            nextNode,
            nextIndex,
            nextIsRightOf,
            placementFlags = getPlacementFlags(srcCDesc.domNode),
            lastVisitedNode,
            pendingLineBreak,
            seenBlockElement = false,
            prevTextInfo;
        
        if (placementFlags == de.cursor.PlacementFlag.INSIDE) {
            
             // If needs simple flip of is rightof flag then return a flipped version
             if (srcCDesc.isRightOf == left) {
                 var cDesc = de.cursor.createCursorDesc(srcCDesc.domNode, srcCDesc.relIndex, !srcCDesc.isRightOf);
                 if (cDesc) return cDesc;
             }
            
        } else if (srcCDesc.isRightOf && placementFlags != de.cursor.PlacementFlag.INSIDE) {
        
            if (left) {
                
                // If scanning left but source cursor is to the right of an element, then need to start the
                // traversal within the elements deepest-right descendant
                while (startNode.lastChild) {
                    startNode = startNode.lastChild;
                }
                if (startNode != srcCDesc.domNode) {
                    startIndex = _nodeLength(startNode, 2) - 1;
                    startIsRightOf = true;
                }
                
            }
        } 
              
        // Begin traversing from source point inclusive
        _visitAllNodes(docBody, startNode, !left, function(domNode) {
            
            var firstVisit = domNode == startNode;
           
            // Skip node that are not displayed / is protected
            if (!_isNodeDisplayed(domNode) || de.doc.isProtectedNode(domNode)) {
                lastVisitedNode = domNode;
                return true;
            }
            
            if (!seenBlockElement && !firstVisit)
                seenBlockElement = _isBlockLevel(domNode);
                
            placementFlags = getPlacementFlags(domNode);

            // Check after nodes
            if (lastVisitedNode) {
                var commonAncestor = _getCommonAncestor(domNode, lastVisitedNode, true);
                var checkANodes = _getAncestors(left ? domNode : lastVisitedNode, commonAncestor, false);
                if (left) { // Need to search from top-down in left search
                    checkANodes.reverse();
                    // Include the current node in the search if it is an after node - except for line breaks
                   if ((placementFlags & de.cursor.PlacementFlag.AFTER) 
                       && !_isAncestor(domNode, lastVisitedNode)
                       &&  _nodeName(domNode) != "br")
                        checkANodes.push(domNode);
                } else {
                    if (!_isAncestor(lastVisitedNode, domNode)
                        && _nodeName(lastVisitedNode) != "br"
                        && !de.doc.isProtectedNode(lastVisitedNode)
                        && !(lastVisitedNode == startNode && startIsRightOf))
                        checkANodes.push(lastVisitedNode);
                }
            
                for (var i in checkANodes) {
                    var node = checkANodes[i];
                    
                    // Is this node an actual after node?
                    if (getPlacementFlags(node) & de.cursor.PlacementFlag.AFTER) {
                        
                        // Check for pending line break
                        if (checkLineBreak(node))
                            return false;
    
                        // Found the next move
                        nextNode = node;
                        nextIndex = 1;
                        nextIsRightOf = true;
                        return false;
                    }
                    
                    // Update block level flag
                    seenBlockElement |= _isBlockLevel(node);
                }
            }
            
            // Is a dom node which is needing a placeholder?
            if (placementFlags == 0) {
                
                if (!de.doc.isNodePackaged(domNode)) {
                    var phType = 0
                    if (_doesNeedESPlaceholder(domNode)) 
                        phType = 1;
                    else if (_doesNeedMNPlaceholder(domNode)) phType = 2;
                    
                    if (phType) {
                    
                        // Create missing placeholder and add it.. prevent the undo manager
                        // from setting the cursor and if there is any undo history then
                        // group this with the last action... and if the dontStoreInsertPHOps option is set then
                        // prevent the undo manager from storing the operations all together
                        de.UndoMan.execute(de.UndoMan.hasUndo() ? de.UndoMan.ExecFlag.GROUP : 0, "InsertHTML", _getOuterHTML(phType == 1 ? de.doc.createESPlaceholder(domNode) : de.doc.createMNPlaceholder()), domNode, domNode.firstChild, 0);
                        
                        // Check for pending linebreak first
                        if (!checkLineBreak(domNode.firstChild)) {
                            // Set next cursor point to the created placeholder
                            nextNode = domNode.firstChild;
                            nextIndex = 0;
                            nextIsRightOf = false;
                        }
                        return false;
                        
                    }
                }
                
            } else if (placementFlags == de.cursor.PlacementFlag.INSIDE) { // AKA A text node
        
                setupMeasuringNodes(domNode);
                
                var relIndex = firstVisit ? startIndex : (left ? _nodeLength(domNode) - 1 : 0),
                    isRightOf;
                    
                // Determine is rightof flag
                if (firstVisit)
                    isRightOf = startIsRightOf;
                else if (prevTextInfo) {
                    // If measured a text position (start point was a text node), 
                    // use its isrightof flag, unless seen a block level
                    // element.
                    if (seenBlockElement) {
                        prevTextInfo = null; // Reset to void checking for a line wrap
                        isRightOf = left;
                    } else isRightOf = prevTextInfo.isRightOf;
                } else isRightOf = left;
                
                // For each charactor in the run of text
                for (; (left && relIndex >= 0) ||
                (!left && relIndex < measureFullText.length); relIndex += left ? -1 : 1) {
                
                    // Measure the charactor. Skip non renderable charactors
                    if (!measureCharactor(relIndex)) {
                        if (firstVisit)
                            isRightOf = !isRightOf;
                        firstVisit = false;
                        continue;
                    }
                    
                    // Determine the charactor position
                    var inf = _getPositionInWindow(measureSpanEl);
                    inf.height = measureSpanEl.offsetHeight;

                    // If there is a pending line break, check to see if it
                    // occurs on the same line as the measured renderable charactor                    
                    if (checkLineBreak(measureSpanEl, inf))
                        return false;

                    // Check if the charactor is at the start or end of a line (line wrap start/end)
                    if (prevTextInfo) {
                        debug.assert(!firstVisit);
                        if (!isOnSameLine(inf, prevTextInfo)) {
                            nextNode = domNode;
                            nextIndex = relIndex;
                            nextIsRightOf = !isRightOf;
                            return false;
                        }
                    }
       
                    // If this isnt the starting point then found the next pos
                    if (!firstVisit || startNode != srcCDesc.domNode) {
                        nextNode = domNode;
                        nextIndex = relIndex;
                        nextIsRightOf = isRightOf;
                        return false;
                    }
                    // Note: no need to check to flip isRightOf flag since this is checked before traversal
                    
                    // Re-setup the measuring nodes (if needs to) due to line break measurements above
                    setupMeasuringNodes(domNode);
                    firstVisit = false;

                    // Set prevTextInfo for testing for line wraps on next text measure
                    prevTextInfo = {
                        domNode : domNode,
                        isRightOf: isRightOf,
                        y: inf.y,
                        height: inf.height
                    };
                    
                    // If the node is a placeholder, only count one movement
                    if (de.doc.isMNPlaceHolder(domNode) || de.doc.isESPlaceHolder(domNode)) {
                        prevDesc.isRightOf = false;
                        break;
                    }
   
                } // End loop: measuring each char in the run of text
                
                // Must restore nodes to avoid traversal going into measure nodes themselves
                restoreMeasuringNodes();
            
            } else if (placementFlags & de.cursor.PlacementFlag.BEFORE) {
            
                // Only set as before node if not the first visit or if searching left and began to right of the node.
                if(!firstVisit || (left && domNode == startNode && startIsRightOf)) {
                    if (!checkLineBreak(domNode)) {
                        nextNode = domNode;
                        nextIsRightOf = false;
                        nextIndex = 0;
                    }
                    return false;                    
                }
            
            } else if (_nodeName(domNode) == "br") { // Check for line breaks (Pure AFTER nodes)
                
                if (!firstVisit) {
                    
                    // Check for any pending line breaks
                    if (checkLineBreak(domNode))                         
                        return false;
                    
                    if (left) {
                        // For LEFT searches check for pending line breaks right away since the
                        // right-node information is always available
                        pendingLineBreak = domNode;
                        if (checkLineBreak(lastVisitedNode, prevTextInfo && prevTextInfo.domNode == lastVisitedNode ? prevTextInfo : null))                         
                            return false;
                    
                    } else { // Right
                        // Set new pending line break
                        pendingLineBreak = domNode;
                    }

                }
            }
            
            lastVisitedNode = domNode;
            
            // For right searches, if the cursor begins to the right on an element then avoid
            // traversing inside the descendants
            if (!left && firstVisit && srcCDesc.isRightOf && placementFlags != de.cursor.PlacementFlag.INSIDE)
                return 1;
            
        }); // End traversal
        
        restoreMeasuringNodes();
        
        // Found a cursor?
        return nextNode ? de.cursor.createCursorDesc(nextNode, nextIndex, nextIsRightOf) : null;

        /**
         * An inner support function. Determines whether a pending line break
         * should be the next cursor position
         * 
         * WARNING: This may restore any current measuring nodes!
         *
         * @param {Node} ele The current element to check pending line breaks against
         *
         * @param {Object} eleDesc Optional (created if not provided). A descriptor, with at least the y and height members set
         *
         * @return {Boolean} True iff the next move is a pending line break (in which case next node/index will be set).
         *
         */
        function checkLineBreak(ele, eleDesc) {

            // If there is a pending line break, check to see if it
            // occurs on the same line as this
            if (pendingLineBreak) {

                // Measure pending line break and element spatial qaulties            
                var lbMeas = measureLineBreak(pendingLineBreak), 
                    eleDesc = eleDesc || (_nodeName(ele) == "br" ? measureLineBreak(ele) : {
                    y: _getPositionInWindow(ele).y,
                    height: ele.offsetHeight
                });

                if (!isOnSameLine(lbMeas, eleDesc)) {
                    // If this before after node is not on the same line as the pending
                    // line break, then count the line break as part of the move.
                    nextNode = pendingLineBreak;
                    nextIndex = 1;
                    nextIsRightOf = true;
                    return true;
                }
                
                pendingLineBreak = null;
            }
            
            return false;
            
        } // End inner checkLineBreak function

    } // End getNextCursorMovement

    
    // See de.cursor.getCursorDescAtXY 
    function getCursorDescAtXY(targetX, targetY, targetNode){
        
	    if (!targetNode) 
            targetNode = _getRenderedNodeAtXY(targetX, targetY);
            
        // If the target was the cursor - recalc the target to the node behind the cursor.
        if (targetNode == cursorDiv) {
            cursorDiv.style.display = "none";
            targetNode = _getRenderedNodeAtXY(targetX, targetY);
            cursorDiv.style.display = "";
        }
        
        if (!targetNode) return null;
        
        // Get the searchspace for the bin search
        var searchSpace = getBinSearchSpace();
  
        // Perform the dual binary search 
        var cDesc = searchBestCursorPos(targetX, targetY, searchSpace);
        
        // Restore the DOM structure
        restoreMeasuringNodes();
        
        return cDesc;
        
        /**
         * An inner supporting function.
         * 
         * @return The search space for the binary search
         */
        function getBinSearchSpace(){
        
            // Build array of all nodes to search inside of targetEl
            var nodesToSearch = [],
                totalPlacementLength = 0,
                wndSize = _getWindowSize();
            
            (function traverse(domNode) {

                var placementFlags = getPlacementFlags(domNode),
                    checkElement,
                    posInWindow;
                
                // Avoid adding nodes which are not in the viewport
                if (domNode.nodeType == Node.ELEMENT_NODE) {
                    checkElement = domNode;
                } else if (domNode.nodeType == Node.TEXT_NODE && placementFlags != 0) {
                    setupMeasuringNodes(domNode);
                    measurePreTextNode.nodeValue = "";
                    measureSpanTextNode.nodeValue = measureFullText;
                    measurePostTextNode.nodeValue = "";
                    checkElement = measureSpanEl;
                }
                
                if (checkElement) {
                    
                    posInWindow = _nodeName(checkElement) == "br" ? 
                        measureLineBreak(checkElement) : 
                        _getPositionInWindow(checkElement);
                    
                    // Is this element above the veiw port?
                    if ((posInWindow.y + checkElement.offsetHeight) <= 0) {
                        restoreMeasuringNodes();
                        return true;
                    }
                    
                    // Is this element below the veiw port?
                    if (posInWindow.y > wndSize.height) {
                        restoreMeasuringNodes();
                        return false;
                    }
                    
                    restoreMeasuringNodes();
                }
                    
                if (placementFlags == de.cursor.PlacementFlag.INSIDE) { // AKA a text node
                    nodesToSearch.push({
                        domNode: domNode,
                        startIndex: totalPlacementLength,
                        endIndex: totalPlacementLength + _nodeLength(domNode) - 1,
                        length: _nodeLength(domNode),
                        placement: de.cursor.PlacementFlag.INSIDE
                    });
                    
                    totalPlacementLength += _nodeLength(domNode);
                    
                } else if (placementFlags & de.cursor.PlacementFlag.BEFORE) {
                    nodesToSearch.push({
                        domNode: domNode,
                        startIndex: totalPlacementLength,
                        endIndex: totalPlacementLength + 1,
                        length: 1,
                        placement: de.cursor.PlacementFlag.BEFORE, // Only store before flag
                        posInWnd : posInWindow // cache this
                    });
                    totalPlacementLength++;
                }
                
                // Recurse in order traversal
                var child = domNode.firstChild;
                var continueTrav = true;
                while(child) {
                    if (!traverse(child)) {
                        continueTrav = false; // Started to traverse in node below the viewport
                        break;
                    }
                    child = child.nextSibling;
                }
                
                // Add after nodes (event if aborting traversal)                
                if (placementFlags & de.cursor.PlacementFlag.AFTER) {
                    nodesToSearch.push({
                        domNode: domNode,
                        startIndex: totalPlacementLength,
                        endIndex: totalPlacementLength + 1,
                        length: 1,
                        placement: de.cursor.PlacementFlag.AFTER, // Only store after flag
                        posInWnd : posInWindow // cache this
                    });
                    totalPlacementLength++;
                    
                }
                
                return continueTrav;
                
            })(targetNode);
            
            return {
                nodes: nodesToSearch,
                totalLength: totalPlacementLength
            };
            
        } // End inner getBinSearchSpace
    
    } // End getCursorDescAtXY
    
    /**
     * IMPORTANT: Measurement nodes are left un-restored after this operation
     * Call restoreMeasuringNodes if you want to restore the dom.
     *
     * @param {Number} targetX The x coord to get the closest cusor pos to
     * @param {Number} targetY The y coord to get the closest cusor pos to
     * @param {Object} A dual bin search space to search.
     * @param {de.cursor.CursorDescriptor} targetLineRef   A reference point to search for a best position directly above or
     *                                                     below the line. Null for full search.
     *                                                     
     * @param {Boolean} aboveLine  If given targetLineRef then set to true to search for best position above the reference point.
     *                             Other false will search below the reference point.
     *                             
     * @return {de.cursor.CursorDescriptor} The closest cursor position it can find. Null if could not find one.
     *
     */
    function searchBestCursorPos(targetX, targetY, searchSpace, targetLineRef, aboveLine){
        
        // Hash table as an associative array, caches measurements
        var nonRenderables = {};

        if (searchSpace.totalLength == 0)         
            return null;
        
        // Sample first position in target element
        var startDesc = getCursorDescFrom(0, 0, 2);
        if (!startDesc)         
            return null;
        
        // Sample last charactor in target element
        var endDesc = getCursorDescFrom(searchSpace.nodes.length - 1, searchSpace.nodes[searchSpace.nodes.length - 1].length - 1, 1);
        
        debug.assert(endDesc != null);
        
        // Check to see if first and last samples are the same
        if (startDesc.domNode == endDesc.domNode &&
        startDesc.absIndex == endDesc.absIndex) {
            // There must be only 1 renderable charactor in the target element
            return validDescriptor(startDesc);
        }
        
        // Store the first samples
        var samples = [startDesc, endDesc];
        
        var best = null;
        
        // Determine closest sample and set as the current best
        best = isCloserToTarget(startDesc, endDesc, targetY) ? startDesc : endDesc;
        
        // Upper and lower are the bounds of the search space in the form of cursor descriptors 
        var upper, lower;
        
        // This algorithm has two passes: The first pass is a binary search to discover the line, or closest line,
        // that the target is on. The second pass is a binary search to home in on the closest charactor to the target
        // on the line that was found to be the best.
        for (var pass = 1; pass <= 2; pass++) { // dual binary search
            if (pass == 1) { // setup first pass: Y DOMAIN
            
                if (targetLineRef) {
                    
                    // If the binary search should find best matching position above or below a line reference point,
                    // then discover all lines within the search space
                    discoverAllLines(startDesc, getNodeIndex(startDesc.absIndex), endDesc, getNodeIndex(endDesc.absIndex));
                    
                    // Select the closest sample that does not fall on the line reference point
                    selectBest();
                    
                    // Now the the target line has been discovered, set the target Y
                    targetY = best.y + (best.height / 2);
                    
                    // Re-calc best abs delta y
                    best.ady = Math.abs(best.y - targetY);
                    
                    // Move to X-bin-search
                    continue;
                    
                } else {
                    // Select the upper and low for the Y range - goto next pass if already on target line
                    if (!selectYRange()) continue;
                }
                
            } else { // setup second pass: X DOMAIN
                var res = selectXRange();
                
                if (typeof res == "boolean") {
                    if (!res) break; // Else the upper and lower range has been selected
                } else break; // the best was found	
            }
            
            // Enter binary search for quickly locating closest line, or charactor
            while (true) {
            
                debug.assert(lower.absIndex < upper.absIndex);
            
                // If lower is next to upper and was doing the line search, then the line search is done... an exact match wasn't found 
                // and the binary search verged towards two charactors side by side but on different lines.                    
                // ...and if was doing the charactor search, then the search has verged at the final point
                if (lower.absIndex == (upper.absIndex - 1)) break;
                
                // Determine current index by halving the search space
                var curAbsIndex = lower.absIndex + Math.floor((upper.absIndex - lower.absIndex) / 2);
                
                // Ensure that the index is not out of bounds
                if (curAbsIndex == lower.absIndex) 
                    curAbsIndex++;
                else if (curAbsIndex == upper.absIndex) curAbsIndex--;
                
                // Locate which node within the target element that the current absolute index is in
                var curNodeIndex = getNodeIndex(curAbsIndex);
                
                // Get the cursor desc at the current node/rel-index
                var current = getCursorDescFrom(curNodeIndex, curAbsIndex - searchSpace.nodes[curNodeIndex].startIndex, 0);
  
                // Upper and Lower are next to each other
                if (!current) break;
                
                // In the first pass all samples are recorded - the initial upper and lower bounds of the next pass
                // will be salvaged from these samples.
                if (pass == 1) samples.push(current);
                
                // Check to see if current is the new best
                if (isCloserToTarget(current, best, targetY)) best = current;
                
                if (pass == 1) { // Line search
                    // Check to see if current matches line. 
                    if (targetY >= current.y && targetY <= (current.y + current.height)) 
                        break; // finished since found a match
    
                    // Narrow search
                    else if (current.y > targetY) 
                        upper = current;
                    else 
                        lower = current;
                    
                } else { // Charactor search
                    // Determine if current is on the same line as the best
                    if (isOnSameLine(current, best)) {
                    
                        // See if target X is on top of current
                        if (targetX >= current.x && targetX <= (current.x + current.width)) 
                            break; // If so, then search complete
    
                        // Otherwise narrow search based on x position
                        else if (current.x > targetX) 
                            upper = current;
                        else 
                            lower = current;
                        
                    }                // If not on same line, then narrow search based on Y coordinates
                    else if (current.y > targetY) 
                        upper = current;
                    else 
                        lower = current;
                    
                }
                
            } // End loop: core binary search for finding line and charactor
        } // End passes
        // FINISHED
        return validDescriptor(best);
        
        // Support functions to follow...
        
        /**
         * @param {Number} absIndex The abs index in the search-space
         *
         * @return {Number} The index within nodes that absIndex resides
         */
        function getNodeIndex(absIndex){
        
            var nodes = searchSpace.nodes;
            
            // If there is one text node, then clearly the current index is that node.
            if (nodes.length == 1)             
                return 0;
            
            // Is it in the first text node?
            if (absIndex >= nodes[0].startIndex && absIndex <= nodes[0].endIndex)             
                return 0;
            
            // Is in last text node?
            if (absIndex >= nodes[nodes.length - 1].startIndex &&
            absIndex <= nodes[nodes.length - 1].endIndex)             
                return nodes.length - 1;
            
            // begin a little binary search to quickely find the node in the search space
            var lo = 0;
            var up = nodes.length - 1;
            var nodeIndex;
            
            while (true) {
            
                var cur = lo + Math.floor((up - lo) / 2);
                if (cur == lo) 
                    cur++;
                else if (cur == up) cur--;
                
                if (absIndex >= nodes[cur].startIndex &&
                absIndex <= nodes[cur].endIndex) {
                    nodeIndex = cur;
                    break;
                    
                    
                } else if (absIndex < nodes[cur].startIndex) 
                    up = cur; // search downward	
    
                else 
                    lo = cur; // search upward
            } // End loop: binary search for locating #text node
            return nodeIndex;
            
        } // End inner getNodeIndex
        
        /**
         * Gets a cursor descriptor from a given position in the search space.
         *
         * Some text nodes in the search space may not support a cursor placement since
         * they can contain only non-renderable symbols. Therefore the returned cursor
         * desc may not be at the given node/index.
         *
         * @param {Number} dir 0 = Both, within upper and lower bounds, 1 = Left only, 2 = Right only.
         * @param {Number} relIndex The relative index
         * @param {Number} nodeIndex The index within the search space nodes
         *
         * @return {Object} the cursor descriptor or NULL if did not find a cursor at the given position that is in bounds.
         */
        function getCursorDescFrom(nodeIndex, relIndex, dir){
        
            var nodes = searchSpace.nodes,
            
                searchLeft = dir == 0 || dir == 1,
            
                // Store the current relative/node index
                // for restoring when switching scan direction
            	origialRelIndex = relIndex,
	            originalNodeIndex = nodeIndex,
            
                // Ideally we would directly measure spatial info at the current node / relative index.
                // However some charactors in text nodes aren't renderable, and thus we must scan left and/or right
                // to find next renderable charactor
                measureEl = null,
                node; // the node within the searchspace nodes
            
            // Find first renderable symbol. 1 pass for non-text nodes, 
            // 1-2 pass for text nodes and searching both dirs: 1st pass search left, 2nd pass search right.
            do {
            
                if (!searchLeft && dir == 0) { // 2nd pass?
                    // Switching direction...
                    nodeIndex = originalNodeIndex;
                    relIndex = origialRelIndex + 1; // exclusive
                    // Check to see if the right of the starting point is in a different node
                    if (relIndex > nodes[nodeIndex].endIndex) {
                        nodeIndex++;
                        relIndex = 0;
                        // Note: If the nodeIndex is out of bounds, the next loop will instantly break
                    }
                }
                
                var reachedSSBounds = false; // refers to search-space upper/lower bounds
                
                // Scan through the nodes
                while (nodeIndex >= 0 && nodeIndex < nodes.length) {
                
                    node = nodes[nodeIndex];
                    
                    // Ignore non-diplayed nodes
                    if (_isNodeDisplayed(node.domNode)) {
                    
                        // Discover the type of node
                        if (node.placement == de.cursor.PlacementFlag.INSIDE) { // AKA Text node
                            // Setup measurement nodes for this text node we are about to search in
                            setupMeasuringNodes(node.domNode);
                            
                            // Scan through charactors.. looking for the first renderable symbol
                            while (relIndex >= 0 && relIndex < measureFullText.length) {
                            
                                // Is the text node/index out of bounds (only when scanning both ways)?
                                if (dir == 0 &&
                                ((lower.domNode == node.domNode && lower.relIndex == relIndex) ||
                                (upper.domNode == node.domNode && upper.relIndex == relIndex))) {
                                    reachedSSBounds = true;
                                    break;
                                }
                                
                                // or have we measured this before and found it was not a renderable char?
                                if (nonRenderables["_" + nodeIndex + '_' + relIndex]) {
                                    relIndex += (searchLeft ? -1 : 1);
                                    continue; // avoid re-measuring unrenderable node
                                }
                                
                                // Measure the current node / index.
                                if (!measureCharactor(relIndex)) {
                                    // Char is not renderable, note the element and store in the 
                                    // hash table (using an accoiative array)
                                    nonRenderables["_" + nodeIndex + '_' + relIndex] = true;
                                    relIndex += (searchLeft ? -1 : 1);
                                    continue;
                                }
                                
                                // Determine position of the charactor. This will be used as a flag
                                // for ending the search for the first non-renderable charactor
                                measureEl = measureSpanEl;
                                break;
                                
                            } // End loop: Searching for renderable charactor
                           
                            
                        } else { // BEFORE and AFTER nodes
                            // Is the text node/index out of bounds (only when scanning both ways)?
                            if (dir == 0 &&
                            ((lower.domNode == node.domNode && lower.placement == node.placement) ||
                            (upper.domNode == node.domNode && upper.placement == node.placement))) {
                                reachedSSBounds = true;
                                break;
                            }
                            
                            measureEl = node.domNode;
                            break;
                        }
                    }
                    
                    // Have we found the next sample, or has the right-scan reached the upper search-space bound?
                    if (measureEl || reachedSSBounds) break;
                    
                    // Setup current node index for scanning the next node
                    nodeIndex += (searchLeft ? -1 : 1);
                    
                    // Setup relative index for searching for next renderable char
                    relIndex = (searchLeft &&
                        nodeIndex >= 0 &&
                        nodeIndex < nodes.length) ? 
                        (nodes[nodeIndex].placement == de.cursor.PlacementFlag.INSIDE ? nodes[nodeIndex].length - 1 : 1) : 0;
                    
                } // End loop: scanning nodes in a particular direction
                
                // Was a sample found?
                if (measureEl) break;
                
            } while (dir == 0 && !(searchLeft = !searchLeft)); // End loop: scanning left and/or right
            
            // If there were no places where the cursor can be placed between lower and upper, 
            // then the search is done.
            if (!measureEl)             
                return null;
                
            var pos, width, height;
            
            // Is their cached spatial calculations for this node?
            if (node.posInWnd) 
                pos = node.posInWnd;
                
            if (_nodeName(measureEl) == "br") {
                pos = pos || measureLineBreak(measureEl);
                height = pos.height;
                width = 0;
            } else {
                pos = pos || _getPositionInWindow(measureEl);
                width = measureEl.offsetWidth;
                height = measureEl.offsetHeight;
            }

            var adxl = Math.abs(pos.x - targetX),
                adxr = Math.abs(pos.x + width - targetX),
                isRightOf;
            
            switch (node.placement) { // searchspace placement flags are separated (cannot have combined flags)
                case de.cursor.PlacementFlag.BEFORE:
                    isRightOf = false;
                    break;
                    
                case de.cursor.PlacementFlag.AFTER:
                    isRightOf = true;
                    break;
                    
                default: // INSIDE/TEXT
                    isRightOf = adxr < adxl;
            }
            
            return {
                domNode: node.domNode,
                relIndex: relIndex,
                absIndex: node.startIndex + relIndex,
                placement: node.placement,
                isRightOf: isRightOf,
                x: pos.x,
                y: pos.y,
                adx: isRightOf ? adxr : adxl,
                ady: Math.min(Math.abs(pos.y - targetY), Math.abs(pos.y + measureEl.offsetHeight - targetY)),
                width: width,
                height: height
            };
            
        } // End inner getCursorDescFrom
        
        /**
         * @return {Boolean}
         * 			True if the range is set.
         * 			False if the start or end sample is on the targets line
         */
        function selectYRange(){
        
            // If the start or end sample is not on the target line, and the target is within the samples y-range, 
            // then find the best line
            if (!((targetY >= startDesc.y && targetY <= (startDesc.y + startDesc.height)) ||
            (targetY >= endDesc.y && targetY <= (endDesc.y + endDesc.height))) &&
            targetY >= startDesc.y &&
            targetY <= (endDesc.y + endDesc.height)) {
            
                // If the target Y does not occur on the start or end sample's line:
                lower = startDesc;
                upper = endDesc;
                
                return true;
            }
            
            return false; // Otherwise, the line-search is done! Next pass...
        } // End inner selectYRange
        /**
         * @return {Boolean, de.cursor.CursorDescriptor}
         * 			False if the best is right at the target.
         * 			Or True if the range has been selected.
         * 			Or A cursor descriptor of the best match if found while setting the range -
         * 			in which case the best var is set
         *
         */
        function selectXRange(){
        
            // Determine the lower and upper bounds to start the charactor binary search with
            if (targetX >= best.x && targetX < (best.x + best.width)) {
                // If the best is right at the target, then the search is done
                return false;
                
            } else if (best.x > targetX) { // search to left of best
                upper = best;
                lower = null;
                for (i in samples) {
                    current = samples[i];
                    /*
                     
                     if (current == best) continue;
                    
                    
                    var isLeftOrAbove;
                    
                    // Check if this sample (current) is to the left or above of best
                    if (isOnSameLine(current, best)) {
                        // If sample is on sample line as best, then check X coords
                        isLeftOrAbove = current.x < best.x;
                        
                    } else 
                        isLeftOrAbove = current.y < best.y; // Check Y coords if not on same line 
                    // If the sample is to the left, or above, of best. Then check to see if the sample
                    // is closer than the current lower to best.
                    if (isLeftOrAbove && (!lower || isCloserToXY(current, lower, best.x, best.y)) && current.absIndex < upper.absIndex) {
                        lower = current;
                    }*/
                   
                   // Select preceeding sample to best in sample set
                   if (current.absIndex < best.absIndex && (!lower || current.absIndex > lower.absIndex))
                       lower = current;
                    
                }
                
                if (!lower) {
                    // The best line must have been the starting sample, thus the best charactor is
                    // the first renderable charactor.
                    return best;
                }
                
                
            } else { // Search to the right of best
                lower = best;
                upper = null;
                for (i in samples) {
                
                    current = samples[i];
                    
                    /*var isRightOrBelow;
                    
                    // Check if this sample (current) is to the right or below of best
                    if (isOnSameLine(current, best)) {
                        // If sample is on sample line as best, then check X coords
                        isRightOrBelow = current.x > best.x;
                        
                    } else 
                        isRightOrBelow = current.y > best.y; // Check Y coords if not on same line 
                    // If the sample is to the right, or below, of best. Then check to see if the sample
                    // is closer than the current upper to best.
                    if (isRightOrBelow && (!upper || isCloserToXY(current, upper, (best.x + best.width), best.y)) && current.absIndex > lower.absIndex) {
                        upper = current;
                    }*/
                   
                    if (current.absIndex > best.absIndex && (!upper || current.absIndex < upper.absIndex))
                        upper = current;
                    
                }
                
                if (!upper) {
                    // The best line must have been the ending sample, thus the best charactor is
                    // the first renderable charactor.
                    return best;
                }
                
            }
            
            return true;
            
        } // End inner selectXRange
        
        /**
         * Sets the best local for searching for the closest cursor position before/after a line reference point.
         */
        function selectBest() {
            
            // Begin with the search space bounds... depending on whether the search should
            // look above or below the line reference point
            best = aboveLine ? startDesc : endDesc;
            
            // Look in all samples... which contain all lines
            for (var i in samples) {
                
                var sample = samples[i];
    
                // Skip samples that are either on the same line as the reference line, or
                // is out of bounds...            
                if (isOnSameLine(sample, targetLineRef) ||
                    (aboveLine && sample.y > targetLineRef.y) ||
                    (!aboveLine && sample.y < targetLineRef.y)) continue;

                // Set new best if sample is better
                if (isCloserToTarget(sample, best, targetY)) {
                    best = sample;
                }
                
            }

        } // End inner selectBest
        
       /**
        * Samples the search space to discover all lines.
        * 
        * @param {de.cursor.CursorDescriptor} lo A cursor desc
        * @param {de.cursor.CursorDescriptor} up A cursor desc
        * 
        * @param {Number} lni Lower node index in search space
        * @param {Number} uni Upper node index in search space
        */
        function discoverAllLines(lo, lni, up, uni){
            
            var stack = [[lo, lni, up, uni]];
            
            while(stack.length > 0) { // simulating recursion - faster and avoids stack overflows
                
                var args = stack.pop();
                lo = args[0];
                lni = args[1];
                up = args[2];
                uni = args[3];

                // Special attention must be payed to tables. They contain inner lines.
                var ni = lni;
                
                // Shrink lower range if lower is a table
                while (ni < searchSpace.nodes.length && _nodeName(searchSpace.nodes[ni].domNode) == "table") {
                    ni++;
                }
                
                if (ni == searchSpace.nodes.length) continue;
                
                if (ni != lni) {
                    lo = getCursorDescFrom(ni, 0, 2);
                    lni = ni;
                    samples.push(lo); // duplicates are ok
                }
       
                ni = uni;
                
                // Shrink upper range if upper is a table
                while (ni >= 0 && _nodeName(searchSpace.nodes[ni].domNode) == "table") {
                    ni--;
                }
                
                if (ni == -1) continue;
                
                if (ni != uni) {
                    up = getCursorDescFrom(ni, searchSpace.nodes[ni].length - 1, 1);
                    uni = ni;
                    samples.push(up); // duplicates are ok
                }
                
                // If lower is directly next to upper or they are on the same line 
                // then the line discovery between these two points is complete
                if (lo.absIndex >= (up.absIndex - 1) || isOnSameLine(up, lo)) continue;
                
                // Determine middle index by halving the search space
                var midIndex = lo.absIndex + Math.floor((up.absIndex - lo.absIndex) / 2);
                 
                // Ensure that the index is not out of bounds
                if (midIndex == lo.absIndex) 
                    midIndex++;
                else if (midIndex == up.absIndex) midIndex--;
                
                // Locate which node within the target element that the current absolute index is in
                var midNodeIndex = getNodeIndex(midIndex);
                
                // Get the cursor desc at the mid point
                lower = lo; // getCursorDescFrom uses this for boundry checks
                upper = up; // getCursorDescFrom uses this for boundry checks   
                var mid = getCursorDescFrom(midNodeIndex, midIndex - searchSpace.nodes[midNodeIndex].startIndex, 0);
                
                // Upper and Lower are next to each other
                if (!mid) continue;
    
                // Record the sample
                samples.push(mid);
    
                // Simulate recursion using a local stack
                stack.push([lo, lni, mid, midNodeIndex]); // left side
                stack.push([mid, midNodeIndex, up, uni]); // right side
                
            } // Next
            
            
        } // End inner discoverAllLines
        
        /**
         * Esnures that the given descriptor is valid.
         * @param {de.cursor.CursorDescriptor} cDesc A desciptor
         * 
         * @return If the given descriptor wasn't valid, it returns a neighbouring cursor placement, 
         *         otherwise the given descriptor is returned.
         */
        function validDescriptor(cDesc) {
            if (!cDesc) return null;
            
            // If the search ended on a line break, then check to make sure
            // that the line break has no cursor placements to the left or right of it...
            if (_nodeName(cDesc.domNode) == "br") {
                var prevDesc = getNextCursorMovement(cDesc,true);
                if (prevDesc && isOnSameLine(cDesc, prevDesc)) {
                    cDesc = prevDesc;
                    cDesc.isRightOf = true;
                } else {
                    var nextDesc = getNextCursorMovement(cDesc,false);
                    if (nextDesc && isOnSameLine(cDesc, nextDesc)) {
                        cDesc = nextDesc;
                        cDesc.isRightOf = false;
                    }
                }
            } else if (de.doc.isMNPlaceHolder(cDesc.domNode))
                cDesc.isRightOf = false;
            
            cDesc.placement = getPlacementFlags(cDesc.domNode);
            
            // Supply the position of the cursor in the actual document
            // rather than just within the window.
            var docScrollPos = _getDocumentScrollPos();
            
            cDesc.docLeft = docScrollPos.left + cDesc.x;
            cDesc.docTop = docScrollPos.top + cDesc.y;
            
            return cDesc;
        } // End inner validDescriptor
    
    } // End searchBestCursorPos


})();
