/*
 * file: Doc.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */
bootstrap.provides("Doc");


var 
    /**
     * @final
     * The protected node classname prefix
     * @type String
     */
    _PROTECTED_CLASS = "sw-protect",

    /**
     * @final
     * The editable section node classname prefix
     * @type String
     */
    _ES_CLASS_PREFIX = "editable",
    
    /**
     * @final
     * The classname use for packaged nodes
     * @type String
     */
    _PACKAGE_CLASS_NAME = "sw-packaged";
    

/**
 * Note: may want to check _doesNeedESPlaceholder first.
 * 
 * @param {Node} domNode A dom node in the document to test.
 * @return {Boolean} True if the given dom node is in need of a modifiable placeholder.
 * 
 * @see _doesNeedESPlaceholder
 */
function _doesNeedMNPlaceholder(domNode){
    if (_isPlaceholderCandidate(domNode))         
        return !_doesContainTanglableDescendant(domNode);
    return false;
}

/**
 * @param {Node} domNode A dom node in the document to test.
 * @return {Boolean} True if the given dom node is in need of a editable section placeholder.
 * 
 * @see _doesNeedMNPlaceholder
 */
function _doesNeedESPlaceholder(domNode){
    if (de.doc.isEditSection(domNode))
		return !_doesContainTanglableDescendant(domNode);
    return false;
}


/**
 * @param {Node} domNode A dom node to test
 * @return {Boolean} True iff the dom node contains a descendant for which the cursor can be placed by.
 */
function _doesContainTanglableDescendant(domNode) {

    var containsTangableNode = false;
    
    // Check if needs a placeholder
    _visitAllNodes(domNode, domNode, true, function(node){
        
        if (node == domNode || de.doc.isProtectedNode(node)) return;
    
        var pflags = de.cursor.getPlacementFlags(node);
        if (pflags == de.cursor.PlacementFlag.INSIDE) {
        
            if (_doesTextSupportNonWS(node) && _nodeLength(node) > 0) {
                
                if (_isAllWhiteSpace(node.nodeValue)) {
                
                    // Validate that the text node is tangable
                    var measureSpan = $createElement("span"),
                    	measureText = document.createTextNode(node.nodeValue);
						
                    measureSpan.appendChild(measureText);
                    node.nodeValue = "";
                    node.parentNode.appendChild(measureSpan);
                    containsTangableNode = measureSpan.offsetHeight != 0 && measureSpan.offsetWidth != 0;
                    node.parentNode.removeChild(measureSpan);
                    node.nodeValue = measureText.nodeValue;
                    
                    // TODO: Opera sometimes (randomly) incorrectly sets the offset width/height to zero
                    // on text nodes...
                    
                } else containsTangableNode = true;
            }
        } else if (pflags) containsTangableNode = true;
        
        return !containsTangableNode;
    });
    
    return containsTangableNode;
} // End function doesContainTanglableDescendant


/**
 * 
 * @param {Node} node A dom node to test
 * @return {Undefined, Boolean} True if node is a placeholder candidate. Undefined if it is not.
 */
var _isPlaceholderCandidate = function(){
	/*
     * A map containing elements the cursor can navigate into which do not contain a "Tangle node."
     * Excludes body
     */
     var placeholderCandidates = $createLookupMap("li,dd,dt,p,td,th,h1,h2,h3,h4,h5,h6,pre,div");
	 
    return function(node){
        return placeholderCandidates[_nodeName(node)] || node == docBody;
    }
}();
    


(function(){
    
    $enqueueInit("Doc", function() {
        
         // Make as subject
        _model(de.doc);
        
        // Preprocess the document: consolidate all existing/initial editable sections
        var es = de.doc.getAllEditSections();
        
        _recordOperations = false;
        for (var i in es) {
            _consolidateWSSeqs(es[i], true);
        }
        _recordOperations = true;
        
        // For dynamically added editable sections, consolidate them too.
        de.doc.addObserver({
            onSectionAdded : function(editSection) {
                // Only consolidate within the editable sections to avoid possibilities of consoldiating
                // surrounding editable DOM with Undo/Redo history.
                _recordOperations = false;
                _consolidateWSSeqs(editSection, false);
                _recordOperations = true;
            }
        });
        
    });

    var 
        propertySetMap = {},
		MN_PH_CLASS = "sw-mn-ph",
		ES_PH_CLASS = "sw-es-ph",
        ES_CLASS_TEST_REGEXP = new RegExp("^" + _ES_CLASS_PREFIX + ".*$"),
		ES_CLASS_MATCH_REGEXP = new RegExp("^" + _ES_CLASS_PREFIX + "-?(.+)$"),
        PROTECTED_NODE_TEST_REGEXP = new RegExp("^" + _PROTECTED_CLASS + "$"),
        PACKAGED_NODE_TEST_REGEXP = new RegExp("^" + _PACKAGE_CLASS_NAME + "$");
    
    // Create the doc namespace
    
    /**
     * @namespace 
     * Provides CSS like language for declaring and customizing editable sections on web pages.
     * 
     * <br><br>
     * Whenever a new editable section has been dynamically added to the document, a "onSectionAdded" event is fired,
     * where the argument is the added editable section.
     * 
     * <br><br>
     * Whenever a editable section has been dynamically removed from the document, a "onSectionRemoved" event is fired,
     * where the argument is the removed editable section.
     * 
     * @borrows de.mvc.AbstractSubject#addObserver as this.addObserver
     * 
     * @borrows de.mvc.AbstractSubject#removeObserver as this.removeObserver
     * 
     * @author Brook Novak
     */
    de.doc = {
    
        /**
         * @param {Node} node A dom node
         *
         * @return {Node}     The first ancestor element of node which is an editable section, inclusive of the given node itself.
         * 					  Null if the node / none of its ancestors are editable sections.
         */
        getEditSectionContainer: function(node){
            return _findAncestor(node, null, this.isEditSection, true);
        },
        
        /**
         * Determines whether a dom node is marked as edit section element.
         *
         * @param {Node} node  The dom node to test
         *
         * @return {Boolean}   True if node is a edit section element
         */
        isEditSection: function(node){
			if (node && node.nodeType == Node.ELEMENT_NODE)
				return _findClassName(node, ES_CLASS_TEST_REGEXP);
            return false;
        },
		
		/**
		 * @return {[Node]} An array of editable sections currently in the document.
		 */
		getAllEditSections : function() {
			
			var editSections = [];
			_visitAllNodes(docBody, docBody, true, function(domNode) {
				if (de.doc.isEditSection(domNode)) editSections.push(domNode);
			});
			return editSections;
			
		},
        
        /**
         * @param {Node} domNode A dom node to test
         * @return {Boolean}     True if the given dom node is a descendant of an editable section.
         */
        isNodeEditable : function(domNode) {
            var es = this.getEditSectionContainer(domNode);
            return es != null && es != domNode;
        },

        /**
         * @param {Node} domNode A dom node to test
         * @return {Node}        The protected nodes container is the node is proected. If the dom is a container then
         *                       it will be returned. Otherwise null will be returned.
         */
        getProtectedNodeContainer : function(domNode) {
            return _findAncestor(domNode, null, function(node) {
                return node.nodeType == Node.ELEMENT_NODE && _findClassName(node, PROTECTED_NODE_TEST_REGEXP);
            }, true);
        },
        
        /**
         * @param {Node} domNode A dom node to test
         * @return {Boolean}     True if the given dom node is, or is an descendant of, a protected node
         */
        isProtectedNode : function(domNode) {
            return this.getProtectedNodeContainer(domNode) != null;
        },
        
        /**
         * @param {Node} domNode  A dom node to test
         * 
         * @return {Node}         The package root node. Null if the node is not packaged.
         */
        getPackageContainer : function(domNode) {
            return _findAncestor(domNode, null, function(node) {
                return node.nodeType == Node.ELEMENT_NODE && _findClassName(node, PACKAGED_NODE_TEST_REGEXP);
            }, true);
        },

        /**
         * A "packaged" node is part of a tree of nodes which are not allowed to be edited, although
         * they are in an editable section.
         * 
         * @param {Node} domNode     A dom node to test
         * @return {Boolean}         True if the given dom node is part of a package.
         */
        isNodePackaged : function(domNode) {
            return this.getPackageContainer(domNode) != null;
        },
        
        /**
         * Declares or overrides a property set for editable sections. 
         * <br><br>
         * Attributes:
         * 
         * 		actionFilter: "[!][actionname1[,actionname2[,...]]]"
         *                     where actual name is case insensitive undoable action name.
         *                     Format action can be followed by sub-action encapsulated in brackets
         * <br><br>
         * If the property name exists, then the existing set will be overridden. 
         * You can override the default property set by using the name "defaultSet". 
         *
         * @example
         * de.doc.declarePropertySet("metadata", {
         * 		inputFilter: "[[a-z][A-Z][1-9]\\s\\n]*",
         * 		actionFilter: "!blockquote,changecontainer,format(link)"
         *                     Accept all actions EXCEPT for blockquote,changecontainer,formatting links
         *                     
         * 	});
         * 
         * @see TODO REFER TO SPEC
         *
         * @param {String} name
         * 					The name of the class being declared.
         *
         *
         * @param {Object} properties 	
         *              A set of attributes that is associated with the given name.
         *
         */
        declarePropertySet: function(name, properties) {
            properties = _clone(properties);
            
            // Build action filter
            if (typeof properties.actionFilter == "string") {
                
                var actionFilter = properties.actionFilter;
                
                // Is the action inclusive or exclusive?
                if (actionFilter.charAt(0) != '!') {
                    properties.afInclusive = true;
                    // Add implicit text editing actions
                    if (actionFilter) 
                        actionFilter += ",";
                    actionFilter += "inserthtml,inserttext,removedom,removetext";
                } else {
                    actionFilter = actionFilter.substr(1);
                    properties.afInclusive = false;
                }
                
                // Break filter into tokens
                var tokens = actionFilter.toLowerCase().split(',');
                
                // Build up the reg exp for quick filtering
                var reStr = "(";

                for (var i in tokens) {
                    
                    reStr += ((i=='0') ? "" : "|");
                    
                    var token = tokens[i];
             
                    var match = /^format\((.+)\)$/.exec(token);
                    
                    // Format actions can have sub-action filters
                    if (match) {
                    
                        var subTokens = match[1].split(',');
                        
                        for (var j in subTokens) {
                            reStr += ((j == '0') ? "" : "|") + "format" + subTokens[j];
                        }
                        
                    } else {
                    
                        // Add the action name to the reg exp set
                        reStr += token;
                    
                        if (token.indexOf("format") == 0) 
                            reStr += ".+"; // If not sub-format actions are defined then declare as all format actions
                    }
                
                } // End loop: parsing action tokens
                
                reStr += ")";
                
                properties.afRE = new RegExp("^" + reStr + "$");
            }
            
			// Set or override a property set
            propertySetMap[name] = properties;
            
            
        },
		
		/**
		 * Declares a batch of property sets in a single call
		 * @param {Object} sets An object containing key-value pairs, when the keys are property set names, 
		 * 					and values are objects containing properties.
		 */
		declarePropertySets: function(sets) {
			for (var tuple in sets) {
				this.declarePropertySet(tuple, sets[tuple]);
			}
		},

	    /**
	     * Retreives the editable property set for a given node. This considers property inheritance
	     *
	     * @param {Node} node A dom node
	     *
	     * @return {Object}	A set of read only dedit atrributes that the given node has inherited.
	     * 			Null if the node is not a (or descendant of a) editable section.
	     */
	    getEditProperties : function (node) {

            // Get the editable section container for this node
            if (!this.isEditSection(node))
                node = this.getEditSectionContainer(node);
            
            if (node) {

    			// Get the property set name for this node
        		var esClassName = _findClassName(node, ES_CLASS_TEST_REGEXP);
                
        		if (esClassName) {
        			var nameMatch = ES_CLASS_MATCH_REGEXP.exec(esClassName);
        		    return nameMatch ? propertySetMap[nameMatch[1]] || {} : {};
        		}

        		return {};
                
            }
            
            return null;
	    },
		
        /**
         * @return {Node} A modifiable node placeholder element.
         *
         * @see The DOM-based Web Editor Specification 1.0: Section 1.4
         */
        createMNPlaceholder: function(){
            var ph = $createElement("span");
            _setClassName(ph, MN_PH_CLASS);
            ph.innerHTML = "&nbsp;";
            return ph;
        },
        
        /**
         * Determines whether a dom node is (part of) a modifiable node placeholder.
         *
         * @param {Node} node A dom node to test
         * 
         * @param {Boolean} immediate 	True to only test if node is a placeholder.
         * 								False to also test if nodes parent is a placeholder.
         *
         * @return {Boolean} True if node is (part of) a modifiable node placeholder. False if it is not
         *
         * @see The DOM-based Web Editor Specification 1.0: Section 1.4
         */
        isMNPlaceHolder: function(node, immediate){
			switch(node.nodeType) {
				case Node.ELEMENT_NODE:
				return _getClassName(node) == MN_PH_CLASS;
				case Node.TEXT_NODE:
				return !immediate && node.parentNode && _getClassName(node.parentNode) == MN_PH_CLASS;
			}
			return false;
        },
		
        /**
         * @param {Node} editSection A editible section to create the editable section placeholder for
         * @return {Node} An editable section placeholder element.
         */
		createESPlaceholder : function(editSection) {
			var phHTML = this.getEditProperties(editSection).phMarkup || "&nbsp;";
            var ph = $createElement("span");
			_setClassName(ph, ES_PH_CLASS);
            ph.innerHTML = phHTML;
            return ph;
		},
		
        /**
         * Determines whether a dom node is (part of) an editable section placeholder.
         *
         * @param {Node} node A dom node to test
         * 
         * @param {Boolean} immediate 	True to only test if node is a placeholder.
         * 								False to also test if the nodes descendants is a placeholder.
         *
         * @return {Boolean} True if node is (part of) a editable section placeholder. False if it is not
         */
        isESPlaceHolder: function(node, immediate){

            while (node) {
                if (node.nodeType == Node.ELEMENT_NODE) {
                    var clsName = _getClassName(node);
                    if (clsName == ES_PH_CLASS) 
                        return true;
                }
				if (immediate) break;
                node = node.parentNode;
            }
            return false;
        },
        
        /**
         * Create and adds a new edit section to the document. Sets the classname for the given editable section.
         * The is required when adding new editable sections <em>after initialization</em> so dedit can track
         * its changes.
         * <br/>
         * Adds placeholders if they are needed.
         * 
         * @param {Node} esEle The editable section to register
         * @param {String} propertySetName The name of the property set to use. Can be null/empty for default set.
         * 
         * @see de.doc.removeEditSection For removing a editable section.
         */
        registerEditSection : function(esEle, propertySetName) {
			
			// Set class name as editable
            var clsName = _getClassName(esEle);
            _setClassName(esEle, (clsName ? clsName + " " : "") + _ES_CLASS_PREFIX + (propertySetName ? "-" + propertySetName : ""));
            
			// Add an editable section placeholder if it needs it
			if (!_doesContainTanglableDescendant(esEle)) 
				esEle.appendChild(this.createESPlaceholder(esEle));
			
			this.fireEvent("SectionAdded", esEle); // NB: Changes module listens
        },
        
        /**
         * Unregisters an editable section from the document.
         * Sets the classname for the given editable section
         * 
         * @param {Node} esEle The edit section to remove from the document.
         */
        unregisterEditSection: function(esEle) {
			
			// Strip classname of editable section prefix
            var clsName = _getClassName(esEle);
			if (clsName)
                _setClassName(esEle, clsName.replace(new RegExp("^|\s" + _ES_CLASS_PREFIX + "\S*$", "g"), ""))

            this.fireEvent("SectionRemoved", esEle); // NB: Changes module listens
        }
        
    }; // End doc namespace
    
})();
