/*
 * file: FormatEnvironment.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */

bootstrap.provides("FormatEnvironment");

/**
 * Stores all format environment variables. These are used by the undoable "Format" action.
 */
var _formatEnvironment = {};

/**
 * Sets a specific format variable in the format environment. Can use to create or override-existing 
 * formatting profiles which are used by the "Format" action.
 * 
 * @param {String} name           The name of the format type. For example "bold" (case insensitive)
 * @param {Function} evalFunc     The evaluation function. TODO: DOCUMENT
 * @param {Function} wrapperFunc  The wrapper function. Must return an inline element. TODO DOCUMENT
 */
function _setFormatEnvVar(name, evalFunc, wrapperFunc) {
    name = name.toLowerCase();
    _formatEnvironment[name + "Eval"] = evalFunc;
    _formatEnvironment[name + "Wrapper"] = wrapperFunc;
};

// Create the default set of formatting variables
(function() {
    
    /**
     * Creates a span with a CSS style
     * @param {String} css   The CSS style name
     * @param {String} val   The style value to set
     * @return {Node}        The created wrapper
     */
    function createCSSSpan(css,val) {
        var wrapper = $createElement("span");
        _setStyle(wrapper, css,val);
        return wrapper;
    }
    
    (function(envVars) {
        for (var i in envVars) {
            _setFormatEnvVar(i, envVars[i][0], envVars[i][1]);
        }
        
    })({
        
        bold: [
            /* Evaluation function */
            function(ele) {
                
                var matches = [];
                
                if (_nodeName(ele) == "b" || _nodeName(ele) == "strong")
                    matches.push({type: 1});
                    
                var fw = ele.style.fontWeight;
                if (fw) {
                    var isBold = fw == "bold";
                    if (!isBold) {
                        fw = parseInt(fw);
                        isBold = (!isNaN(fw) && fw >= 700)
                    }
                    if (isBold)
                        matches.push({type: 3, match:"fontWeight"});
                }
                
                return matches.length > 0 ? {
                    strip: matches, 
                    inline: $createElement("strong"), // Extracted inline equivalent
                    value:true
                    } : null;
               
            }, 
            
            /* Wrapper */
            function() {return $createElement("strong");}
        ],
        
        italics: [
            /* Evaluation function */
            function(ele) {
                
                var matches = [];
                
                if (_nodeName(ele) == "i" || _nodeName(ele) == "em")
                    matches.push({type: 1});
    
                if (ele.style.fontStyle == "italic")
                    matches.push({type: 3, match:"fontStyle"});
                
                return matches.length > 0 ? {
                    strip: matches, 
                    inline: $createElement("em"), // Extracted inline equivalent
                    value:true
                    } : null;
            }, 
            
            /* Wrapper */
            function() {return $createElement("em");}
     
        ],
        
        underline: [
            /* Evaluation function */
            function(ele) {
                
                var matches = [];
                
                if (_nodeName(ele) == "u")
                    matches.push({type: 1});
    
                if (ele.style.textDecoration == "underline")
                    matches.push({type: 3, match:"textDecoration"});
                    
                return matches.length > 0 ? {
                    strip: matches, 
                    inline: createCSSSpan("textDecoration", "underline"), // Extracted inline equivalent
                    value:true
                    } : null;
            }, 
            
            /* Wrapper */
            function() {
                return createCSSSpan("textDecoration", "underline");
            }
     
        ],
        
        strike: [
            /* Evaluation function */
            function(ele) {
                
                var matches = [];
                
                if (_nodeName(ele) == "strike")
                    matches.push({type: 1});
    
                if (ele.style.textDecoration == "line-through")
                    matches.push({type: 3, match:"textDecoration"});
                    
                return matches.length > 0 ? {
                    strip: matches, 
                    inline: createCSSSpan("textDecoration", "line-through"), // Extracted inline equivalent
                    value:true
                    } : null;
            }, 
            
            /* Wrapper */
            function() {
                return createCSSSpan("textDecoration", "line-through");
            }
            
            
        ],
        
        color: [
        
            /* Evaluation function */
            function(ele) {
                
                if (ele.style.color && ele.style.color.length > 0)
                    return {
                        strip:[{type: 3, match:"color"}],
                        inline: createCSSSpan("color", ele.style.color),
                        value : ele.style.color
                    };
                
            }, 
            
            /* Wrapper */
            function(value) {
                return createCSSSpan("color", value);
            }

        ],
        
        backcolor: [
        
            /* Evaluation function */
            function(ele) {
                
                if (ele.style.backgroundColor && ele.style.backgroundColor.length > 0)
                    return {
                        strip:[{type: 3, match:"backgroundColor"}],
                        inline: createCSSSpan("backgroundColor", ele.style.backgroundColor),
                        value : ele.style.backgroundColor
                    };
                
            }, 
            
            /* Wrapper */
            function(value) {
                return createCSSSpan("backgroundColor", value);
            }

        ],
        
        fontsize: [
            /* Evaluation function */
            function(ele) {
                
                if (_nodeName(ele) == "small" || _nodeName(ele) == "big") {
                    var val = _nodeName(ele) == "small" ? "smaller" : "larger";
                    return {
                        strip:[{
                            type: 1
                        }],
                        inline: createCSSSpan("fontSize", val),
                        value : val
                    };
                 }
                    
                if (ele.style.fontSize && ele.style.fontSize.length > 0)
                    return {
                        strip:[{type: 3, match:"fontSize"}],
                        inline: createCSSSpan("fontSize", ele.style.fontSize),
                        value: ele.style.fontSize
                    };
            }, 
            
            /* Wrapper */
            function(value) {
                return createCSSSpan("fontSize", value);
            }
        ],
        
        fontfamily: [
            /* Evaluation function */
            function(ele) {
                if (ele.style.fontFamily && ele.style.fontFamily.length > 0)
                    return {
                        strip:[{type: 3, match:"fontFamily"}],
                        inline: createCSSSpan("fontFamily", ele.style.fontFamily),
                        value: ele.style.fontFamily
                    };
            }, 
            
            /* Wrapper */
            function(value) {
                return createCSSSpan("fontFamily", value);
            }
        ],
        
        link: [
        
            /* Evaluation function */
            function(ele) {
                if (_nodeName(ele) == "a")
                    return {
                        strip:[{type: 1}],
                        inline: ele.cloneNode(false),
                        value: {
                            url: ele.href,
                            title: ele.title
                        }
                    };
            }, 
            
            /* Wrapper */
            function(value) {
                var wrapper = $createElement("a");
                wrapper.href = value.url;
                wrapper.title = value.title;
                return wrapper;
            }
        ]
    });

})();

