/*
 * file: Fragment.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */

bootstrap.provides("Fragment");

/**
 * @class
 * 
 * DOMFragment's are used to describe a range of DOM nodes. They are similar
 * W3C DOM Ranges, except they have a more rich/expressive range and contain information 
 * for reversing operations safely.
 *  
 * @constructor
 * @private
 * 
 * Use de.DOMFragment.buildFragment
 * 
 * @param {Node} domNode The dom node to wrap
 * @param {Number} posInParent The position of the domNode in is parent child list.
 * 
 */
var _DOMFragment = function() {
    

    var cls = function(domNode, posInParent) {
		
		/**
		 * The wrapped dom node
		 * @type Node
		 */
		this.node = domNode;
    	
    	/**
    	 * A read only member: The position of the domNode in is parent child list.
    	 * @type Number
    	 */
    	this.pos = posInParent;
    	
    	/**
    	 * A read only member: The de.dom.DOMFragment's children. Never null.
    	 * @type [de.dom.DOMFragment]
    	 */
    	this.children = [];
        
        /**
         * True indicates that this node has descendants (including self) that are outside of the
         * fragment range. Flase indicates that the node is completely within the fragments range.
         * 
         * @type Boolean
         */
        this.isShared = false;
        
        /**
         * The dom fragments parent. Null for root.
         * 
         * @type de.dom.DOMFragment
         */
        this.parent = null;
    }
    
    cls.prototype = {
    
        /**
         * Applies a function to nodes in order starting from this node and all its children.
         *
         * @param {Function} func A function applied to the nodes. 1 argument is given: a visited node.
         * 						The traversing will be stopped by returning false.
         */
        visit: function(func){

            if (func(this) === false)             
                return false;
            
            for (var i in this.children) {
                if (!this.children[i].visit(func))                 
                    return false;
            }
            
            return true;
        },
        
        
        
        /**
         *
         * Removes the fragment from the document.
         *
         */
        disconnect: function(){
            this.visit(function(currentFrag) {
               if (!currentFrag.isShared) 
                   removeFragment(currentFrag, false); 
            });    
        },


        /**
         * An extension of de.dom.DOMFragment#disconnect. This routine furthermore collapses nodes after the
         * disconnection: where nodes the have been left in the document -- within the fragment range --
         * are removed/copied/moved in a way that a typical word-proccessor would behave.
         * <br><br>
         * The fragment must not be disconnected prior to this call.
         * 
         * @return {Node} The first migrated node. Null if nothing was migrated.
         */
        collapse : function() {
            
            // Get the first common ancestor between the start and end fragments... this may not be the root.
            var firstCommonAncestor = _getCommonAncestor(this.getStartFragment().node, this.getEndFragment().node, false),
                fcaFrag = this,
                stopMigratingSiblings = false, // used for inner function
            	fragRoot = this, // used for inner function
            	hasCollapsibleBoundAncestor = false; // used for inner function
                
            while(fcaFrag.node != firstCommonAncestor) {
                fcaFrag = fcaFrag.children[0];
            }

            // Remove the fragments inclusive range from the document
            fragRoot.disconnect();
 
            // Insert any placeholders that needs placing
            insertPlaceholders();

            // Remove all migratable nodes and build migration trees
            var migrations = migratePath(fcaFrag, true, nextMigrationPoint(fragRoot.getStartFragment())),
                i, j;
            
            // Add migration trees into their migration points
            for (i in migrations) {
                
                if (migrations[i].migrantRoots.length > 0) {
                    
                    // Discover the insertion index in the migration point
                    var mPointNode = migrations[i].migrationPoint.node,
                        insertIndex = migrations[i].migrationPoint.children[0].pos, // Start-bound's index
                        containsNonPHTangable = false,
                        seenPlaceholder = 0;
                        
                    // If the start-bound child is still in the document, then increase the index
                    if (isInDocument(migrations[i].migrationPoint.children[0].node)) 
						insertIndex++;

                    // Insert the migrants
                    for (j=0; j<migrations[i].migrantRoots.length; j++) {
						_execOp(_Operation.INSERT_NODE, migrations[i].migrantRoots[j], mPointNode, insertIndex + j);
                    }
                    
                    // Check to see if migration point has any redundant placeholders, which may have been uneccessarely added, 
                    // or just moved, via the collapse operation.
                    
                    // First check if there are any tangable descendants of the migration point which are not
                    // immediate children that are placeholders...
                    if (_isPlaceholderCandidate(mPointNode)) {
                        _visitAllNodes(mPointNode, mPointNode, true, function(node){
                            if (node == mPointNode) return;
                            if (!(de.doc.isMNPlaceHolder(node, false) && 
                                (node.nodeType == Node.TEXT_NODE ? node.parentNode.parentNode == mPointNode : node.parentNode == mPointNode)) && 
                                de.cursor.getPlacementFlags(node)) {
                                containsNonPHTangable = true;
                                return false;
                            } 
                            
                        });
                    }
                    
                    // Scan through the migration points immediate nodes
                    for (var k = 0; k < mPointNode.childNodes.length; k++) {
                        
                        var domNode = mPointNode.childNodes[k];
                        
                        if (de.doc.isMNPlaceHolder(domNode, true) || de.doc.isESPlaceHolder(domNode, true)) {
                            
                            // If we have already seen a placeholder in this migration point, or the migration point is not
                            // a placeholder candidate, or the migration point contains tangable nodes that are not immediate
                            // children who are placeholders... then this placeholder is redundant
                            if (seenPlaceholder || !_isPlaceholderCandidate(mPointNode) || containsNonPHTangable) 
								_execOp(_Operation.REMOVE_NODE, domNode); // Remove it from the document

                            seenPlaceholder = 1;
                        }
                    }
                }
            }

            // Return the first migrated node
            return migrations.length > 0 && migrations[0].migrantRoots.length > 0 ? migrations[0].migrantRoots[0] : null;
            
           /**
            * Removes/clones nodes from the document and creates migration trees -- paired with their
            * migration points, for which they should be appended to.
            * 
            * @param {de.dom.DOMFragment} pathRoot     The top-most fragment of the path to travel up to.
            * @param {Object} isEndBoundPath           True iff the path is the end-boundry path
            * @param {de.dom.DOMFragment} curMigPoint  The current migration point
            * @return {[Object]}                       The migrations to make
            */
           function migratePath(pathRoot, isEndBoundPath, curMigPoint) {
               
                var curMigration = {
                    /* 
                     * A migration point is a dom node in the start-bound path where nodes in the
                     * end bound path can move(migrate) to.
                     */
                     migrationPoint : curMigPoint,
                     
                     /*
                      * Migration roots are disconnected dom-trees which should be connected with the
                      * migration point in this instance by adding as children ... which is done outside
                      * of this function scope.
                      */
                     migrantRoots : []
                }
                
                var migrations = [curMigration];
        
                // Follow a path starting from the end node and moving up to the path's root
                for (
                    var fragment = isEndBoundPath ? pathRoot.getEndFragment() : pathRoot.getStartFragment(); 
                    fragment != (isEndBoundPath ? pathRoot : pathRoot.parent); 
                    fragment = fragment.parent) {
    
                    var domNode, nextNode;

                    // Check to see if this fragment has an ancestor who is collapsible and is on the end-bound path.
                    if (isEndBoundPath) {
                        hasCollapsibleBoundAncestor = false;
                        for (var f = fragment.parent; f != pathRoot; f = f.parent) {
                            if (f.isShared && isCollapsible(f.node)) {
                                hasCollapsibleBoundAncestor = true;
                                break;
                            }
                        }
                    }
                    
                    if (!isEndBoundPath || fragment.isShared) {
                    
                        domNode = fragment.node;
                        nextNode = domNode.nextSibling;

                        // Should this node be removed? I.E: Is the node collapsable?
                        if (isCollapsible(domNode)) {
        
                            stopMigratingSiblings = true;
                            
							// All descendants of a collapsiable node on the bound path should be migrated
							debug.assert(!isEndBoundPath || (isEndBoundPath && !domNode.firstChild));
							
                            // Remove this collapsible node if it is on the end-bound path and is shared
                            if (isEndBoundPath)
                                _execOp(_Operation.REMOVE_NODE, fragment.node);
							
                        }
        
                        // Check to see if this node should be migrated, or should stay
                        if (!isEndBoundPath || canDirectlyMigrate(domNode)) {

                            // Check if need to remove this migrant or duplicate it
                            if (domNode.firstChild) 
                                domNode = domNode.cloneNode(false);
                            else 
								_execOp(_Operation.REMOVE_NODE, domNode);
                            
                            // Is the current migration point valid?
                            while (curMigration.migrationPoint != fcaFrag 
                                && !_isValidRelationship(domNode, curMigration.migrationPoint.node)
                                && _nodeName(domNode) != "div") {
                                // One exception: don't allow migration of div element: the generic container
                            
                                // Search for next valid migration point... 
                                curMigration.migrationPoint = nextMigrationPoint(curMigration.migrationPoint);
                            }

                            // Add this node to all migrations
                            for (var i = 0; i < migrations.length; i++) {
                                
                                // Link up dom node to its children
                                for (var j in migrations[i].migrantRoots) {
									_execOp(_Operation.INSERT_NODE, migrations[i].migrantRoots[j], domNode);
                                }
                                
                                // Set the new roots for this tree-level
                                migrations[i].migrantRoots = [domNode];
        
                                // Duplicate node for next migration point
                                if (i < (migrations.length - 1)) 
                                    domNode = domNode.cloneNode(false);
									
                            }
                            
                        } else if (!isCollapsible(domNode)) { // Fragment in boundry path that cannot be migrated
                        
                            stopMigratingSiblings = true; 
                            if (!domNode.firstChild) 
                                // The un-migratible node has been left childless, remove it from the document
                                _execOp(_Operation.REMOVE_NODE, fragment.node);
                        }
                        
                    } else {  // Is on end bound path and fragment not shared and has been disconnected. i.e. completely removed from document
                        
						// Determine next node...
						
                        // If the parent has been disconnected, then there is no use searching through the next siblings
                        if (!fragment.parent.isShared) 
							continue;
                        
                        // See if this fragments parent is part of the starting bound
                        var startBound = fragRoot;
                        while (startBound.children.length > 0 && startBound != fragment.parent) {
                            startBound = startBound.children[0];
                        }

                        if (startBound == fragment.parent) {
                            debug.assert(startBound.children.length > 0);
                            
                            // If the start bound path shares the same parent for the current fragment 
                            // (which is on the end path), then the next sibling is not neccesarily the first
                            // remaining child in this fragment's parent.
                            nextNode = null;
                            if (!startBound.insertedPH) { // Watch out for these, since they are added on the fly, not in fragment structure
                                
                                if (isInDocument(startBound.children[0].node)) {
                                    // The first child of the start bound is in the document
                                    nextNode = startBound.node.childNodes.length > (startBound.children[0].pos + 1) ?
                                        startBound.node.childNodes[startBound.children[0].pos + 1] : null;
                                } else {
                                    // The first child of the start bound has been removed
                                    nextNode = startBound.node.childNodes.length > startBound.children[0].pos ?
                                        startBound.node.childNodes[startBound.children[0].pos] : null;
                                }
                                
                            }

						 // Get the first remainding child in this disconnected fragment's parent
                        } else nextNode = fragment.parent.node.firstChild; // May be null/not exist

                    }
                    
                    // If the parent cannot be migrated from the end-bounds path, then don't migrate it's children to 
                    // the left of the end-bounds.
                    if (isEndBoundPath && !canDirectlyMigrate(fragment.parent.node) && !isCollapsible(fragment.parent.node)) 
                        stopMigratingSiblings = true;
                        
                    // If this fragment is for the starting iteration of a new migrant fragment, then avoid recursing into
                    // it's sibling dom nodes that are not in a fragment yet... this will be handled afterwards
                    else if (!isEndBoundPath && fragment == pathRoot) 
						continue;
                    
                    // Get this fragments child-index in it's parent
                    var nextFragIndex = fragment.getIndexInParent();
 
                    // Set/reset the migration index for this fragments parent. Only applicable on the bound path
                    if (isEndBoundPath) fragment.parent.migrantIndex = 0;
                    
                    while (nextNode && (hasCollapsibleBoundAncestor || !stopMigratingSiblings)) { // For each adjacent path

                        domNode = nextNode; // For bound path, valid for non-bound path
                        nextNode = domNode.nextSibling; // For bound path, valid for non-bound path
                        nextFragIndex++; // For non-bound path
                        
                        var adjacentMigrations;
                        
                        if (isEndBoundPath) {
                            
                            var migrantFragRoot;
                            
                            // If this operation is being repeated, then get the migration tree created from the first time
                            // this operation was performed
                            if (fragment.parent.migrants && fragment.parent.migrants.length > fragment.parent.migrantIndex) {
                                
                                migrantFragRoot = fragment.parent.migrants[fragment.parent.migrantIndex];

                            } else { // First time this operation has been performed
                            
                                // Build a fragment to capture the structure of the current document's state
                                migrantFragRoot = _buildFragment(
                                    domNode.parentNode, 
                                    domNode, 
                                    0, 
                                    domNode, 
                                    _nodeLength(domNode, 1));
          
                                // Link it to the parent fragment
                                if (!fragment.parent.migrants) fragment.parent.migrants = [];
                                fragment.parent.migrants.push(migrantFragRoot);
                            }
                            
                            fragment.parent.migrantIndex++;
         
                            adjacentMigrations = migratePath(migrantFragRoot.children[0], false, migrations[migrations.length-1].migrationPoint); // Recurse
                            
                        } else { // Fragment already created...
                        
                            adjacentMigrations = migratePath(
                                fragment.parent.children[nextFragIndex], 
                                false,
                                 migrations[migrations.length-1].migrationPoint); // Recurse
                        }

                        // Merge the adjacent migrations into the current migrations
                        for (var i in adjacentMigrations) {
                            if (adjacentMigrations[i].migrantRoots.length == 0) continue;
                            
                            // Search for a matching migration point
                            var wasMerged = false;
                            for (var j in migrations) {
      
                                if (migrations[j].migrationPoint == adjacentMigrations[i].migrationPoint) {
                                    // Append the adjacent migrant roots for this migration point (at the current tree-level)
                                    migrations[j].migrantRoots = migrations[j].migrantRoots.concat(adjacentMigrations[i].migrantRoots);
                                    wasMerged = true;
                                    break;
                                }
                            }
                            
                            // Must be a new migration point... add to the current set of migration points
                            if (!wasMerged) migrations.push(adjacentMigrations[i]);
      
                        }
                        
                        // Keep the current migration updated
                        curMigration = migrations[migrations.length-1]; // TODO: NEEDED?
                        
                    } // End loop: recursing over adjacent paths
            
                } // End loop: travelling up path to root
               
               return migrations;
           } // End inner migratePath
    
           /**
            * Looks up the start-bound path from the given migration point.
             * @param {_Fragment} currentPoint Must not be the first Common Ancestor - must be on the start bound path.
             * @return {_Fragment} The next migration point
             */
            function nextMigrationPoint(currentPoint) {
                
                debug.assert(currentPoint != fcaFrag);
                
                do {
                    
                    currentPoint = currentPoint.parent;
                    
                } while (currentPoint != fcaFrag && !(
                 
                       // currentPoint.isShared && // NOTE: Cannot used isShared to determine if exists in document... since cna be re-inserted
						isInDocument(currentPoint.node) && /* i.e: shared nodes or re-inserted nodes due to placeholders */
						isCollapsible(currentPoint.node)));

                return currentPoint;
            } // End inner nextMigrationPoint
                   
            /**
             * @param {Node} domNode
             * @return {Boolean} True if domNode is classed as "collapsible"
             */
            function isCollapsible(domNode) {
                return _isGenericBlockLevel(domNode) || _nodeName(domNode) == "li";
            } // End inner isCollapsible
            
            /**
             * @param {Node} domNode
             * @return {Boolean} True if domNode can directly be migrated from the boundry path into a migration point
             */
            function canDirectlyMigrate(domNode) {
                return domNode.nodeType == Node.TEXT_NODE || _isInlineLevel(domNode);
            } // End inner canDirectlyMigrate
            
            function isInDocument(domNode) {
                return _isAncestor(docBody, domNode);
            }

            /**
             * If the start-bound path contains any placeholder candidates either left without tangable nodes 
             * or are fully disconnected, then placeholders are inserted / fragments are reconnected
             */
            function insertPlaceholders() {
            
                for (var fragment = fragRoot.getStartFragment(); 
                    fragment != null; 
                    fragment = fragment.parent ? fragment.parent : null) {
                        
                    var domRef = fragment.node;
                        
                    if (_isPlaceholderCandidate(domRef) || de.doc.isEditSection(domRef)) {
                        
                        // Is this still in the document?
                        if (fragment.isShared) {
							
							var phType = 0
			                if (_doesNeedESPlaceholder(domRef))
								phType = 1;
							else if (_doesNeedMNPlaceholder(domRef))
								phType = 2;
 
                            if (phType) {
                                
                                // The disconnection has left a placeholder candidate in the document without a tangable node.

                                // Create a new placeholder and add it
                                var ph = phType == 1 ? 
									de.doc.createESPlaceholder(domRef) : 
									de.doc.createMNPlaceholder();
									
								_execOp(_Operation.INSERT_NODE, ph, domRef);
								
								// Mark that this fragment inserted a placeholder
								fragment.insertedPH = 1;
                            }  
                            
                        } else { // disconnected

							// Get rid of any children that were part of the diconnection
                            while (domRef.firstChild) {
								_execOp(_Operation.REMOVE_NODE, domRef.firstChild);
                            }
                        
                            // Create a new placeholder and add it
							_execOp(_Operation.INSERT_NODE, de.doc.createMNPlaceholder(), domRef);
							
							// Mark that this fragment inserted a placeholder
							fragment.insertedPH = 1;
							
                            // Link this and all it's disconnect ancestors back into the document...
                            while (!fragment.isShared) {

                                if (fragment.parent.isShared) {
                                    
                                    // Reconnect this with it's parent
									_execOp(_Operation.INSERT_NODE, fragment.node, fragment.parent.node, fragment.pos);

                                } else { 
                                
                                    // Make sure the parent has all its children removed
                                    var parentDomNode = fragment.parent.node;
                                    while(parentDomNode.firstChild) {
										_execOp(_Operation.REMOVE_NODE, parentDomNode.firstChild);
                                    }
                                    
                                    // Reconnect this with it's parent
									_execOp(_Operation.INSERT_NODE, fragment.node, parentDomNode);
									
									
                                    
                                }
                                
                                if(!fragment.parent) break;
                                fragment = fragment.parent;
                            }

                        }
                        
                        break; // No need to search ancestors for inserting placeholders as this point
                        
                    }
                }
            }

        },
        

        /**
         * @return {de.dom.DOMFragment} The start fragment of the fragment's range.
         */
        getStartFragment : function() {
            var startFrag = this;
            while (startFrag.children.length > 0) {
                startFrag = startFrag.children[0];
            }
            return startFrag;
        },
        
        /**
         * @return {de.dom.DOMFragment} The end fragment of the fragment's range.
         */
        getEndFragment : function() {
            var endFrag = this;
            while (endFrag.children.length > 0) {
                endFrag = endFrag.children[endFrag.children.length - 1];
            }
            return endFrag;
        },
        
        /**
         * @return {Integer} The zero-based index of this fragment in it's parent. Null if this is a root fragment.
         */
        getIndexInParent : function() {
            
            if (!this.parent) return null;
            
            var index = 0;
            while (this != this.parent.children[index]) {
                index++;
            }
            return index;
        },
        
        /**
         * @return {Boolean} True if the start node was a split text node, so that it has a 
         *                   preceeding text node which was the orginal node, false otherwise.
         */
        wasStartSplit : function() {
            return this.getStartFragment().preSplitNode ? true : false;
        },
        /**
         * @return {Boolean} True if the end node was a split text node, so that it has a 
         *                   proceeding text node which was the new split node, false otherwise.
         */
        wasEndSplit : function() {
            return this.getEndFragment().postSplitNode ? true : false;
        },
        
        /**
         * @return {Node} The previous text node which will split at the start fragment.
         *                 Null if this framgent didn't split the start point
         */
        getPreSplitNode : function() {
            var startFrag = this.getStartFragment();
            return startFrag.preSplitNode ? startFrag.preSplitNode : null;
        },
        
        /**
         * @return {Node} The following text node which will split at the end fragment.
         *                 Null if this framgent didn't split the end point
         */
        getPostSplitNode : function() {
            var endFrag = this.getEndFragment();
            return endFrag.postSplitNode ? endFrag.postSplitNode : null;
        },
        
        /**
         * Gets the adjusted node/index of the given tuple to point to a valid position in the DOM
         * which they may have been left pointing to invalid indexes due to the construction of this fragment.
         * 
         * @param {Object} node
         * 
         * @param {Object} index
         * 
         */
        getAdjustedNodeIndex : function(node, index) {
                
            // Check if node belongs in the formatting fragments' split text nodes, and see if it's index is out of bounds
            if (node.nodeType == Node.TEXT_NODE && index >= _nodeLength(node)) {
            
                var startFrag = this.getStartFragment(), 
                    endFrag = this.getEndFragment();
                        
                // Was the start node split? And did the node previously point to this split node?
                if (startFrag.wasStartSplit()) {
                    var preSplitNode = startFrag.getPreSplitNode();
                    if (node == preSplitNode) {
                        index -= _nodeLength(preSplitNode);
                        node = startFrag.node;
                    }
                }
                
                // Was the end node split? And did the node previously point to the split node?
                if (endFrag.wasEndSplit() && node == endFrag.node && index >= _nodeLength(node)) {
                    index -= _nodeLength(endFrag.node);
                    node = endFrag.getPostSplitNode();
                }
            }
            
            return {node:node,index:index};
            
        },
        
        /**
         * Gets the original node/index values before this fragment was built.
         * 
         * @param {Object} node
         * @param {Object} index
         */
        getOriginalNodeIndex : function(node, index) {
     
            if (node.nodeType == Node.TEXT_NODE) {
            
                var startFrag = this.getStartFragment(),
                    endFrag = this.getEndFragment();
                
                // Was the end node split? And does the node point to the added split node?
                if (node == endFrag.getPostSplitNode()) {
                    index += _nodeLength(endFrag.node);
                    node = endFrag.node;
                }
                
                // Was the start node split? And does the node point to the added split node?
                if (startFrag.wasStartSplit() && node == startFrag.node) {
                    var preSplitNode = startFrag.getPreSplitNode();
                    index += _nodeLength(preSplitNode);
                    node = preSplitNode;
                }
            }                    
            
            return {node:node,index:index};
        }

    };
	
	// @DEBUG ON
	// Add tostring method for debugging
	cls.prototype.toString = function() {
		return _nodeName(this.node) + (this.isShared ? "[SHARED]" : "[NOT-SHARED]");
	}
	// @DEBUG OFF

    /**
     * Removes the fragments dom node from the document.
     * 
     * @param {de.dom.DOMFragment} fragment
     * @param {Boolean} forceRemoveDoc Set to true to always remove the fragments node from the document.
     *                                 Set to false to only remove if the fragment has a shared parent.
     */
    function removeFragment(fragment, forceRemoveDoc) {
        if (forceRemoveDoc || (fragment.parent && fragment.parent.isShared))
			_execOp(_Operation.REMOVE_NODE, fragment.node);
    }
    
    return cls;
    
}();

/**
 * Builds a fragment. 
 * <br><br>
 * If startNode / endNode are the same, then startIndex / endIndex must be different.
 * If startNode or endNodes are placeholders, they are extended to include the whole placeholder in the range.
 * <br>
 * The start node/index tuple must occur before the end node/index tuple when traversing inorder.
 * <br>
 * The range will always been extended to the deepest descendants.
 * 
 * @param {Node} commonAncestor (Optional) A common ancestor of the start and end nodes, can be as high as possible.
 *                              If null then the closest common ancestor will be used.
 *
 * @param {Node} startNode      The starting dom node of the fragments range.
 *
 * @param {Number} startIndex 	The inclusive start index in the start node.
 * 								Ranges from 0 to the text length for text nodes.
 *                              Where 0 indicates that the range begins at the first char, and text length
 *                              indicates that the range begins directly after the text node, but not including it.
 *                              <br>
 *                              Ranges from 0 to 1 for elements.
 *                              Where 0 indicates that the range includes the element and its decendants,
 *                              and 1 indicates that the range excludes the element and its decendants.
 *                              
 *
 * @param {Node} endNode        The ending dom node of the fragments range.
 *
 * @param {Number} endIndex 	The inclusive end index in the end node.
 * 								Ranges from 0 to the text length for text nodes.
 *                              Where 0 indicates that the range ends just before the text node, but not including it, 
 *                              and text length indicates that the range ends at the last charactor in the text run.
 *                              <br>
 *                              Ranges from 0 to 1 for elements.
 *                              Where 0 indicates that the range excludes the element and its decendants,
 *                              and 1 indicates that the range includes the element and its decendants.
 *
 * @return {de.dom.DOMFragment} The fragments root FragmentNode, which will always be the commonAncestor. Never null.
 */
function _buildFragment(commonAncestor, startNode, startIndex, endNode, endIndex) {
    
    debug.assert(startNode != endNode || startIndex < endIndex, "Invalid range");
    
	debug.assert(
		!(startNode.nodeType == Node.TEXT_NODE && (startIndex < 0 || startIndex > _nodeLength(startNode))),
		"Start index out of range"
	);
    
	debug.assert(
		!(startNode.nodeType == Node.ELEMENT_NODE && (startIndex < 0 || startIndex > 1)),
		"Start index out of range"
	);
	
	debug.assert(
		!(endNode.nodeType == Node.TEXT_NODE && (endIndex < 0 || endIndex > _nodeLength(endNode))),
		"End index out of range"
	);

	debug.assert(
		!(endNode.nodeType == Node.ELEMENT_NODE && (endIndex < 0 || endIndex > 1)),
		"End index out of range"
	);
    
	debug.assert(
		!(endNode == startNode && endIndex == startIndex),
		"Invalid range"
	);
    
    // Set common ancestor if not given
    if (!commonAncestor)
        commonAncestor = _getCommonAncestor(startNode, endNode);
    
    // Make sure the range extends to the deepest descendants
    var adjustBoundry = false;
    while (startNode.firstChild) {
        startNode = startIndex == 0 ? startNode.firstChild : startNode.lastChild;
        adjustBoundry = true;
    }
    if (adjustBoundry && startIndex > 0) 
        startIndex = _nodeLength(startNode, 1);
   
    adjustBoundry = false;
    while (endNode.firstChild) {
        endNode = endIndex == 0 ? endNode.firstChild : endNode.lastChild;
        adjustBoundry = true;
    }
    if (adjustBoundry && endIndex > 0) 
        endIndex = _nodeLength(endNode, 1);
    
    // Make sure placeholders are fully included within the inclusive range
    if (de.doc.isMNPlaceHolder(startNode) || de.doc.isESPlaceHolder(startNode)) 
		startIndex = 0;
		
 	if (de.doc.isMNPlaceHolder(endNode) || de.doc.isESPlaceHolder(endNode)) 
		endIndex = _nodeLength(endNode, 1);
        
    // @DEBUG ON
    // Check that end occurs after the start
    var foundEnd = false;
    _visitAllNodes(commonAncestor, startNode, true, function(domNode) {
        foundEnd = domNode == endNode;
        return !foundEnd;
    });
    debug.assert(foundEnd, "Invalid range: end node not found after start node");
    // @DEBUG OFF
    
    var affectedNodes = [];
    
    // Get all the nodes between the start and end nodes.
    _visitAllNodes(commonAncestor, startNode, true, function(node){
        affectedNodes.push(node);
        return node != endNode;
    });
    
    // Add the missing nodes to the start of the array
    var ancestors = _getAncestors(startNode, commonAncestor, false, true);
    ancestors.reverse();
    affectedNodes = ancestors.concat(affectedNodes);
    
    // to traverse the nodes inorder from the common ancestor through
    // the start and end nodes range, just iterate forwards on the array:
    var fragRoot;
    
    for (var i in affectedNodes) {
    
        // Build up a fragment by traversing through the affected nodes in order
        var node = affectedNodes[i], preSplitNode = 0, postSplitNode = 0;
        
        // Perform split operations on boundry text nodes
        if ((node == startNode || node == endNode) && node.nodeType == Node.TEXT_NODE) {

            // Because the first/last nodes are text nodes, we may have to split them up.
            if (node == startNode && startIndex > 0 && startIndex < _nodeLength(node)) {
                
                preSplitNode = node;
                
                // Leave the start node's stable references as is, since the remainding text
                // node within the range is a newly created text node...
                node = _execOp(_Operation.SPLIT_TEXT_NODE, node, startIndex);

                // Update endnode if range is all within same node
                if (endNode == startNode) {
                    endNode = node;
                    endIndex -= _nodeLength(startNode);
                }
                
                startNode = node;
                // No need to worry about start index
            }
            
            if (node == endNode && endIndex > 0 && endIndex < _nodeLength(node)) {
				postSplitNode = _execOp(_Operation.SPLIT_TEXT_NODE, node, endIndex);
			}
            
        }

		// Create a new fragment node
        var fragment = new _DOMFragment(node, _indexInParent(node));
        
		// Has the root been set?
        if (!fragRoot) 
            fragRoot = fragment;
        else { // Set up the parent relationship...
                
            fragment.parent = null;
            fragRoot.visit(function(frag) {
                if (frag.node == fragment.node.parentNode)
                    fragment.parent = frag;
                return fragment.parent == null;
            });
            fragment.parent.children.push(fragment);
        }
        
        // If the node is a boundry node and is excluded, then set a flag to
        // note that such nodes are to be outside of the fragment range.
        fragment.isShared = !(preSplitNode || postSplitNode) && (
            (node == startNode && startIndex == _nodeLength(node, 1))
             ||
            (node == endNode && endIndex == 0)
             );
             
         if (preSplitNode)
             fragment.preSplitNode = preSplitNode;
             
         if (postSplitNode)
             fragment.postSplitNode = postSplitNode;
        
    } // End iterating over affected nodes

    // Determine which nodes are shared with the fragment's range and document and which aren't
    markSharedNodes(fragRoot);
    
    return fragRoot;
    
    /**
     * An inner support function.
     * Sets the "isShared" flag for a fragment and all it's descendant fragments
     * @param {de.dom.DOMFragment} currentFrag The fragment to mark from
     */
    function markSharedNodes(currentFrag) {

        var isAnyChildShared = false;
        
        // Scan children first (disconnecting post order)
        for (var ch in currentFrag.children) {
        
            var childFrag = currentFrag.children[ch];
            
            // Recurse
            markSharedNodes(childFrag);
            
            // Detect if any children of the current node are shared
            isAnyChildShared |= childFrag.isShared;
        }
        
        // Determine if this fragment is shared. The root is a special case... thus explicitly must
        // be declared as being shared. isShared will already be true if the fragment is a boundry 
        // node (start/end) that has been excluded from the range.
        currentFrag.isShared |= (isAnyChildShared ||
            currentFrag == fragRoot ||
            currentFrag.node.childNodes.length != currentFrag.children.length);
            
    } // End inner markSharedNodes
    
}; // End buildFragment

// Expose build fragment
de.buildFragment = _buildFragment;
