/*
 * file: Spell.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */

bootstrap.provides("Spell");

$enqueueInit("Spell", function() {
    
    // Setup auto-selection of marked errors
    _addHandler(_engine == _Platform.GECKO ? window : document, "mouseup", function() {
        
        var sel = de.selection.getRange();
        if (sel && !sel.endNode) {
            
            // The user clicked in an editable section... i.e. did not range a selection
            var spellNode = de.spell.getMarkedAncestor(sel.startNode);
            if (spellNode) {
                
                // The user click inside inline content marked as a spelling error
                // so select the whole word
                de.selection.setSelection({
                    startNode:spellNode,
                    startIndex:0,
                    endNode:spellNode,
                    endIndex:1
                });
            }
        }
        
    });
    
}, "Selection"); // add mouseup event after selection's event registration

(function() {
    
    var 
        /**
         * The classname of the wrappers used for marking spelling mistakes
         * @final
         * @type String
         */
        SPELL_MARK_CLASS_NAME = "sw-spell-error",
        SPELL_MARK_CLASS_NAME_RE = /^sw-spell-error$/;
    
    /**
     * @namespace
     */
    de.spell = {
        
        /**
         * @param {[Element]} editableSections (Optional) The editable sections to get words from.
         *                                     If not provided then all editable sections will be queried.
         * 
         * @return {[String]} The <em>set</em> of words found in all the given editable sections.
         */
        getWords : function(editableSections) {
            
            if (!editableSections)
                editableSections = de.doc.getAllEditSections();
            
            var words = [], wmap = {};
            
            for (var i in editableSections) {
                visitWords(editableSections[i], function(word) {
                    if (!wmap[word]) { // Unique word?
                        words.push(word);
                        wmap[word] = 1; // Track words to build a set
                    }
                });
            }
            
            return words;
            
        },
        
        /**
         * Scans a given set of editable sections for spelling errors, for each spelling error found
         * it wraps them with a span with the spelling error class. 
         * 
         * Calling this may result in a undoable action executed
         * via the undo manager (where the users can undo themselves).
         * 
         * @param {[String]|String} errors An array of miss-spelt words - or a string of whitespace seporated words
         * 
         * @param {[Element]} editableSections (Optional) An array of editable sections to mark errors for.
         *                                     If not provided then all editable sections will be queried.
         */
        markWords : function(errors, editableSections) {
            
            if (!editableSections)
                editableSections = de.doc.getAllEditSections();
                
             // Convert error array/string into a lookup map
             errors = $createLookupMap(typeof errors == "string" ?  errors.replace(/\s/g,',') : errors.join(','));
             
             var foundError = 0;
             
             for (var i in editableSections) {
                visitWords(editableSections[i], function(word, startNode, startIndex, endNode, endIndex) {
                    
                    // Is this word considered an error?
                    if (errors[word]) {
                        // Wrap the word and get the fragment of the selected word.
                        // Group all the actions together as one unit.
                        var frag = de.UndoMan.execute(
                            foundError ? de.UndoMan.ExecFlag.GROUP : 0, 
                            'SpellMark',
                            startNode, 
                            startIndex, 
                            endNode, 
                            endIndex + 1);
                            
                        foundError = 1;
                        return frag.children[frag.children.length-1].node; // Return the (exclusive) resume node as the last text node that made up the word
                    }
                    
                    
                });
            }
            
        },
        
        /**
         * Clears all words marked with spelling errors in a given set of editable sections.
         * This operation will add a single undoable action to the undo manager if any
         * marked spelling errors are found.
         * 
         * @param {[Element]} editableSections (Optional) An array of editable sections to clear marked errors from.
         *                                     If not provided then all editable sections will be queried.
         *                                     
         * @return {Boolean} Evaluates true if some errors were unmarked (i.e. an undoable action was added
         *                   to the undo managaer).
         */
        clearAllMarks : function(editableSections) {
            
            if (!editableSections)
                editableSections = de.doc.getAllEditSections();
            
            var errorWrappers = [], i;
            
            for (i in editableSections) {
                
                // Find all spelling error wrappers in each editable section
                _visitAllNodes(editableSections[i], editableSections[i], true, function(domNode) {
                    if (de.spell.isSpellErrorWrapper(domNode))
                        errorWrappers.push(domNode);   
                });
            }
            
            // Remove all the wrappers
            for (i in errorWrappers) {
                de.UndoMan.execute(i == '0' ? 0 : de.UndoMan.ExecFlag.GROUP, "SpellUnmark", errorWrappers[i]);
            }
            
            // If nothing was found then zero will be returned.
            return errorWrappers.length;
            
        },
        
        /**
         * Clears a specific marked spelling error.
         * This will add an action to the undo manager.
         * 
         * @param {Element} markNode The spelling error wrapper to "ignore" (i.e. remove)
         */
        ignoreError : function(markNode, execFlags) {
            de.UndoMan.execute(execFlags ? execFlags : 0, 'SpellUnmark', markNode);
        },
        
        /**
         * Replaces an error with a correction.
         * This will add an action to the undo manager.
         * 
         * @param {Element} markNode   The spelling error wrapper to "correct" (i.e. replace)
         * @param {String} correction  The word to replace the spelling error with
         */
        correctError : function(markNode, correction) {
            de.UndoMan.execute('SpellCorrect', markNode, correction);
        },
        
        /**
         * @param {Node} node A dom node
         * @return {Element} The spelling error wrapper which is either the given dom node or a ancestor of the given dom node.
         */
        getMarkedAncestor : function(node) {
            return this.isSpellErrorWrapper(node) ? node : _findAncestor(node, this.isSpellErrorWrapper);
        },
        
        /**
         * @param {Node} node  The node to test
         * @return {Boolean}   True iff the node is a spelling error wrapper element.
         */
        isSpellErrorWrapper : function(node) {
            return _findClassName(node, SPELL_MARK_CLASS_NAME_RE) ? true : false;
        },
        
        /**
         * Strips spelling wrapper tags from HTML
         * 
         * @param {String} html HTML Markup.
         * 
         * @return {String} The given HTML with all spelling error wrappers removed.
         */
        stripSpellWrapperHTML : function(html) {
            
            var re = /<span class\s*=\s*(?:"|')sw-spell-error(?:"|')\s*>([^<]+)<\/span>/i;
                
           // TODO: Can I just use replace rather than dealing with loops all the time?
           
           // Remove spelling wrappers via RE
            while (match = re.exec(html)) {
                html = html.substr(0, match.index) + match[1] + html.substr(match.index + match[0].length);
            }

            return html;
        }
        
        
    };

    /**
     * Scans for words in an editable section.
     * 
     * @param {Element} editableSection    The editable section to scan words for
     * 
     * @param {Function} callback          The callback function to send words to. Takes 5 args:
     *                                     word (string) -         The word that was found
     *                                     startNode (Node) -      The start text node of the word
     *                                     startIndex (Number) -   The start index of the word within the start text node 
     *                                     endNode (Node) -        The end text node of the word
     *                                     endIndex (Number) -     The end index of the word within the end text node 
     *                                     
     *                                     Return the node to exclusively resume the in-order traversal from
     */
    function visitWords(editableSection, callback) {
        
        var wordBreakerChars = /^[^\w']$/,
            resumeNode = editableSection,
            startNode,
            startIndex,
            endNode,
            endIndex,
            curWord;
        
        while (resumeNode) {
            
            // Continue visiting nodes within the editable section...
            _visitAllNodes(editableSection, resumeNode, true, function(domNode) {
                
                // Skip resume node (exclusive start point)
                if (resumeNode == domNode)
                    resumeNode = 0;  // Clear current resume node
                
                else {
                    
                     // For all text nodes within an element (e.g. not within comments, styles or scripts)...
                    if (domNode.nodeType == Node.TEXT_NODE && domNode.parentNode.nodeType == Node.ELEMENT_NODE) {
                        
                        // Only allow words to extend between adjacent text nodes .. otherwise
                        // wrapping the words will become over complex and bloat the code.
                        if (endNode && domNode.previousSibling != endNode) {
                            if (checkWord())
                                return false; // NB: Will revisit this node since will exclusively resume at the previous sibling
                        }
                       
                        var str = domNode.nodeValue, i;
                       
                        // For each charactor in this text node's string
                        for (i = 0; i < str.length; i++) {
                            
                            var c = str.charAt(i);
             
                            // Is this charactor a part of a word or a word breaker?
                            if (wordBreakerChars.test(c)) {
                                
                                // If so then check for a pending word...
                                if (checkWord())
                                    return false;
                                    
                            } else { // Found word symbol
                                
                                // Set start node/index for start of new word
                                if (!startNode) {
                                    startNode = domNode;
                                    startIndex = i;
                                    curWord = "";
                                }
                                
                                // Build word
                                curWord += c;
                                
                                // Track end point
                                endNode = domNode;
                                endIndex = i;
                            }
                            
                        } // End loop: parsing words in text node
                        
                    }
                    
                }
                
            });
            
        } // End loop: visiting nodes within the given editable section

        // Check any pending word....        
        checkWord();

        /**
         * Checks if there is a pending word, if there is then the callback is invoked.
         * @return {Node} The node to exclusively resume from if the callback changed the DOM.
         */
        function checkWord() {
            
            // Is there a word pending? Exclude words in a package or within a protected heirarchy
            if (startNode 
                && endNode
                && !de.doc.isProtectedNode(startNode) 
                && !de.doc.isNodePackaged(startNode)) {
                
                // Invoke callback
                resumeNode = callback(curWord, startNode, startIndex, endNode, endIndex);
                
                // Reset start/end point.
                startNode = endNode = 0;
                
                // Return result
                return resumeNode;
            }
            
            // Reset start/end point.
            startNode = endNode = 0;
        }
        
    } // End visitWords function
    
    
    _registerAction("SpellMark", {
        
        /**
         * An undoable action: wraps a adjacent group of text nodes with a spelling error wrapper
         * 
         * @param {Node} startNode      
         * @param {Number} startIndex
         * @param {Node} endNode
         * @param {Number} endIndex
         */
        exec : function(startNode, startIndex, endNode, endIndex) {
           
            // Build a fragment - isolating the word within the text node
            var frag = _buildFragment(null, startNode, startIndex, endNode, endIndex),
                 wrapper = $createElement("span");
            
            // Wrap the text with a span
            _setClassName(wrapper, SPELL_MARK_CLASS_NAME);
            _execOp(_Operation.INSERT_NODE, wrapper, frag.node, frag.children[0].pos);
            
            for (var i = 0; i < frag.children.length; i++) {
                var migrant = frag.children[i].node;
                _execOp(_Operation.REMOVE_NODE, migrant);
                _execOp(_Operation.INSERT_NODE, migrant, wrapper);
            }
            
            return frag;
        }
        
    });
    
    _registerAction("SpellUnmark", {
    
        /**
         * Removes a wrapper
         * 
         * @param {Element} markedNode A mark wrapper element to remove...
         */
        exec : function(markedNode) {
            
            debug.assert(markedNode && de.spell.isSpellErrorWrapper(markedNode));
    
            // Decide whether to remove the wrapper
            /*var shouldRemove = _getClassName(node) != SPELL_MARK_CLASS_NAME;
            if (!shouldRemove) 
                shouldRemove = _doesHaveElementStyle(node);
                */
            
            // TODO: REFACTOR THIS OPERATION... IT IS USED EVERYWHERE
            while(markedNode.firstChild) {
                var migrant = markedNode.firstChild;
                _execOp(_Operation.REMOVE_NODE, migrant);
                _execOp(_Operation.INSERT_NODE, migrant, markedNode.parentNode, _indexInParent(markedNode));
            }
            _execOp(_Operation.REMOVE_NODE, markedNode);
            
        }
        
    });
    
    _registerAction("SpellCorrect", {
        
        /**
         * Replaces an error with a correction.
         * 
         * @param {Element} markNode   The spelling error wrapper to "correct" (i.e. replace)
         * @param {String} correction  The word to replace the spelling error with
         */
        exec : function(markedNode, correction) {
            _execOp(_Operation.INSERT_NODE, document.createTextNode(correction), markedNode.parentNode, _indexInParent(markedNode));
            _execOp(_Operation.REMOVE_NODE, markedNode);
        }
        
    });

    
})();

