/*
 * file: Typing.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */
/*
 * TODO: RESTRUCTURE TO SUPPORT ADDING AND REMOVING OF COMMANDS IN SOURCE
 * ...ONCE PROJECT ACTIONS ARE ALL DONE... intertwine with actions... as well as command filtering
 *  
 * Implements typical word processor key stroke behaviour in the web browser
 */
bootstrap.provides("Typing");

/**
 * The typing singleton is just a model which fires a single event:
 * onTyping, which passes three arguments:
 * 	    typingEvent:   An object containing a cancel member, when flagged as true then the typing module will
 *                     not execute typing logic.
 *                     
 *      event:         The DOM event
 *      
 *      normalizedKey: The string containing the normalized key
 *      
 * This is called in a keystroke event on the document.
 * 
 * @see _addHandler
 */     
 
de.Typing = {}; // Model

(function(){

    $enqueueInit("Typing", function() {
        _addHandler(document, "keystroke", onKeyStroke);
        _model(de.Typing);
    }, "MVC");
    
    function onKeyStroke(e, normalizedKey) {
        
        debug.println("Key-Stroke: '" + normalizedKey + "'");
        
        // Fire typing event
        var typingEvent = {cancel:false};
        de.Typing.fireEvent("Typing", typingEvent, e, normalizedKey);
    
        // Cancel event if requested
        if (typingEvent.cancel)
            return;

        // Is the CTRL (or Apple-func key for mac) down?
        if (de.events.Keyboard.isAcceleratorDown(e)) {
        
            switch (normalizedKey.toLowerCase()) {
                case "z":
                    de.UndoMan.undo();
                    return false;
                    
                case "y":
                    de.UndoMan.redo();
                    return false;
            }
            
        }

        // Don't manipulate the selection if it is not editable
        if (!de.selection.isRangeEditable() || !de.cursor.exists()) 
            return;
			
		var targetES = de.doc.getEditSectionContainer(de.cursor.getCurrentCursorDesc().domNode);
			
        // Is the CTRL (or Apple-func key for mac) down?
        if (de.events.Keyboard.isAcceleratorDown(e)) {
        
            switch (normalizedKey.toLowerCase()) {

                case "b":
                    toggleFormat("bold");
                    return false;
                    
                case "i":
                    toggleFormat("italics");
                    return false;
                    
                case "u":
                    toggleFormat("underline");
                    return false;
                    
                case "a": // Select all
                    de.selection.selectAll(targetES);
                    return false;
                
                    
            }
            
        } else if (!e.ctrlKey && !e.metaKey && !e.altKey) {
        
            var keyStr = normalizedKey;

            if (keyStr) {
            
                if (keyStr.length > 1) {
                
                    switch (keyStr) {
                        case "Space":
                            keyStr = " ";
                            break;
                        
                        case "Tab": // Indentation / promotion
                            
							if (_isBlockLevel(targetES) || targetES == docBody) {
                                
                                var firstBlock = _findAncestor(de.cursor.getCurrentCursorDesc().domNode, docBody, _isBlockLevel, 1);
                                
                                if (firstBlock && _nodeName(firstBlock) == "li") {
                                    
                                    if (e.shiftKey)
                                        de.UndoMan.execute("DemoteItem");
                                    else de.UndoMan.execute("PromoteItem");
                                    
                                } else de.UndoMan.execute("Indent", !e.shiftKey);
 
								
							}
                            
                            return false;
                            
                        case "Delete":
                        case "Backspace":
                            
                            if (de.selection.remove()) {
                                return false;
                                
                            } else {
                            
                                // Get the cursor descriptor
                                var cursorDesc = de.cursor.getCurrentCursorDesc(),
                                    preDesc,
                                    postDesc;
                                
                                if (keyStr == "Backspace") {
                                
                                    if (cursorDesc.isRightOf && (cursorDesc.placement == (de.cursor.PlacementFlag.AFTER | de.cursor.PlacementFlag.BEFORE))) {
                                        // Delete before/after nodes completely if cursor is directly after them
                                        preDesc = _clone(cursorDesc);
                                        preDesc.isRightOf = false;
                                    } else {
                                        preDesc = de.cursor.getNextCursorMovement(cursorDesc, true);
                                    }
                                    
                                    postDesc = cursorDesc;
                                    
                                } else {
                                    // Delete is just like backspace on the right charactor.
                                    preDesc = cursorDesc;
                                    
                                    if (!cursorDesc.isRightOf && (cursorDesc.placement == (de.cursor.PlacementFlag.AFTER | de.cursor.PlacementFlag.BEFORE))) {
                                        // Delete before/after nodes completely if cursor is directly after them
                                        postDesc = _clone(cursorDesc);
                                        postDesc.isRightOf = true;
                                    } else {
                                        postDesc = de.cursor.getNextCursorMovement(cursorDesc, false);
                                    }
                                    
                                }
                                
                                if (preDesc && postDesc) {
                                
                                    // Check that the range is valid
                                    var isEditable = de.doc.isNodeEditable(postDesc.domNode);
                                    if (isEditable && postDesc.domNode != preDesc.domNode) isEditable &= de.doc.isNodeEditable(preDesc.domNode);
                                    
                                    if (isEditable) {
                                    
                                        var preIndex;
                                        if (preDesc.domNode.nodeType == Node.TEXT_NODE) {
                                            preIndex = preDesc.relIndex;
                                            if (preDesc.isRightOf) preIndex++;
                                        } else {
                                        
                                            if (_nodeName(preDesc.domNode) == "br") {
                                            
                                                // If the pre desc is a line break, then count how many other line breaks there are between the pre/post desc
                                                var brCount = 0;
                                                _visitNodes(_getCommonAncestor(preDesc.domNode, postDesc.domNode), preDesc.domNode, true, null, function(domNode){
                                                    if (_nodeName(domNode) == "br") brCount++;
                                                    return domNode != postDesc.domNode;
                                                });
                                                
                                                // If there are more than one line breaks, then exclude the starting line break... since only
                                                // one line break should be removed
                                                preIndex = brCount > 1 ? 1 : 0;
                                                
                                            } else 
                                                preIndex = preDesc.isRightOf ? 1 : 0;
                                        }
                                        
                                        var postIndex;
                                        if (postDesc.domNode.nodeType == Node.TEXT_NODE) {
                                            postIndex = postDesc.relIndex;
                                            if (postDesc.isRightOf) postIndex++;
                                        } else {
                                        
                                            // If the pre desc is a line break, then count how many other line breaks there are between the pre/post desc
                                            if (_nodeName(postDesc.domNode) == "br") {
                                                var brCount = 0;
                                                _visitNodes(_getCommonAncestor(preDesc.domNode, postDesc.domNode), preDesc.domNode, true, null, function(domNode){
                                                    if (_nodeName(domNode) == "br" && !(domNode == preDesc.domNode && preIndex == 1)) brCount++;
                                                    
                                                    return domNode != postDesc.domNode;
                                                });
                                                
                                                // If there are more than one line breaks, then exclude the ending line break... since only
                                                // one line break should be removed
                                                postIndex = brCount > 1 ? 0 : 1;
                                                
                                            } else 
                                                postIndex = postDesc.isRightOf ? 1 : 0;
                                            
                                        }
                                        
                                        // Get virtual range since cursor node/index may probably be affected by the selection
                                        var preVNI = de.selection.getVirtualNodeIndex(preDesc.domNode, preIndex),
                                            postVNI = de.selection.getVirtualNodeIndex(postDesc.domNode, postIndex);
                                            
                                            
                                        // See if can use light-weight remove text action
                                        if (preVNI.node == postVNI.node && preVNI.node.nodeType == Node.TEXT_NODE) 
											de.UndoMan.execute("RemoveText", preVNI.node, preVNI.index, postVNI.index - preVNI.index);
										else {
											// Check to see that the range does not remove any editable sections
							                var ca = _getCommonAncestor(preVNI.node, postVNI.node, true);
							                if (de.doc.isEditSection(ca) || de.doc.isNodeEditable(ca)) 
												de.UndoMan.execute("RemoveDOM", preVNI.node, preVNI.index, postVNI.node, postVNI.index);
										}
                                    }
                                }
                                
                                // signal that no further event processing from further up the event stack is needed
                                return false;
                                
                            }
                                
                        case "Enter":
                        
                            de.selection.remove();
                            var cursorDesc = de.cursor.getCurrentCursorDesc();
                            debug.assert(cursorDesc != null);
                			
                            var index = cursorDesc.relIndex;
                            if (cursorDesc.domNode.nodeType == Node.TEXT_NODE) 
                                index += (cursorDesc.isRightOf ? 1 : 0);
                            else index = cursorDesc.isRightOf ? 1 : 0;
                            
							if (!de.doc.getEditProperties(targetES).singleLine) {
								if (e.shiftKey) {
								
									var lbHTML = "<br>";
									// Line breaks may need place holders too
									if (!(cursorDesc.domNode.nodeType == Node.TEXT_NODE &&
									index < _nodeLength(cursorDesc.domNode))) lbHTML += _getOuterHTML(de.doc.createMNPlaceholder());
									
									de.UndoMan.execute("InsertHTML", lbHTML, cursorDesc.domNode.parentNode, cursorDesc.domNode, index);
									
								} else if (_isBlockLevel(targetES) || targetES == docBody) {
									// Should never allow block container to be created in inline editable sections! (Invaid HTML)
									de.UndoMan.execute("SplitContainer", cursorDesc.domNode, index);
								}
							}
                            
                            return false;
                            
                        case "Home":
                        case "End":
                            
                            // Set the cursor to either the beggining or end of the current editable section.
                            
                            var isHome = keyStr == "Home", 
                                boundDesc = de.cursor.getNearestCursorDesc(targetES, isHome ? 0 : 1, !isHome, !isHome);
                            
                            if (boundDesc) {
                                de.cursor.setCursor(boundDesc);
                                de.cursor.scrollToCursor();
                                return false;
                            }
                            
                        default:
                            return true;
                    
                    }
                }
                
                // The key press is a printable charactor.
                
                // Remove any selection there might be
                de.selection.remove();
                
                // Get the cursor descriptor
                var cursorDesc = de.cursor.getCurrentCursorDesc();
                debug.assert(cursorDesc != null);

                // Calculate the cursor index
                var index = cursorDesc.relIndex;
                if (cursorDesc.domNode.nodeType == Node.TEXT_NODE && cursorDesc.isRightOf) 
                    index++;
                else if (cursorDesc.domNode.nodeType == Node.ELEMENT_NODE) 
                     index = cursorDesc.isRightOf ? 1 : 0;
    
                // Perform a text insert action via the undo manager
                de.UndoMan.execute("InsertText", cursorDesc.domNode, keyStr, index);
                
                return false;
                
            }
            
            
        }
        
    }
    
    /**
     * 
     * @param {String} formatType
     * @return {Boolean} True if something was formatted, false if not
     */
    function toggleFormat(formatType) {
        // Perform the action via the undo manager
        return de.UndoMan.execute("Format", formatType, !de.selection.getEditState([formatType]).formatStates[formatType]) ? true : false;
    }
    
    
})();
