/*
 * file: UndoMan.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */

// @DEPENDS: MVC collections.DoublyLinkedList

bootstrap.provides("UndoMan");
    
/**
 * Registers an undoable action to the action repository.
 * @param {String} name    The unique name of the action
 * @param {Object} action  The action object
 */
var _registerAction = function() {alert("hit!!!");};

(function(){

    var actionRepository = {},
        history = new _DoublyLinkedList(),
    
    /* Always points to the next action to be undoed.
       Null if there is no undo history.
       Note that it can be null if there is redo history...*/
        currentActionNode = null,
        
        /* Non-zero if executing an action within an action. It represents the action exec depth, typically it would be 0-1, but sometimes 2. */
        execActionDepth = 0;

    // Setup registor action logic
    _registerAction = function(name, action) {
        debug.assert(!actionRepository[name], "Already registered action for " + name);
        actionRepository[name] = action;
    };
    
    /**
     * @class
     *
     * The undo manager singleton subject provides facilities for executing, undoing and redoing
     * de.actions.UndoableAction's.
     * <br><br>
     * Before a action is executed/undon/redone, a "onBeforeAction" event is fired, 
     * where the argument is the action about to be executed/undon/redone.
     * <br><br>
     * After a action is executed/undon/redone, a "onAfterAction" event is fired,
     * where the argument is the action that has been executed/undon/redone.
     * 
     * @borrows de.mvc.AbstractSubject#addObserver as this.addObserver
     * 
     * @borrows de.mvc.AbstractSubject#removeObserver as this.removeObserver
     */
    de.UndoMan = {
        /**
         * Add cap to avoid consuming too much memory... < 0 = unlimited
         */
        maxHistoryCount : 100, 
        
        ExecFlag : {
			
            GROUP : 1, // @REPLACE de.UndoMan.ExecFlag.GROUP 1
            
            UPDATE_SELECTION : 2, // @REPLACE de.UndoMan.ExecFlag.UPDATE_SELECTION 2
            
            /**
             * If provided then the undo manager will not store the undoable operations for undoing/redoing -
             * It will leave the current operation list in tact after execution and thus the action
             * will not be undone/redone directly by the undo manager.
             * 
             * Used for executing actions within an action, or other internal specialized situations.
             */
            DONT_STORE : 4 // @REPLACE de.UndoMan.ExecFlag.DONT_STORE 4
        },
        
        /**
         * Exposure of _registerAction internal.
         * @see _registerAction
         */
        registerAction : _registerAction,
        
        /**
         * 
         * @param {Number} flags (Optional) NOTE: If currently executing in an action, then the DONT_STORE flag
         *                                  will be automatically set. This allows actions to be combined into one
         * 
         * @param {String} actionName
         * 
         * @return {Object} Action specific result
         */
        execute: function(flags, actionName) {
            
            // Setup arguments
            var args = Array.prototype.slice.call(arguments);

            // Set default flags            
            if (typeof flags != "number") {
                flags = de.UndoMan.ExecFlag.UPDATE_SELECTION;
                 actionName = args[0];
                 args.shift();
            } else args.splice(0, 2);
            
            // If already executing in a action
            if (execActionDepth)
                flags = de.UndoMan.ExecFlag.DONT_STORE;
                
            // @DEBUG ON
            // This can be helpful!!
            if (execActionDepth) {
                debug.println("WARNING: executing an action within an action - Undo man setting exec flags to DONT_STORE");
            }
            // @DEBUG OFF
            
            if((flags & de.UndoMan.ExecFlag.GROUP) && !currentActionNode)
                _error("Cannot group action to nothing");
            
            // Check that the action exists   
            if (!actionRepository[actionName])
                _error("Unknown action called \"" + actionName + "\"");
        
            var result, 
                action = actionRepository[actionName], 
                actionData = new _ActionData(actionName, flags, de.selection.getRange(false), de.selection.getRange(true));

            // Apply action filtering on selection
            if (actionData.selBefore) {
                var eProps = de.doc.getEditProperties(actionData.selBefore.startNode);

                if (eProps && eProps.afRE) {
                    var reEval = eProps.afRE.test(actionName.toLowerCase() + (actionName == "Format" ? args[0].toLowerCase() : ""));
                    if (reEval != eProps.afInclusive)
                        return;
                }

            }

			// Notify observers
			this.fireEvent("BeforeExec", actionData);
            
            // Safety check: there shouldn't be any operations in the current op list if storing them here
            debug.assert((flags & de.UndoMan.ExecFlag.DONT_STORE) || !_getOperations());
			
            // Execute the operation
            execActionDepth++;
            try {
                result = action.exec.apply(actionData, args);
            } finally {
                execActionDepth--;
            }

			// Add to undo history? I.E: Not returning ops
            if (!(flags & de.UndoMan.ExecFlag.DONT_STORE)) {
                
                var opList = _getOperations();
            
                // Did anything occur?
                if (!opList || opList.length == 0) {
                    
                    this.fireEvent("AfterExec", actionData);
                    
                    // Restore selection to state before action
                    if (actionData.selBefore)
                        de.selection.setSelection(actionData.selBefore.startNode, actionData.selBefore.startIndex, actionData.selBefore.endNode, actionData.selBefore.endIndex, true);
                    
                    return result;
                }
                
    			// Destroy any redo-history
    			// If the current undo marker is at the very beggining of the
    			// list, then reset the list.
    			if (!currentActionNode) 
    				history.clear(); /* Note: If history is already clear then this is ok. */
    			else if (currentActionNode.next) 
                    history.chop(currentActionNode);
                
                // Store the action's operations
                actionData.opList = opList;
                history.push(actionData);
                
                // Update the undo marker
                currentActionNode = history.tail;
                
                // Check for max history
                if (history.length > this.maxHistoryCount && this.maxHistoryCount > -1 && !(action.flags & de.UndoMan.ExecFlag.GROUP)) 
                    history.removeAtIndex(0);
                
            }
            
            // Should update the selection?
			if (flags & de.UndoMan.ExecFlag.UPDATE_SELECTION) {
            
                var selAfter = actionData.selAfter;
                if (!selAfter)
                    de.selection.clear();
            
                else de.selection.setSelection(selAfter.startNode, selAfter.startIndex, selAfter.endNode, selAfter.endIndex, true);
            
            } else {
               
                // No need to keep the selection snapshots
               delete actionData["selBefore"];  
               
               if (actionData["selAfter"]) 
                   delete actionData["selAfter"];
            } 
			
            // No need to keep order selection range (only used for action exec benifit)	
            delete actionData["selBeforeOrdered"];
                
			// Notify observers
			this.fireEvent("AfterExec", actionData);

            return result;
            
        },
   
        /**
         * Undos the last action.
         */
        undo: function(){
			
			do {
	        
	            // If the currentActionNode is already before the head, or there is
	            // no history, then return.
	            if (!currentActionNode)                 
	                return;
	            
            	var actionData = currentActionNode.data;
	            
	            // Notify observers
	            this.fireEvent("BeforeUndo", actionData);
	
	            try {
	            
                    // Undo the operation
                    _undoOperations(actionData.opList);
	                
	                // Shift the action node along
	                currentActionNode = currentActionNode.prev;
	                
	            } catch (err) {
                    
	                // If the undo failed, the all undo/redo history can become out of
	                // sync with the DOM. Therefore lose the history to avoid bugs from
	                // snowballing into something worse.
	                this.clear();
                    
	                throw err;
	            }
				
				// Restore selection
				if (actionData.flags & de.UndoMan.ExecFlag.UPDATE_SELECTION) {
                    var selBefore = actionData.selBefore;
                    if (!selBefore)
                        de.selection.clear();
                    else de.selection.setSelection(selBefore.startNode, selBefore.startIndex, selBefore.endNode, selBefore.endIndex, true);
				}
	            
	            // Notify observers
	            this.fireEvent("AfterUndo", actionData);
				
			} while (currentActionNode && (actionData.flags & de.UndoMan.ExecFlag.GROUP));
            
        },
        
        /**
         * Redo's the last undo.
         */
        redo: function(){
			
			var firstRedo = true;
			
            while(1) {
        
	            if (history.length == 0 || currentActionNode == history.tail)                 
	                return;
				
                // Is the undo-pointer back to the very start of the history?	
				var curAction = currentActionNode ? currentActionNode.next : history.head;
				var actionData = curAction.data;
				
				// Only continue redoing if the action if grouped
				if (!firstRedo && !(actionData.flags & de.UndoMan.ExecFlag.GROUP))
					return; 
					
				currentActionNode = curAction;
				
	            // Notify observers
	            this.fireEvent("BeforeRedo", actionData);
	
	            // Re-execute the operations
	            try {
					_redoOperations(actionData.opList);
	            } catch (err) {
                    
	                // If the undo failed, the all undo/redo history can become out of
	                // sync with the DOM. Therefore lose the history to avoid bugs from
	                // snowballing into something worse.
	                this.clear();
	                throw err;
	            }
				
				// Set the new selection if it was requested to update the selection
	            // with this action
	            if (actionData.selAfter) {
                    var selAfter = actionData.selAfter;
                    if (!selAfter)
                        de.selection.clear();
                    else de.selection.setSelection(selAfter.startNode, selAfter.startIndex, selAfter.endNode, selAfter.endIndex, true);
				}

				firstRedo = false;
				
	            // Notify observers
	            this.fireEvent("AfterRedo", actionData);
	            
			} 
            
        },
        
        /**
         * Clears all undo/redo history
         */
        clear : function() {
            history.clear();
            currentActionNode = null;
        },
		
		/**
		 * @return {Boolean} True if there is any undo history.
		 */
		hasUndo : function() {
			return currentActionNode != null;
		}, 
		
		/**
		 * @return {Boolean} True if there is any redo history.
		 */
		hasRedo : function() {
			return history.length > 0 && currentActionNode != history.tail;
		}

    }; // End undo manager singleton
    
    // Make undo manager a model
    _model(de.UndoMan);
    
})();


var _ActionData = function() {
		
	var cls = function(name, flags, selBefore, selBeforeOrdered) {
		this.name = name;
		this.flags = flags;
        this.selBefore = selBefore;
        this.selBeforeOrdered = selBeforeOrdered;
		/* this.opList = null */
	}
	
	cls.prototype = {
		
		/**
		 * @return {Node} The top-most editable section changed by this action. Undefined if there was none.
		 */
		getEditSection : function() {
			
			if (this.opList) {
				// Infer from operations list
				for (var i in this.opList) {
					var op = this.opList[i];
					for (var mem in op) {
						// Is this a dom node still in the document body?
						if (_isDOMNode(op[mem]) && _isAncestor(docBody, op[mem])) {
							var esNode = de.doc.getEditSectionContainer(op[mem]);
							if (esNode)
								return esNode;
						}
					}
				}
			}
		}
		
	};
	
	return cls;
	
}();

