/*
 * file: WhitespaceUtil.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */

bootstrap.provides("WhitespaceUtil");

var _consolidateWSSeqs, _normalizeNBSP, _convertWSToNBSP;

/*
 * @see http://www.w3.org/TR/html401/struct/text.html
 */
(function() {
    
    /* Elements which can be physically separated by white space. */
    var breakableElements = $createLookupMap("button,img,iframe,map,object"),
    
    /* Inline elements which cannot be regarded as part of a whitespace sequence. */
        nonWSInlineElements = $createLookupMap("br,button,img,iframe,map,object,select,textarea,applet");

    /**
     * This does not create any undoable operations.
     * 
     * @param {Node} targetNode A node to convert all whitespaces to NBSP entities in 
     *                 text nodes which can support non whitespace and has normal whitespace breaking
     */
    _convertWSToNBSP = function(targetNode) {
        _visitTextNodes(targetNode, targetNode, true, function(textNode) {
            if (_doesTextSupportNonWS(textNode) && getWSStyle(textNode) == "normal")
                textNode.nodeValue = textNode.nodeValue.replace(/[\t\n\r ]/g, _NBSP);
        });
    };
        
    /**
     * Consolidates white space. This creates undoable operations.
     * 
     * @param {Node} targetNode         The DOM node and all it's descendants to consolidate.
     * 
     * @param {Boolean} extendRange     If the first text node begins with whitespace, then the first
     *                                  whitespace sequence may start before the target node. If the last text
     *                                  node in the range ends with whitespace, then the last whitespace sequence
     *                                  may end after the target node. Set to true to allow consolidation outside of 
     *                                  the target node, false will truncate whitespace sequences within the target node.
     *         
     */
    _consolidateWSSeqs = function (targetNode, extendRange) {

        // Get the first text node within target node - that is editable
        var ftn;
        _visitTextNodes(targetNode, targetNode, true, function(textNode){
            if (_doesTextSupportNonWS(textNode) && _nodeLength(textNode) > 0) {
                ftn = textNode;
                return false;
            }
        });
        
        // If there are no text nodes then there is nothing to consolidate
        if (!ftn) return;

        // If the first text node contains a whitespace... extend range backward...
        // possibly before the targetnode... to ensure that all preceeding whitespace
        // that is part of the first node/index sequence is included. May over estimate but
        // that is ok.
        var currentNode = targetNode;
        var ignorePreceedingWS = false;
        
        if (extendRange && _isAllWhiteSpace(ftn.nodeValue.charAt(0))) {
            
            _visitAllNodes(null, ftn, false, function(domNode) {
                
                // Skip start node
                if (domNode == ftn) return;

                if (domNode.nodeType == Node.TEXT_NODE) {
                    
                    // Text nodes that do not support not whitespace shouldn't be consolidated...
                     if (!_doesTextSupportNonWS(domNode))
                         return false;
                         
                    // Adjust new node to start consolidating from
                    currentNode = domNode;
    
                    // If the text node contains a nonWS charactor then the range has been extended enough
                    if (!_isAllWhiteSpace(domNode.nodeValue)) {
                        // Set flag to ignore any preceeding whitespace at the starting node (see later)
                        ignorePreceedingWS = true;
                        return false;
                    }
   
                } 
                
                // If the node is not inline, then WS sequences can't spill over these
                else if (!_isInlineLevel(domNode)) return false;
                
            });  
        }

        var seenTargetNode = _isAncestor(targetNode, currentNode),
            currentIndex = 0;

        // Keep traversing through the target node's descendants until all whitespace sequences are 
        // consolidated or completely removed
        while(currentNode) {

            // Get the next whitespace sequence
            var seq = nextWSSequence(currentNode, currentIndex, ignorePreceedingWS, targetNode, seenTargetNode, false, false, extendRange);
            ignorePreceedingWS = false;
            seenTargetNode = seq.seenTargetNode;
            currentNode = seq.resumeNode;
            currentIndex = seq.resumeIndex;
            
            // Was there a whitespace sequence? if so, and the sequence is not using "pre" wrapping then
            // there might be something to consolidate
            if (seq.startNode && getWSStyle(seq.startNode) != "pre") {
                
                // If the whitespace sequence breaks two inline/text elements apart, then adjust the range
                // so that it leaves one whitespace behind
                if (isBreaker(seq.startNode, seq.startIndex, seq.endNode, seq.endIndex)) {
                    
                    // If the whitespace sequence is just one in length, then there is nothing to consolidate
                    if (seq.startNode == seq.endNode && seq.startIndex == (seq.endIndex - 1)) {
                        seq.startNode = null;
                    } else {
                        
                        // Increment start node / index by one whitespace to leave one white space behind
                        // If the start index is larger/equal to the start nodes text length,
                        // the fragment range will include the start node, but exclude it from removal.
                        seq.startIndex++;
                    }
                } 
                
                // Is there anything to consolidate?
                if (seq.startNode) {
                                  
                    // Create the fragment and disconnect it from the document
                    var seqFrag = _buildFragment(_getCommonAncestor(seq.startNode, seq.endNode, false), seq.startNode, seq.startIndex, seq.endNode, seq.endIndex);
                    seqFrag.disconnect();

                    // Keep the current node/index pointer updated
                    var updateTargetNode = currentNode == targetNode;
                    if (currentNode) {
                    
                        var startFrag = seqFrag.getStartFragment(),
                            endFrag = seqFrag.getEndFragment(),
                            updated = false,
                            wasStartSplit = seqFrag.wasStartSplit(), 
                            wasEndSplit = seqFrag.wasEndSplit();
                        
                        
                        // Is this node the same as the start node of the fragment, and was the start node split?
                        if (currentNode == seq.startNode && wasStartSplit) {
                        
                            debug.assert(startFrag.getPreSplitNode() == seq.startNode);
                            debug.assert(_nodeLength(seq.startNode) == seq.startIndex);
                            
                            // Does the index need updating?
                            if (currentIndex >= _nodeLength(seq.startNode)) {
                            
                                var remTextLen = _nodeLength(startFrag.node);
                                
                                // Does the index fall in the removed range?
                                if (currentIndex < (_nodeLength(seq.startNode) + remTextLen))

                                    // If adjusting left, then simply truncate the index to the end of the start node
                                    currentIndex = _nodeLength(seq.startNode) - 1;
                                    
                                // Was both the end node AND start node split at the same node?.. and the node/index
                                // is pointing in the remaining text (right most)?
                                else if (currentNode == seq.endNode && wasEndSplit) {
                                
                                    // Adjust the node to become the remaining text
                                    currentNode = endFrag.getPostSplitNode()
                                    
                                    // Set the index to become relative to the split end node
                                    currentIndex -= (_nodeLength(seq.startNode) + remTextLen);

                                } else 
                                    assert(false);
                            } 
                            
                            updated = true;
                            
                        // Otherwise is this node the same as the end node of the fragment, and was the end node split?
                        } else if (currentNode == seq.endNode && wasEndSplit) {
                        
                            var remTextLen = _nodeLength(endFrag.node);
                            
                            // Does the index fall outside the removed range?
                            if (currentIndex >= remTextLen) {
                            
                                // Adjust the node to become the remaining text
                                currentNode = endFrag.getPostSplitNode();
                                
                                // Set the index to become relative to the split end node
                                currentIndex -= remTextLen;
                                
                                updated = true;
                                
                             // If not, then the node/index should be set to the start or end bounds node/index
                            } else currentNode = null;
                            
                        }
                        
                        if (!updated) {
                            // Determine if the disconnection of the fragment removed this dom node
                            var wasRemoved;
                            if (currentNode) {
                                wasRemoved = false;
                                seqFrag.visit(function(frag){
                                    if (!frag.isShared && frag.node == currentNode) {
                                        wasRemoved = true;
                                        return false;
                                    }
                                });
                            } else 
                                wasRemoved = true;
                            
                            if (wasRemoved) {
                            
                                // If the node was removed, then set the node/index to the starting bounds
                                var frag = startFrag;
                                while (!frag.isShared) {
                                    frag = frag.parent;
                                }
                                
                                // Set the node to become the first shared node on the starting bound...
                                currentNode = frag.node;
                                
                                // The index should be at the end of the start bound if the very-end of start bound still remains in the document,
                                // Otherwise the index should be set to the beggining of the start bound. 
                                // It is possible for the very-end of the start fragment to still be included because if the
                                // sequence is a breaker, then the start index can be incremented exclude the start node.
                                currentIndex = frag == startFrag ? _nodeLength(currentNode, 1) : 0;
                                
                                // If the shared node contains child nodes, then set the current node to become the child at which the startbounds
                                // proceeded from
                                if (currentNode.childNodes.length > 0 && frag.children.length > 0 && frag.children[0].pos > 0) {
                                    currentNode = currentNode.childNodes[frag.children[0].pos - 1];
                                    // Set index to the end of selected node
                                    currentIndex = _nodeLength(currentNode, 1);
                                } 
                            }
                        }
                        
                        // If the current node is the same as the target node, the target node
                        // is a text node that has been split - so update this aswell
                        if (currentNode && updateTargetNode) targetNode = currentNode;
                    }
                    
                }

            }
              
        } // End loop: consolidating whitespaces in target node

    };


    /**
     * Converts any NBSP entities within a given node (and in some cases just outside of the node)
     * into whitespace, only if the conversion won't collapse the whitespace.
     * 
     * This will not create any undoable operations
     * 
     * @param {Node} targetNode The node to normalize all containing non breaking spaces
     */
    _normalizeNBSP = function(targetNode) {
        
        var currentNode = targetNode, 
            currentIndex = 0;

        while (currentNode) {
        
            // Get the next whitespace sequence.. including NBPS's
            var seq = nextWSSequence(currentNode, currentIndex, false, targetNode, true, true, true, true);
            currentNode = seq.resumeNode;
            currentIndex = seq.resumeIndex;
            
            // Is there a whitespace sequence?
            if (seq.startNode) {

                var isWSSeqBreaker = isBreaker(seq.startNode, seq.startIndex, seq.endNode, seq.endIndex),
                    seqTextNodes = [];
                
                //debug.println("Found ws sequence - wordbreaker=" + isWSSeqBreaker + ", endIndex=" + seq.endIndex);
                
                // Get all text nodes in the whitespace sequence into an array
                _visitTextNodes(_getCommonAncestor(seq.startNode, seq.endNode, false), seq.startNode, true, function(textNode) {
                    seqTextNodes.push(textNode);
                    if (textNode == seq.endNode) return false;
                });
                
                // For each text node in the whitespace sequence....
                for (var i = 0; i < seqTextNodes.length; i++) {
                
                    var textNode = seqTextNodes[i];
                    
                    // For each charactor in the whitespace sequence
                    for (var index = (i == 0 ? seq.startIndex : 0); index < (i == (seqTextNodes.length - 1) ? seq.endIndex : _nodeLength(textNode)); index++) {
                    
                      //  debug.println("Checking whitespace at index " + index + "...");
                        
                        if (textNode.nodeValue.charAt(index) == _NBSP) {
                            
                         //   debug.println("Found NBSP at index " + index);
                        
                            // Keep NBSP if the NBSP is at the start or end of the sequence, and the sequence is not
                            // a word breaker
                            if (!(!isWSSeqBreaker &&
                            ((i == 0 && index == seq.startIndex) ||
                            (i == (seqTextNodes.length - 1) && index == (seq.endIndex - 1))))) {
                            
                                // Keep the NBSP if preceded by a whitespace
                                var ch;
                                if (index == 0)
                                    ch = (i > 0) ? seqTextNodes[i - 1].nodeValue.charAt(_nodeLength(seqTextNodes[i - 1]) - 1) : null;
                                else 
                                    ch = textNode.nodeValue.charAt(index - 1);
                                
                                if (!ch || !_isAllWhiteSpace(ch)) {
                                
                                    // Keep the NBSP if proceeded by a whitespace
                                    if (index == (_nodeLength(textNode) - 1)) 
                                        ch = (i < (seqTextNodes.length - 1)) ? seqTextNodes[i + 1].nodeValue.charAt(0) : null;
                                    else 
                                        ch = textNode.nodeValue.charAt(index + 1);
                                    
                              //      debug.println("ch = " + (ch ? ch : "NULL"));
                                    
                                    if (!ch || !_isAllWhiteSpace(ch)) {
                                        
                                        // Otherwise... replace the non breaking space with a whitespace
                                        
                                 //       debug.println("Replacing NBSP at index " + index + " (node length = " + _nodeLength(textNode) + ")");
                                        
                                        textNode.nodeValue = textNode.nodeValue.substr(0, index) + " " + textNode.nodeValue.substr(index + 1);
                                    }
                                }
                                
                            }
                            
                        }
                    } // End loop: iterating over whitespaces in ws seqence
                } // End loop: Iterating over text nodes in ws sequence

            }
            
        } // End loop: searching for whitespace sequences in target node
        
    };

    /**
     * Discovers the start and end points of the next whitespace seqeunce from a given point (inclusive)
     * 
     * @param {Node} initNode The node to search from (towards the right)
     * @param {Number} initIndex The index to search from.
     * @param {Boolean} ignorePreceedingWS True to ignore the initial whitespaces encountered
     * @param {Node} targetNode The node at which the search should reside within.
     * @param {Boolean} seenTargetNode Flag as true if the target node has been visited. 
     * @param {Boolean} includeNBSP True to include non breaking spaces as whitespace, false to only count whitespace.
     * @param {Boolean} ignoreInternalSingleWS True to ignore any single whitespace sequences that are definatly breaking two words apart
     * @param {Boolean} extendRange True to allow the sequences to go past the target node for boundry cases.
     * 
     * @return {Object} An object with the following members:
     *         seenTargetNode - true if the target node was encountered.
     *         resumeNode 	  - The node to resume the search for remaining ws sequences in the target node
     *         resumeIndex 	  - The index to resume the search for remaining ws sequences in the target node
     *         startNode      - The start node of the sequence. Null if there was none.
     *         startIndex 	  - The start index of the sequence (if there was one)
     *         endNode        - The end node of the sequence, if there was one.
     *         endIndex       - The end index of the sequence, if there was one.
     */
    function nextWSSequence(initNode, initIndex, ignorePreceedingWS, targetNode, seenTargetNode, includeNBSP, ignoreInternalSingleWS, extendRange) {

        var resumeNode = null,
            resumeIndex = initIndex,
            startNode, startIndex, endNode, endIndex, startWSStyle, curWSStyle;
        
        // Locate the next whitespace sequence from the current node onwards (if any).
        _visitAllNodes(null, initNode, true, function(domNode) {
            
            // Has the search space exhausted? I.E: Has the traversal gone past the target node's descendants -
            // and at this point isn't looking for any whitespace to consolidate?
            if (seenTargetNode && domNode != targetNode && (!startNode || !extendRange) && !_isAncestor(targetNode, domNode)) // Case: if target is text node, then it can split... and prematurely end search
                return false; // Finished consolidating/removing ws
    
            // Update flag if domnode is the target node
            seenTargetNode |= (domNode == targetNode);
            
            // Set helper: the whitespace CSS style for the current visited dom node
            curWSStyle = getWSStyle(domNode);

            if (startNode) {

                // If building a whitespace sequence, check to see if the ancestors of the starting node - up to
                // the common ancestor of the start node and this current node - can be contained in a whitespace
                // sequence.
                var ca = _getCommonAncestor(domNode, startNode, false);
                var ancestors = _getAncestors(startNode, ca, false, false);
                var terminateSeq = false;
                for (var i in ancestors) {
                    if (!(_isInlineLevel(ancestors[i]) && !nonWSInlineElements[_nodeName(ancestors[i])])) {
                        terminateSeq = true;
                        break;
                    }
                }

                // Whitespace sequences cannot contain different breaking mechanisms.
                if (terminateSeq || curWSStyle != startWSStyle) {
                    resumeNode = domNode;
                    resumeIndex = 0;
                    return false;
                }
            }

            if (domNode.nodeType == Node.TEXT_NODE) {
                
                if (domNode.parentNode.nodeType != Node.COMMENT_NODE) {
                
                    if (!_doesTextSupportNonWS(domNode)) {
                    
                        debug.assert(!ignorePreceedingWS);
                        
                        // If there is potentially something to consolidate, abort this traversal
                        if (startNode) {
                            // Record current position to resume traversal after consolidation
                            resumeNode = domNode;
                            resumeIndex = 0;
                            return false;
                        }
                        
                    } else {
                    
                        // Iterate over charactors in the text run
                        while (resumeIndex < _nodeLength(domNode)) {
                            var ch = domNode.nodeValue.charAt(resumeIndex);
                            if (_isAllWhiteSpace(ch) || (includeNBSP && ch == _NBSP)) {
                            
                                if (!ignorePreceedingWS) {
                                
                                    // Note start/end node/index of whitespace sequence
                                    if (startNode) {
                                        endNode = domNode;
                                        endIndex = resumeIndex + 1;
                                    } else {
                                        startNode = domNode;
                                        startWSStyle = curWSStyle;
                                        startIndex = resumeIndex;
                                        endNode = null;
                                    }
                                }
                                
                            } else { // Non whitespace charactor
                                ignorePreceedingWS = false;
                                
                                // Is there a current sequence that has more than 1 whitespace, or one that resides at the start of
                                // this text run?
                                if (endNode || (startNode && (ignoreInternalSingleWS || startIndex == 0))) {
                                
                                    // Record current position to resume traversal after consolidation
                                    resumeNode = domNode;
                                    return false;
                                    
                                // Ignore any previous single-whitespace sequences, that do not reside at the start of the text run
                                } else 
                                    startNode = null;
                            }
                            
                            resumeIndex++;
                            
                        } // End loop: iterating over charactors in text run
                    }
                    
                } else ignorePreceedingWS = false;

            } else { // Not a text node
                
                ignorePreceedingWS = false;
                
                if (domNode.nodeType != Node.COMMENT_NODE) {
                    // Whitespace sequences can contain a subset of inline elements.
                    if (startNode &&
                    !(_isInlineLevel(domNode) && !nonWSInlineElements[_nodeName(domNode)])) {
                        resumeNode = domNode;
                        resumeIndex = 0;
                        return false;
                    }
                }
                
                // The element at this point can be part of the current whitespace sequence...
                
            }
                
            resumeIndex = 0;

        }); // End visit
        
        // If sequence is one in length, must set end position
        if (startNode && !endNode) {
            endNode = startNode;
            endIndex = startIndex + 1;
        }
        
        return {
            seenTargetNode : seenTargetNode,
            resumeNode : resumeNode,
            resumeIndex : resumeIndex,
            startNode : startNode,
            startIndex : startIndex,
            endNode : endNode,
            endIndex : endIndex
        };
        
        
     }
 
    /**
     * @param {Node} startNode        The starting text node of the whitespace sequence
     * @param {Number} startIndex     The starting index of the whitespace sequence
     * @param {Node} endNode          The ending text node of the whitespace sequence
     * @param {Number} endIndex       The ending index of the whitespace sequence
     * @return {Boolean}              True iff the given whitespace sequence breaks two words/breakable-elements apart.
     */
    function isBreaker(startNode, startIndex, endNode, endIndex) {
        
        var startWSStyle = getWSStyle(startNode);

        // Look to the left
        if (startIndex == 0) {
            var found = false;
            _visitAllNodes(null, startNode, false, function(domNode) { 

                if (domNode == startNode) return; // Skip initial text node
                
                var res = scan(domNode, startNode);
                
                if (!res && found) {
                    // Check that all ancestors up to and excluding the common ancestor of this dom node
                    // and the start node, are all nodes which are breaked by whitespace
                    var ca = _getCommonAncestor(domNode, startNode, false);
                    var ancestors = _getAncestors(domNode, ca, false, false);

                    for (var i in ancestors) {
                        // Reset found flag
                        found = false;
                        
                        // Check ancestor if it does not break on whitespace...
                        if (!scan(ancestors[i], startNode) && !found) 
                            return false;
                    }
                    
                    // Restore flags
                    res = false;
                    found = true;
                        
                    
                }
                return res;
            });
            
            if (!found) return false;
        }

        // Look to the right
        if (endIndex == _nodeLength(endNode)) {
            var found = false;
            _visitAllNodes(null, endNode, true, function(domNode){
                if (domNode == endNode) return; // Skip initial text node

                var res = scan(domNode, endNode);
                
                if (!res && found) {
                    // Check that all ancestors up to and excluding the common ancestor of this dom node
                    // and the end node, are all nodes which are breakable by whitespace
                    var ca = _getCommonAncestor(domNode, endNode, false);
                    var ancestors = _getAncestors(endNode, ca, false, false);

                    for (var i in ancestors) {
                        // Reset found flag
                        found = false;
                        
                        // Check ancestor if it does not break on whitespace...
                        if (!scan(ancestors[i], endNode) && !found) 
                            return false;
                    }
                    // Restore flags
                    found = true;
                    res = false;
                    
                }
                
                return res;
            });
            
            if (!found) return false;
        }
        
        return true;
        
        /**
         * Inner helper function.
         * 
         * Sets the "found" local to true if domnode is considered breakable (in it's context)
         * 
         * @param {Node} domNode The node to check
         * @param {Node} initialNode The end or start node of the scan
         * @return {Boolean} True to continue scanning, false to abort...a result was found.
         */
        function scan(domNode, initialNode) {

            if (domNode.nodeType == Node.TEXT_NODE) {
                if (_nodeLength(domNode) > 0) {
                    found = _doesTextSupportNonWS(domNode); // Non-WS nodes are not breakable
                    return false;
                }
            
            } else if (breakableElements[_nodeName(domNode)]) {
                found = !_isAncestor(domNode, initialNode); // WS Doesn't break from within breakable nodes to outside of them
                return false;
                
            // If hit a block level element or line break before a breakable node, then the sequence must be
            // leading or trailing text.
            } else if(_isBlockLevel(domNode) || _nodeName(domNode) == "br") 
                return false; 

            // Keep looking...
            return true;
        }// End inner scan
        
    }
    
    /**
     * @param {Node} node  A node to get it's whitespace CSS style for.
     * @return {String}    The CSS white-space style for the given node,
     *                     never null/always is a style.
     */
    function getWSStyle(node) {

        var style = _getComputedStyle(node, "white-space");
        
        if (!style) {
            
            // Check if descends from PRE
            do {
                if (_nodeName(node) == "pre") {
                    style = "pre";
                    break;
                }
                node = node.parentNode;
            } while (node && node.nodeType == Node.ELEMENT_NODE);
            
            // Set as normal
            if (!style) style = "normal";
        }
        
        return style;
  
    }
    
})();

