/*
 * file: FormatAction.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */

// @DEPENDS: UndoMan
bootstrap.provides("actions.FormatAction");

_registerAction("Format", {
    
    /**
     * An undoable formatting action. Formats a range of DOM nodes in the document by
     * adjusting their CSS styles.
     * 
     * @param {String} type            The type of formatting you want to apply. Case insensitive. 
     *                                 See TODO for a list of standard formatting actions available.
     * 
     * @param {Object} value           The value to set - dependant on format type. null or empty to clear formatting.
     * 
     * @param {Object} range           (Optional) The DOM Range to apply the formatting to. If omitted then the current
     *                                 selection will be used. Must have members:
     *                                 
     * 	                               - startNode      The starting dom node of the fragments range.
     *
     *                                 - startIndex 	The inclusive start index in the start node.
     * 								        Ranges from 0 to the text length for text nodes.
     *                                      Where 0 indicates that the range begins at the first char, and text length
     *                                      indicates that the range begins directly after the text node, but not including it.
     *                                      <br>
     *                                      Ranges from 0 to 1 for elements.
     *                                      Where 0 indicates that the range includes the element and it's decendants,
     *                                      and 1 indicates that the range excludes the element and it's decendants.
     *
     *
     * 	                                - (optional) endNode        The ending dom node of the fragments range. Omit to select word at given start node/index.
     *
     * 	                                - (optional) endIndex 	The inclusive end index in the end node.  Omit to select word at given start node/index.
     * 								        Ranges from 0 to the text length for text nodes.
     *                                      Where 0 indicates that the range ends just before the text node, but not including it,
     *                                      and text length indicates that the range ends at the last charactor in the text run.
     *                                      <br>
     *                                      Ranges from 0 to 1 for elements.
     *                                      Where 0 indicates that tmFragment(frag) {
     *                                      and 1 indicates that the range includes the element and it's decendants.
     * 
     * @return {_Fragment}            The fragment which represents the range in the DOM which was formatted. Undefined if nothing formatted.
     */
    exec:function(type, value, range) {
        
        type = type.toLowerCase();
        
        // Check if the format type exists
        if (typeof _formatEnvironment[type + "Wrapper"] == "undefined") {
            debug.assert(false, 'Unknown format request: "' + type + '"');
            return;
        }

        var clear = value ? false : true, isWordRange = false, orderSelAfter = true, formatRange;

        // Auto-set range to current selectoin if none is provided
        if (!range) {
        
            if (!this.selBefore)                 
                return; // Nothing to format
            range = _clone(this.selBeforeOrdered);
            formatRange = _clone(range);
            
            // Because the range is based on the current selection, set the selection
            // after to maintain the selection before ordering (i.e. keep the cursor
            // at the same end of the selection).
            orderSelAfter = this.selBefore.inOrder;
            
        } else 
            formatRange = _clone(range);
        
        // Auto-select word-range - if the range is a single node/index tuple
        if (!formatRange.endNode) {
            
            var anchorRange = null;
            
            // Special case: instead of selecting word for links, select the whole anchor
            if (type == "link") {
                
                var current = formatRange.startNode;
                
                while (current && de.doc.isNodeEditable(current)) {
                    if (current.nodeType == Node.ELEMENT_NODE && _nodeName(current) == "a") {
                        anchorRange = {
                            startNode : current,
                            startIndex : 0,
                            endNode : current,
                            endIndex : 1 /* Since a element, 1 indicates complete right-most deepest descandant */
                        };
                        break;
                    }
                    current = current.parentNode;
                }
                
            }
            
            formatRange = anchorRange ? 
                anchorRange
                : de.selection.getWordRangeAt(formatRange.startNode, formatRange.startIndex >= _nodeLength(formatRange.startNode) ? _nodeLength(formatRange.startNode) - 1 : formatRange.startIndex);
            
            if (!formatRange)
                return; // Nothing to format
                
            isWordRange = orderSelAfter = true;
        }
        
        // Execute the action
        var fragmentRoot = formatDOMExec(
            formatRange.startNode, 
            formatRange.startIndex, 
            formatRange.endNode, 
            formatRange.endIndex, 
            _formatEnvironment[type + "Wrapper"](value || ""), 
            clear, 
            _formatEnvironment[type + "Eval"]);
            
        // Update selection if requested
        if (this.flags & de.UndoMan.ExecFlag.UPDATE_SELECTION) {
            
            if (isWordRange) {
            
                // Must get adjusted node/index since fragment will most likely of split some text nodes
                var adjustedRange = fragmentRoot.getAdjustedNodeIndex(range.startNode, range.startIndex);
      
                // Set the selection to be the single point
                this.selAfter = {
                    startNode: adjustedRange.node,
                    startIndex: adjustedRange.index
                };
            
            } else {
            
                var startFrag = fragmentRoot.getStartFragment(), endFrag = fragmentRoot.getEndFragment();
                
                var startIndex = 0, 
                endIndex = _nodeLength(endFrag.node, 1);
                
                this.selAfter = {
                    startNode: orderSelAfter ? startFrag.node : endFrag.node,
                    startIndex: orderSelAfter ? startIndex : endIndex,
                    endNode: orderSelAfter ? endFrag.node : startFrag.node,
                    endIndex: orderSelAfter ? endIndex : startIndex
                };
            }
            
        }
        
        
        /**
         * The workhorse for the format action.
         */
        function formatDOMExec(startNode, startIndex, endNode, endIndex, inlineWrapper, clear, evalElement) {
           
            var isRecursing = arguments.length > 7,
                startExclusive = startIndex == _nodeLength(startNode, 1),
                endExclusive = endIndex == 0;
                
            // Build up fragment - split any text nodes where neccessary
            var fragmentRoot = _buildFragment(_getCommonAncestor(startNode, endNode, false), 
                startNode, 
                startIndex, 
                endNode, 
                endIndex);
                
            var startFrag = fragmentRoot.getStartFragment(),
                endFrag = fragmentRoot.getEndFragment(),
                rootNode = fragmentRoot.node; // Keep track of the root node... it may change if it is removed 
                
            // Mark fragments which should not be included for formatting
            if (startExclusive)
                markFrags(startFrag);
            if (endExclusive)
                markFrags(endFrag);
        
            function markFrags(markFrag) {
                do {
                    markFrag.dontFormat = 1;
                    markFrag = markFrag.parent;
                } while(markFrag && markFrag.children.length == 1);
            }
        
            // PASS 1: Clear all related formatting
            
            // Clear all related formatting within fragment
            fragmentRoot.visit(function(frag){
                if (!frag.dontFormat)
                    removeFormatting(frag.node);
            });
            
            // Clear all related formatting up to the document root (include body since can have styles/classes)
            if (!isRecursing) {
                var domNode = rootNode; // NB: Include root in removal since it may no longer be the fragment root
                while (domNode) {
                    removeFormatting(domNode);
                    if (domNode == docBody) break;
                    domNode = domNode.parentNode;
                } // End clear formatting to to root element
            }
            
            if (!clear) { // Should new formatting be applied? I.E. Not just clearing formatting
                
                // PASS 2: Set new formatting
                
                // Rebuild fragment since structure may have changed (this won't create any extra operations)
                startNode = startFrag.node;
                endNode = endFrag.node;
                var newFragmentRoot = _buildFragment(
                    rootNode, 
                    startNode, 
                    startExclusive ? _nodeLength(startNode, 1) : 0, 
                    endNode, 
                    endExclusive ? 0 : _nodeLength(endNode, 1)
                    );
                    
                if (startExclusive)
                    markFrags(newFragmentRoot.getStartFragment());
                if (endExclusive)
                    markFrags(newFragmentRoot.getEndFragment());
                
                // NB: Keeping old fragment root in order to determine new cursor position (see below)
                
                // Traverse the fragment structure and encapsulate inline groups where needed
                (function trav(frag){
                
                    var igroups = [];
                    
                    // For all sub tress
                    for (var i in frag.children) {
                        
                        var child = frag.children[i];
           
                        if (shouldEncap(child)) 
                            igroups.push(child.node);
                        else {
                            // If cannot wrap the subtree due to containing at least one block level element, then
                            // recurse deeper
                            encapsulateIGroup(); // wrap anything pending
                            trav(child);
                        }
                    }
                    
                    // Encap remaining run if any
                    encapsulateIGroup();
                    
                    /**
                     * Checks the local igroups array and moves any nodes in the array into a wrapper
                     * then clears the array.
                     */
                    function encapsulateIGroup(){
                    
                        if (igroups.length > 0) {
                        
                            // Create inline wrapper and inset at beginning of inline group run
                            var wrapper = inlineWrapper.cloneNode(false);
                            _execOp(_Operation.INSERT_NODE, wrapper, frag.node, _indexInParent(igroups[0]));
                            
                            // Move run of inline groups into the wrapper
                            for (var i in igroups) {
                                var toEncap = igroups[i];
                                _execOp(_Operation.REMOVE_NODE, toEncap);
                                _execOp(_Operation.INSERT_NODE, toEncap, wrapper);
                            }
                            
                            // Reset igroup run
                            igroups = [];
                            
                        }
                    } // End encapsulateIGroup function
                    
                    /**
                     * @param {Node} node The node to test
                     * @return {Boolean} True iff the fragments node should be encapsulated.
                     */
                    function shouldEncap(frag) {
                        
                        // Should this fragment be excluded from formatting?
                        if (frag.dontFormat)
                            return false;
                            
                        // Check that all node inclusive don't contain block-levels 
                        var allin = true;
                        _visitAllNodes(frag.node, frag.node, true, function(domNode) {
                            allin = !_isBlockLevel(domNode);
                            return allin;
                        })
                        
                        if (allin) {
                        
                            // Last check: the fragment mimics the actual document structure.. i.e. not a partial range.
                            // This avoids encapsulating nodes which are not in the format range
                            var isStructureSame = true;
                            (function checkStructure(curFrag, curNode) {
                                
                                // Avoid including fragments which aren't actually in the format range
                                 if (curFrag.dontFormat) {
                                     isStructureSame = false;
                                     return;
                                 }
                                
                                if (curNode.firstChild) {
                                    
                                    isStructureSame = curNode.childNodes.length == curFrag.children.length;
                                    
                                    if (isStructureSame) {
                                        for (var j = 0; j < curNode.childNodes.length; j++) {
                                            checkStructure(curFrag.children[j], curNode.childNodes[j]);
                                            if (!isStructureSame)
                                                break;
                                        }
                                    }
                                    
                                } else {
                                    isStructureSame = curFrag.children.length == 0;
                                }
                                
                            })(frag, frag.node);
                            
                            return isStructureSame;
                            
                        }
                        
                        return false;
                    } // End isAllInline function
                    
                })(newFragmentRoot); // End encapsulating inline groups
                
            } // End if: not clearing formatting
            
            return fragmentRoot;
            
            /**
             * Removes immediate formatting for the given node.
             * May result in removal of node... the local range may be adjusted when removals occur
             * 
             * @param {Node} node a node to remove formatting from
             */
            function removeFormatting(node) {
                
                if (node.nodeType == Node.ELEMENT_NODE) {
                    
                    var shouldRemove = false,
                        res = evalElement(node); // Determine the actions to take to strip formatting at this element
                        
                    if (res) { // Does this element contain related formatting?
                        
                        for (var i in res.strip) {
                        
                            // Strip formatting
                            switch (res.strip[i].type) {
                                case 1: // Name match
                                    shouldRemove = true;
                                    break;
                                    
                                case 2: // Class match
                                
                                    // Remove class... but keep extra classes
                                    var cls = _getClassName(node);
                                    if (cls) { // Strip matched class
                                        var clss = cls.split(' ');
                                        for (var j in clss) {
                                            if (clss[j] == res.strip[i].match) {
                                                clss.splice(j, 1);
                                                break;
                                            }
                                        }
                                        cls = cls.length > 0 ? clss.join(' ') : "";
                                        
                                        _execOp(_Operation.SET_CLASS, node, cls);
                                    }
                                    break;
                                    
                                case 3: // Style match
                                    // Erase the matched style
                                    _execOp(_Operation.SET_CSS_STYLE, node, res.strip[i].match, "");
                                    break;
                                    
                                // @DEBUG ON
                                default:
                                    debug.assert(false, 'Unknown result type "' + res.strip[i].type + '" for stripping formatting');
                                // @DEBUG OFF
                            }
                            
                            // If going to remove then no need to strip anything.
                            if (shouldRemove) break;
                            
                        } // End loop: stripping formatting
                        
                        // Get left-most child of the current sub tree
                        var leftMost = node.firstChild;
                        while (leftMost.firstChild)
                            leftMost = leftMost.firstChild;
                            
                        // Get right-most child of the current sub tree 
                        var rightMost = node.lastChild;
                        while (rightMost.lastChild)
                            rightMost = rightMost.lastChild;
                            
                        // Check if the node should be auto-removed - only if the element is a generic span and has no class/styles.
                        if (!shouldRemove && _nodeName(node) == "span") {
                            shouldRemove = !_getClassName(node);
                            if (!shouldRemove) 
                                shouldRemove = _doesHaveElementStyle(node);
                        }
                            
                        if (shouldRemove) {
                            
                            // Remove all children and connect with parent
                            var migrations = [];
                            while (node.firstChild) {
                                var migrant = node.firstChild;
                                _execOp(_Operation.REMOVE_NODE, migrant);
                                _execOp(_Operation.INSERT_NODE, 
                                        migrant, 
                                        node.parentNode, 
                                        _indexInParent(migrations.length > 0 ? migrations[migrations.length - 1] : node) + 1
                                        );
                                migrations.push(migrant);
                            }
        
                            // If has no children, then ignore ... code wll bloat and it's too complex to bother tidying such cases
                            if (migrations.length> 0) {
        
                                // If node is the current root node then the root must be adjusted
                                if (node == rootNode)
                                    rootNode = node.parentNode;
                        
                                // Remove the actual node
                                _execOp(_Operation.REMOVE_NODE, node);
                                
                            }
        
                        }
                        
                        // Ensure that all of the elements descendants that it not within the formatting range
                        // still inherits the formatting that was just stripped using recursion
                        if (!isRecursing) {
                            
                            // Recurse on left-range?
                            var isInFormatRange = false;
                            fragmentRoot.visit(function(f){
                                isInFormatRange = (!f.dontFormat && f.node == leftMost);
                                return !isInFormatRange;
                            });
                            
                            if (!isInFormatRange) { // Do recursion
        
                                formatDOMExec(
                                    leftMost, 
                                    0, 
                                    startFrag.node, 
                                    startExclusive ? _nodeLength(startFrag.node, 1) : 0, 
                                    res.inline, 
                                    false, 
                                    evalElement, 
                                    true);
                            }
                            
                            // Recurse on right-range?
                            isInFormatRange = false;
                            fragmentRoot.visit(function(f){
                                isInFormatRange = (!f.dontFormat && f.node == rightMost);
                                return !isInFormatRange;
                            });
                            
                            if (!isInFormatRange) { // Do recursion
                                
                                formatDOMExec(
                                    endFrag.node, 
                                    endExclusive ? 0 : _nodeLength(endFrag.node, 1), 
                                    rightMost, 
                                    _nodeLength(rightMost, 1), 
                                    res.inline, 
                                    false, 
                                    evalElement, 
                                    true);
                            }
                        }
                        
                    }
                    
                }
                
            } // End inner function removeFormatting
        
        } // End inner function formatDOMExec
            
    } // End exec
});




