/*
 * file: IndentAction.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */

// @DEPENDS: UndoMan
bootstrap.provides("actions.IndentAction");

(function() {

    /**
     * The amount of pixels to adjust margins when increasing and decreasing indents.
     * @type Number
     */
    var INDENT_WIDTH = 20,
        undentableBlocks = $createLookupMap("dt,dd,caption,colgroup,col,thead,tfoot,tbody,tr,td,th,legend,optgroup,option,area,frame");
    
    _registerAction("Indent", {
        /**
         * An undoable indentation action. Increases or decreases left indents on container elements in a given range.
         * If nodes in the given range should be indented but do not have a container, a new container is created for them.
         * 
         * @author Brook Novak
         * 
         * @param {Boolean} increase    True to increase indents in range, false to decrease indents.
         * 
         * @param {Node} startNode      (Optional) The starting dom node of the range to align.
         *                              If not provided then the current selection will be used.
         *                              If provieded must also provide endNode
         *
         * @param {Node} endNode        (Optional) The ending dom node of the range to align. Can be the same as start node
         *                              If not provided then the current selection will be used.
         * 
         *
         */
        exec : function(increase, startNode, endNode) {

            // Auto-set range if not provided.
            if (!startNode) {
                
                if (!this.selBefore)
                    return; // Nothing to select
                    
                if (this.selBefore.endNode) {
                    startNode = this.selBeforeOrdered.startNode;
                    endNode = this.selBeforeOrdered.endNode;
                } else 
                    startNode = endNode = this.selBefore.startNode;
                
            }
            
            debug.assert(endNode, "Supplied start node but not the end node");

            var containers;
            
            // First check for special case: If the ranges first block level common ancestor
            // is a list item, then indent the list item rather than normalizing within a list item.
            var ca = _getCommonAncestor(startNode, endNode);
            for (var level = 0; level < 2; level++) {
                while (ca != docBody && !_isBlockLevel(ca)) {
                    ca = ca.parentNode;
                }
                
                if (ca == docBody)
                    break;
                    
                if (_nodeName(ca) == "li") {
                    containers = [ca];
                    break;
                }
                ca = ca.parentNode;
            }
            
            if (!containers) 
                // Normalize containers in range and get list of all the containers
                containers = _getNormalizedContainerRange(startNode, endNode);
            
            // Record all indentable containers in the given range
            for (var i in containers) {
                if (!undentableBlocks[_nodeName(containers[i])]) {
    
                    var con = containers[i];
                    
                    // Determine the containers left margin in pixels
                    var marginToParse = (con.style.marginLeft && con.style.marginLeft.toLowerCase() != "auto") ?
                        con.style.marginLeft : null;
                        
                    // If margin is default/auto, get the computed style to make sure
                    if (!marginToParse)
                        marginToParse = _getComputedStyle(con, "margin-left");
                        
                    marginToParse = (marginToParse && marginToParse.toLowerCase() != "auto") ?
                        marginToParse : "0";
                    
                    // Parse intentation numberic value
                    var marginPixels;
                    if (marginToParse.indexOf("%") != -1) {
                        // Convert percentage into pixels by using the containers relative offset
                        marginPixels = con.offsetLeft;
                        
                    } else marginPixels = parseInt(marginToParse);
                    
                    var newMarginLen = marginPixels + ((increase ? 1 : -1) * INDENT_WIDTH);
                    
                    // Clip margin to zero
                    if (newMarginLen < 0) 
                        newMarginLen = 0;
                    
                    // Assign new margin
                    if (marginPixels != newMarginLen)
                        _execOp(_Operation.SET_CSS_STYLE, con, "marginLeft",  newMarginLen + "px");
                    
                }
            }
            
            this.selAfter = this.selBefore;
      
        }
    });
        
})();
