/*
 * file: ItemizeAction.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */

// @DEPENDS: UndoMan
bootstrap.provides("actions.ItemizeAction");

(function() {
    
    /* 
     * Lookup map containing element names which when itemized should be converted to a list item rather than being
     * a descendant of the list items.
     */
    var convertContainers = $createLookupMap("p, div");
    
    _registerAction("Itemize", {
            
        /**
         * An undoable itemize action. Creates/converts/destroys a list of items in a given range
         * 
         * @author Brook Novak
         * 
         * 
         * @param {Boolean} bullets    	True to itemize as bullet list, false to itemize as numbered list.
         * 
         * @param {Node} startNode      (Optional) The starting dom node of the range to align.
         *                              If not provided then the current selection will be used.
         *                              If provieded must also provide endNode
         *
         * @param {Node} endNode        (Optional) The ending dom node of the range to align. Can be the same as start node
         *                              If not provided then the current selection will be used.
         *
         */
        exec : function(bullets, startNode, endNode) {
            
            // Auto-set range if not provided.
            if (!startNode) {
                
                if (!this.selBefore)
                    return; // Nothing to select
                    
                if (this.selBefore.endNode) {
                    startNode = this.selBeforeOrdered.startNode;
                    endNode = this.selBeforeOrdered.endNode;
                } else 
                    startNode = endNode = this.selBefore.startNode;
                
            }
            
            debug.assert(endNode, "Supplied start node but not the end node");
            
            // Perfom execute (recursive operation)
            (function exec(start, end, destroy, normalizedRange) {

                // Get master container
                var masterContainer = _findAncestor(
                    _getCommonAncestor(start, end, true),
                    docBody,
                    _isBlockLevel,
                    true
                ) || docBody, destroyListItems, itemizeContainers;
                
                // Deepen start/end range
                var deepStart = start, deepEnd = end;
                while (deepStart.firstChild) {
                    deepStart = deepStart.firstChild;
                }
                while (deepEnd.lastChild) {
                    deepEnd = deepEnd.lastChild;
                }
                
                // Handle special case 1: range within tables
                if (_isTableElement(masterContainer) && 
                    _nodeName(masterContainer) != "td" && 
                    _nodeName(masterContainer) != "th") {
                    
                    // Get all the table's cells within the range
                    var tabCells = [];
                    _visitAllNodes(masterContainer, masterContainer, true, function(domNode) {
                        if (domNode.nodeType == Node.ELEMENT_NODE && 
                          (_nodeName(domNode) == "td" || _nodeName(domNode) == "th")) {
                           tabCells.push(domNode);
                           return (domNode != deepEnd && !_isAncestor(domNode, deepEnd));
                       }
                       return domNode != deepEnd;
                    });
                   
                    // Check if all cells contains nothing but list items and normalize their contents
                    var areAllListEles = true, normalizedRanges = [];
                    for (var i in tabCells) {
                        var res = areAllListElements(tabCells[i], tabCells[i]);
                        if (res) {
                            areAllListEles &= res.allListEles;
                            normalizedRanges.push(res.nrange);
                        } else normalizedRanges.push(0);
                    }
                    
                    // Recurse on table cell
                    for (var i in tabCells) {
                        if (normalizedRanges[i])
                            exec(tabCells[i], tabCells[i], areAllListEles, normalizedRanges[i]);
                    }
                    
                    return;
                }
                
                // If master container is a list container itself, adjust master container to its parent
                if (_nodeName(masterContainer) == "ul" || _nodeName(masterContainer) == "ol") 
                    masterContainer = masterContainer.parentNode;
                
                // Handle special case 2: range within a list item
                var ca = masterContainer;
                for (var level = 0; level < 2 && ca; level++) {
                    if (_nodeName(ca) == "li") {
                        
                        // Determine if need to destroy list item or convert it to a different type
                        if ((normalizedRange && destroy) ||
                        (!normalizedRange &&
                        ((bullets && _nodeName(ca.parentNode) == "ul") ||
                        (!bullets && _nodeName(ca.parentNode) == "ol")))) {
                            destroyListItems = [ca]; // the list item is the target type
                        } else {
                           itemizeContainers = [ca];
                           masterContainer = ca.parentNode;
                        }
                        break;
                        
                    }
                    ca = _findAncestor(ca.parentNode, docBody, _isBlockLevel, true);
                }
                
                // Check if all containers in range are list-items/list-containers
                if (!normalizedRange && !destroyListItems && !itemizeContainers) {
                    var res = areAllListElements(start, end);
                    
                    if (res) {
                    
                        normalizedRange = res.nrange;
                        
                        // If the range contains all list elements then destroy them
                        if (res.allListEles) 
                            destroyListItems = normalizedRange;
                        else 
                            itemizeContainers = normalizedRange;
                        
                    } else 
                        return;
                    
                } else if (normalizedRange) { 
                    // If recursing on a table cell then the destroy/create operation will be predetermined
                    if (destroy)
                        destroyListItems = normalizedRange;
                    else 
                        itemizeContainers = normalizedRange;
                }
                
                if (destroyListItems) {
                    
                    // Tear down all list-items / list-containers.
                    for (var i in destroyListItems) {
                        
                        var cont = destroyListItems[i];
                        var contName = _nodeName(cont);
                        
                        if (contName == "li") {
                            convertLI(cont, true);
                            
                        } else if (contName == "ul" || contName == "ol") {
                            
                            // Determine range of list items within list container
                            var li = _isAncestor(cont, start) ? _findAncestor(start, cont) : cont.firstChild,
                                endLI = _isAncestor(cont, end) ? _findAncestor(end, cont) : cont.lastChild;
                                
                            while (li) {
                               var removeLI = li;
                               li = li == endLI ? null : li.nextSibling;
                               if (_nodeName(removeLI) == "li") // ensure not white space
                                   convertLI(removeLI, true);
                            } // End loop: destroying range of list items in list container
                            
                        }
                        
                    } // End loop: destroying list items
                    
                } else { // Creation of list items
                
                    debug.assert(itemizeContainers ? true : false);
                
                    // Create the new list container
                    var listContainer;
                    
                    // Itemize all containers in range
                    for (var i in itemizeContainers) {
                        
                        var cont = itemizeContainers[i];
                        var cRoot = cont == masterContainer ? cont : _findAncestor(cont, masterContainer),
                            contName = _nodeName(cont);
                            
                        if (contName == "li") {
                            
                            if (i == '0') {
                                listContainer = convertLI(cont, false);
                            } else {

                                var liParent = cont.parentNode;
                                _execOp(_Operation.REMOVE_NODE, cont);
                                _execOp(_Operation.INSERT_NODE, cont, listContainer);
                                
                                // If the list items container is left without any other list items, remove it
                                var child = liParent.firstChild;
                                while (child && _nodeName(child) != "li") {
                                    child = child.nextSibling;
                                }
                                
                                if (!child)
                                    _execOp(_Operation.REMOVE_NODE, liParent);
                            }
                            
                        } else {
                        
                            if (i == '0') {
                                
                                listContainer = $createElement(bullets ? "ul" : "ol"); 
                                
                                // Determine where to insert the list container
                                if (contName == "ul" || contName == "ol") {
                                    
                                    var shouldInsertBefore = false;
                                    
                                    var firstLI = cont.firstChild;
                                    while (firstLI && _nodeName(firstLI) != "li") {
                                        shouldInsertBefore |= (firstLI == start); // possible start node is white space
                                        firstLI = firstLI.nextSibling;
                                    }
                                    
                                    if (firstLI)
                                        shouldInsertBefore |= _isAncestor(firstLI, start);
                                    
                                    // Insert the list container before this list container if
                                    // the range starts at the first list item.
                                    _execOp(_Operation.INSERT_NODE, listContainer, cRoot.parentNode, _indexInParent(cRoot) + (shouldInsertBefore ? 0 : 1));
                                    
                                        
                                } else 
                                    _execOp(_Operation.INSERT_NODE, listContainer, cRoot.parentNode, _indexInParent(cRoot) + 1);

                            }
                            
                            if (contName == "ul" || contName == "ol") { // List container?
                                
                                // Migrate list items in the container within the start/end range
                                var lcChild = _isAncestor(cont, start) ? _findAncestor(start, cont) : cont.firstChild, lcEndChild = _isAncestor(cont, end) ? _findAncestor(end, cont) : cont.lastChild;
                                
                                while (lcChild) {
                                
                                    var migrateLI = _nodeName(lcChild) == "li" ? lcChild : null;
                                    lcChild = lcChild == lcEndChild ? null : lcChild.nextSibling;
                                    
                                    // NB: Keeping white space... since undo history might rely on them
                                    if (migrateLI) {

                                        // Migrate the list item into the new list item container
                                        _execOp(_Operation.REMOVE_NODE, migrateLI);
                                        _execOp(_Operation.INSERT_NODE, migrateLI, listContainer);
                                        
                                        // If the list-container is encapsulated by inline elements before the master container,
                                        // then the migrated list items children must also be encapsulated.
                                        // TODO: PHASE THIS OUT - SUPPORTS INVLAID HTML
                                        if (cRoot != cont) {
                                            var liChild = migrateLI.firstChild;
                                            while (liChild) { // For all list items immediate children
                                                if (liChild.nodeType != Node.ELEMENT_NODE && liChild.nodeType != Node.TEXT_NODE) continue;
                                                
                                                // Create a cloned sub-tree of the list containers inline parents
                                                var inode = cont.parentNode, iSubTree = [];
                                                while (inode) {
                                                    iSubTree.push(inode.cloneNode(false));
                                                    inode = inode == cRoot ? null : inode.parentNode;
                                                }
                                                
                                                // Connect up unary inline tree
                                                for (var k = iSubTree.length - 1; k > 0; k--) {
                                                    iSubTree[k].appendChild(iSubTree[k - 1]);
                                                }
                                                
                                                // Insert the cloned inline sub tree in the migrated list item
                                                // such that it encapsulated this list item child
                                                _execOp(_Operation.INSERT_NODE, iSubTree[iSubTree.length - 1], migrateLI, _indexInParent(liChild));
                                                _execOp(_Operation.REMOVE_NODE, liChild);
                                                _execOp(_Operation.INSERT_NODE, liChild, iSubTree[0]);
                                                
                                                liChild = iSubTree[iSubTree.length - 1].nextSibling;
                                            } // End loop: encapsulating list item's contents with inline elements
                                        }
                                        
                                    }
                                } // End loop: migrating list items from an existing list container to the new list container
                                // Check if the container should be removed from the document.. if
                                // it no longer contains any list items
                                var shouldRemoveCont = true;
                                lcChild = cont.firstChild;
                                while (lcChild) {
                                    if (_nodeName(lcChild) == "li") {
                                        shouldRemoveCont = false;
                                        break;
                                    }
                                    lcChild = lcChild.nextSibling;
                                }
                                
                                if (shouldRemoveCont) 
                                    _execOp(_Operation.REMOVE_NODE, cRoot); // Remove by the root
                                
                            } else {
                            
                                // Create/add a new list item
                                var listItem = $createElement("li");
                                _execOp(_Operation.INSERT_NODE, listItem, listContainer);
           
                                // Migrate contents
                                _execOp(_Operation.REMOVE_NODE, cRoot);
                                _execOp(_Operation.INSERT_NODE, cRoot, listItem);
                                
                                // Should this container element be converted into a list item? i.e.
                                // removed from within the list item it was just added to?
                                if (convertContainers[contName]) {
                                    
                                    var parent = cont.parentNode; // Might not be the list container
                                    
                                    // Remove container from the document
                                    _execOp(_Operation.REMOVE_NODE, cont);
                                    
                                    // Migrate children into the containers old parent
                                    while (cont.firstChild) {
                                        var miNode = cont.firstChild;
                                        _execOp(_Operation.REMOVE_NODE, miNode);
                                        _execOp(_Operation.INSERT_NODE, miNode, parent);
                                    }
                                    
                                }
                                
                            }
                            
                        }
                        
                    } // End loop: itemizing containes in range
                    
                }        

                /**
                 * Converts a list item, either migrates its contents up one level or changes the list item type from
                 * numbers to bullets or vice verse.
                 * 
                 * @param {Node} liEle A list item ("li") element.
                 * @param {Boolean} destroy True to migrate contents up one level, false to change type.
                 * 
                 * @return {Node} If destroy is false, it will return the new list container created for changing the list item type.
                 */
                function convertLI(liEle, destroy) {
                    
                    var sib = liEle.previousSibling, 
                        doesHavePreceedingLI = false,
                        doesHaveProceedingLI = false,
                        lcEle = liEle.parentNode,
                        didSplit = false,
                        newLCont = null;
                        
                    // Determine if the list item is preceeded with other list items
                    while(!doesHavePreceedingLI && sib) {
                        doesHavePreceedingLI |= _nodeName(sib) == "li";
                        sib = sib.previousSibling;
                    }
                    
                    // Determine if the list item is proceeded with other list items
                    sib = liEle.nextSibling
                    while(!doesHaveProceedingLI && sib) {
                        doesHaveProceedingLI |= _nodeName(sib) == "li";
                        sib = sib.nextSibling;
                    }
                    
                    if (doesHavePreceedingLI) {
                        lcEle = splitLIContainer(lcEle, liEle, !destroy, true);
                        didSplit = true;
                    }
                    
                    if (destroy) {
                    
                        // If the list item's contents are not going to be migrated within a lower level list item, 
                        // then the range needs to be normalized such that inline groups are place in paragraphs.
                        if (_nodeName(lcEle.parentNode) != "li") 
                            _getNormalizedContainerRange(liEle, liEle);
                        
                        // Migrate the list item's contents before its list container
                        while (liEle.firstChild) {
                            var liCh = liEle.firstChild;
                            _execOp(_Operation.REMOVE_NODE, liCh);
                            _execOp(_Operation.INSERT_NODE, liCh, lcEle.parentNode, _indexInParent(lcEle));
                        }
                        
                        // Remove the list item from its container
                        _execOp(_Operation.REMOVE_NODE, liEle);
                        
                    } else { // change type

                        if (didSplit) {
                            
                            if (doesHaveProceedingLI) {
                                // Get next list item element in split container
                                var nextLI = liEle.nextSibling;
                                while (_nodeName(nextLI) != "li") {
                                    nextLI = nextLI.nextSibling;
                                }
                                
                                // If previously split the container and migrated some remaining li's .. then split
                                // again on the new container - into a container of the original type.
                                splitLIContainer(lcEle, nextLI, true, true);
                            }
                                
                            newLCont = lcEle;

                        } else {
                            
                            // Insert a new container before and migrate the list item
                            newLCont = splitLIContainer(lcEle, liEle, true, false);
                            
                        }
                        
                    }
                    
                    // Check if the list item container has any list items left
                    sib = lcEle.firstChild;
                    while (sib && _nodeName(sib) != "li") {
                        sib = sib.nextSibling;
                    }
                                            
                    // If the li element's container is left without a li, remove it from the document...
                    if (!sib) {
                        
                        // Remove all ancestors of the container which do not have multiple children to keep html tidy
                        var remEle = lcEle;
                        while (remEle.parentNode.childNodes.length == 1 &&
                               !de.doc.isEditSection(remEle.parentNode) &&
                               remEle.parentNode != docBody) {
                            remEle = remEle.parentNode;
                        }
                        
                        _execOp(_Operation.REMOVE_NODE, remEle);
                        
                    }
                    
                    return newLCont;
                    
                } // End inner function: convertLI
                
                /**
                 * Splits a list item container in two. Records operation.
                 * @param {Node} lcEle          A ul or ol element
                 * @param {Node} liEle          An li element to split from within lcEle
                 * @param {Boolean} flipType    True to have the split container to have a different type than lcEle
                 * @param {Boolean} downward    True to split container from liEle downward - so new container is after lcEle.          
                 *                              False to split container from liEle upward - so new container is before lcEle.
                 *                              
                 * @return {Node} The split container.
                 */
                function splitLIContainer(lcEle, liEle, flipType, downward) {

                    // Split container in two
                    var splitC = flipType ? $createElement(_nodeName(lcEle) == "ol" ? "ul" : "ol") : lcEle.cloneNode(false);
                    
                    // Migrate this and all following li's into the split container
                    if (downward) {
                        
                        while (liEle) {
                            var migrateNode = liEle;
                            liEle = liEle.nextSibling;
                            _execOp(_Operation.REMOVE_NODE, migrateNode);
                            _execOp(_Operation.INSERT_NODE, migrateNode, splitC);
                        }
                        
                        _execOp(_Operation.INSERT_NODE, splitC, lcEle.parentNode, _indexInParent(lcEle) + 1);
                        
                    } else { // Upward
                        while (liEle) {
                            var migrateNode = liEle;
                            liEle = liEle.previousSibling;
                            _execOp(_Operation.REMOVE_NODE, migrateNode);
                            _execOp(_Operation.INSERT_NODE, migrateNode, splitC, 0);
                        }
                        
                        _execOp(_Operation.INSERT_NODE, splitC, lcEle.parentNode, _indexInParent(lcEle));
                    }

                    return splitC;
                    
                } // End inner function splitLIContainer
                

                /**
                 * Normalizes the given range and checks if all containers are list-items/list-containers of a certain type
                 * (ol or ul)
                 * 
                 * @param {Node} start    The start of the range to normalize. This range will be deepend
                 * @param {Node} end      The end of the range to normalize. This range will be deepend
                 * @return {Object}           Either Null if the range does not have containers. Or an object with
                 *                             a memeber called "allListEles" which is true iff all
                 *                             containers in the normalized range are list-items/list-containers - and
                 *                             another memeber "nrange" which contains the normalized range.
                 */
                function areAllListElements(start, end) {
                   
                    // Normalize the range
                    var normalizedRange = _getNormalizedContainerRange(start, end);
                    
                    // Anything to itemize?
                    if (normalizedRange.length == 0)
                        return null;
                        
                    // Check if range is all list elements
                    var isAllListEles = true;
                    for (var i in normalizedRange) {
                        var cname = _nodeName(normalizedRange[i]);
                        
                        // Get list item container name
                        if (cname == "li")
                            cname = _nodeName(normalizedRange[i].parentNode);
                        
                        if ((bullets && cname != "ul") || (!bullets && cname != "ol")) {
                            isAllListEles = false;
                            break;
                        }
                    }
                    
                    return {allListEles: isAllListEles, nrange: normalizedRange};
                    
                } // End inner areAllListElements
                
            })(startNode, endNode); // End recursive exec function
            
            this.selAfter = this.selBefore;
            
        } // End execute function
    });
    
})();
