/*
 * file: ModifyTableAction.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */

// @DEPENDS: UndoMan
bootstrap.provides("actions.ModifyTableAction");

(function() {
    
    _registerAction("ModifyTable", {
        
        /**
         * An undoable action. Provides a range of table modification operations
         * 
         * @author Brook Novak
         * 
         * @param {String} op           The operation name (case insensitive). Can be: 
         *                                  insert-rows-before-n,  Where n is the amount
         *                                  insert-rows-after-n,  Where n is the amount
         *                                  delete-rows,
         *                                  insert-cols-before-n,  Where n is the amount
         *                                  insert-cols-after-n,  Where n is the amount
         *                                  delete-cols,
         *                                  merge-cells,
         *                                  split-cells,
         *                                  delete-table                  
         * 
         * @param {Node} startNode      The starting dom node of the range to align.
         *
         * @param {Node} endNode        The ending dom node of the range to align. Can be the same as start node
         * 
         * @throws {Error}              If given operation is unknown/malformed.
         *
         */
        exec : function(op, startNode, endNode) {
            
            var tableNode = null,
                selectedNodes = [];
                
            // Find first occurance of a table (it does not make sense to modify multiple tables)
            _visitAllNodes(_getCommonAncestor(startNode, endNode), startNode, true, function(domNode) {
                
                // Find table
                if (!tableNode)
                   tableNode = _findAncestor(domNode, docBody, function(node){ return (_nodeName(node) == "table"); }, true);
               
                // Record selected nodes
                selectedNodes.push(domNode);
                
                return domNode != endNode; // Traverse only in range
            });
            
            // Is a table selected?
            if (tableNode) {
                
                var tableSelection = {
                    start : null, /* [row-index,col-index,node] */
                    end : null /* [row-index,col-index,node] */
                };
                
                // Get selection within table
                for (var y = 0; y < tableNode.rows.length; y++) { // For each row
                    var row = tableNode.rows[y];
                    for (var x = 0; x < row.cells.length; x++) { // For each cell within row
                        var cell = row.cells[x];
                        
                        // Determine if this cell is selected
                        for (var i in selectedNodes) {
                            if (selectedNodes[i] == cell || _isAncestor(cell, selectedNodes[i])) {
                                
                                // Update table selection info
                                if (!tableSelection.start) 
                                    tableSelection.start = [parseInt(y),parseInt(x), cell];
                                tableSelection.end = [parseInt(y),parseInt(x), cell];
                                break;
                            }
                        }
                    }
                }
                
                if (tableSelection.start) { // Is anything in the table selection / does have content to modify?
                    
                    // Determine table operation
                    op = op.toLowerCase().split('-');
                    
                    switch(op[0]) {
                        case "insert":
                            
                            if (op.length == 4) {
                                
                                // Determine amount of rows / cols to insert
                                var amount = parseInt(op[3]);
                                if (isNaN(amount))
                                    _error(_ErrorCode.BAD_ARGS);
                                
                                // Insert the rows/cols
                                if (op[1] == "rows")
                                    insertRows(tableNode, tableSelection, op[2] == "before", amount);
                                else if (op[1] == "cols") 
                                    insertCols(tableNode, tableSelection, op[2] == "before", amount);
     
                                else _error(_ErrorCode.BAD_ARGS);
                                
                            }
    
                            else _error(_ErrorCode.BAD_ARGS);
                        
                            break;
                            
                        case "delete":
                            
                            if (op.length != 2) 
                                _error(_ErrorCode.BAD_ARGS);
                            
                            switch(op[1]) {
                                case "table":
                                    deleteTable(tableNode);
                                    break;
                                    
                                 case "rows":
                                    deleteRows(tableNode, tableSelection);
                                    break;
                                    
                                 case "cols":
                                    deleteCols(tableNode, tableSelection);
                                    break;
                                    
                                 default:
                                    _error(_ErrorCode.BAD_ARGS);
                            }
                            
                            break;
                        
                        case "merge":
                            break;
                            
                        case "split":
                            break;
                        
                        default:
                            _error(_ErrorCode.BAD_ARGS);
                    }
                    
                    
                    
                    // TODO: UPDATE CURSOR
                
                }
                
                
            }
            

        }
    });
    
    /**
     * Inserts a new table cell containing a placeholder.
     * @param {Node} row      The tr element to insert a row into
     * @param {Number} index  The index of the cell to insert.
     */
    function insertCell(row, index) {
        
        // Insert a new cell
        var newCell = _execOp(_Operation.INSERT_CELL, row, index);
        
        // Add placeholder markup
        de.UndoMan.execute(
            de.UndoMan.ExecFlag.DONT_STORE, 
            "InsertHTML", "<p>" + _getOuterHTML(de.doc.createMNPlaceholder()) + "</p>", 
            newCell);
    }
    
    /**
     * Inserts one or more rows into a table
     * @param {Node} table        The table to insert rows into
     * @param {Object} selection  The table selection
     * @param {Boolean} isBefore  True to insert before selection, false after
     * @param {Number} amount     The amount of rows to add
     */
    function insertRows(table, selection, isBefore, amount) {
        
        // Determine insertion index
        var newRowIndex;
        if (isBefore) 
            newRowIndex = selection.start[0];
        else { // insert after - must consider row-span
            
            newRowIndex = selection.end[0] + 1;
            var rspan = selection.end[2].getAttribute("rowspan");
            if (rspan) newRowIndex += (parseInt(rspan)-1);
        }
        
        // Clamp
        if (newRowIndex > table.rows.length)
            newRowIndex = table.rows.length;
        
        // Determine column count for create cols per new row
        var columnCount = getColumnCount(table);
                                
        for (var i = 0; i < amount; i++) {

            // Create the new row
            var newRow = _execOp(_Operation.INSERT_ROW, table, newRowIndex);
            
            // Insert empty cells to fill out row
            for (var j = 0; j < columnCount; j++) {
                insertCell(newRow, 0);
            }
            
        }
    }
    
    /**
     * 
     * @param {Object} selection    The table selection
     * @param {Boolean} isBefore    True to get column index before selection, false after
     * @return {Number}             The column index before/after the given selection
     */
    function getColumnIndex(selection, isBefore) {
        
         var index;
        
        if (isBefore) {
            index = Math.min(selection.start[1], selection.end[1]);
        }  else { // insert after - must consider col-span
            
            var startIndex = selection.start[1],
                endIndex = selection.end[1];
            
            var cspan = selection.start[2].getAttribute("colspan");
            if (cspan) startIndex += (parseInt(cspan)-1);
            
            cspan = selection.end[2].getAttribute("colspan");
            if (cspan) endIndex += (parseInt(cspan)-1);
            
            index = Math.max(startIndex, endIndex);
        }
        
        return index;
    }
    
    /**
     * @param {Node} table     The table to check
     * @return {Number}        The amount of columns in the given table
     */
    function getColumnCount(table) {
        var columnCount = 0;
        for (var y = 0; y < table.rows.length; y++) { 
            columnCount = Math.max(columnCount, table.rows[y].cells.length);
        }
        return columnCount;
    }
    
    /**
     * Inserts one or more columns into a table
     * @param {Node} table        The table to insert columns into
     * @param {Object} selection  The table selection
     * @param {Boolean} isBefore  True to insert before selection, false after
     * @param {Number} amount     The amount of columns to add
     */
    function insertCols(table, selection, isBefore, amount) {
        
        // Determine column insertion index
        var newColIndex = getColumnIndex(selection, isBefore);
        if (!isBefore)
            newColIndex++; // Insert AFTER
        
        for (var i = 0; i < amount; i++) {
            
            for (var y = 0; y < table.rows.length; y++) { // For each row in table
                var row = table.rows[y];
                
                // Determine insert index ... may not be newColIndex due to colspans
                var spannedIndex = -1,
                    insertIndex = newColIndex;
                    
                for (var x = 0; x < row.cells.length; x++) { // For each cell in current row
                    var cell = row.cells[x];
                    
                    spannedIndex ++;
                    
                    // Add column span if has span length more than 1
                    var cspan = cell.getAttribute("colspan");
                    if (cspan) {
                        spannedIndex += (parseInt(cspan) - 1);
                    }
                    
                    // Break if found insert index
                    if (spannedIndex >= newColIndex) {
                        insertIndex = parseInt(x);
                        break;
                    }
                } // End loop: each cell
                
                // Insert the new cell which makes up the new column in this row
                insertCell(row, insertIndex);
                
            } // End loop: each row
        
        } // End loop: insertion amount
        
    }
    
    /**
     * Deletes an entire table
     * @param {Node} table The table element to delete
     */
    function deleteTable(table) {
        _execOp(_Operation.REMOVE_NODE, table);
    }
    
    /**
     * Deletes all selected rows in table. 
     * Deletes entire table if all rows selected.
     * 
     * @param {Node} table        The table element to delete rows from
     * @param {Object} selection  The table selection
     */
    function deleteRows(table, selection) {
        
        // Is all table selected?
        if (selection.start[0] == 0 && selection.end[0] >= (table.rows.length-1)) 
            deleteTable(table);
            
        else {
            for (var i = 0; i <= (selection.end[0] - selection.start[0]); i++) {
                _execOp(_Operation.DELETE_ROW, table, selection.start[0]);
            }
        }
    }
    
    /**
     * Deletes all selected columns in table. 
     * Deletes entire table if all columns selected.
     * 
     * @param {Node} table        The table element to delete columns from
     * @param {Object} selection  The table selection
     */
    function deleteCols(table, selection) {
        
        var colStartIndex = getColumnIndex(selection, true),
            colEndIndex = getColumnIndex(selection, false),
            columnCount = getColumnCount(table);
  
        // Check if need to delete entire table
        if (colStartIndex == 0 && colEndIndex >= (columnCount-1))
            deleteTable(table);
            
        else {
            for (var i = 0; i <= (colEndIndex - colStartIndex); i++) {
                
                for (var y = 0; y < table.rows.length; y++) { // For each row in table
                    var row = table.rows[y];
                    
                    // Determine deletion index ... may not be colStartIndex due to colspans
                    var spannedIndex = -1,
                        deleteIndex = colStartIndex;
                        
                    for (var x = 0; x < row.cells.length; x++) { // For each cell in current row
                        var cell = row.cells[x];
                        
                        spannedIndex ++;
                        
                        // Add column span if has span length more than 1
                        var cspan = cell.getAttribute("colspan");
                        if (cspan) {
                            spannedIndex += (parseInt(cspan) - 1);
                        }
                        
                        // Break if found insert index
                        if (spannedIndex >= colStartIndex) {
                            deleteIndex = parseInt(x);
                            break;
                        }
                    } // End loop: each cell
                    
                    // Delete the cell
                    _execOp(_Operation.DELETE_CELL, row, deleteIndex);
                    
                } // End loop: each row               
                
            } // End loop: deleting columns
        }
    }
    
    
    
    
    function mergeCells(table, selection) {
        
        /*
        var mergeWidth = (selection.end[1] - selection.start[1]) + 1,
            mergeHeight = (selection.end[0] - selection.start[0]) + 1,
            mergeTarget = ?;
        
        for (var y = 0; y < mergeHeight; y++) { // Merge all adjacent cells in y-axis (colums)
            
            // Merge this row into merge target
            for (var x = 1; x < mergeWidth; x++) { // Merge row into target
                
                _execOp(_Operation.DELETE_CELL, table, selection.start[0]);
                
                x += colspanextra;
                
            }
        }*/
        
        
    }
    
    
    function splitCells(selectedNodes) {
        
        for (var i in selectedNodes) {
            var node = selectedNodes[i];
            
            if (_nodeName(node) == "td" || _nodeName(node) == "th") { // Is cell?
            
               // Does this selected cell have a column span?
               var span = node.getAttribute("colspan");
               if (span)
                   _execOp(_Operation.SET_ATTRIBUTE, node, "colspan", ""); // Erase span

               var span = node.getAttribute("rowspan");
               if (span) 
                   _execOp(_Operation.SET_ATTRIBUTE, node, "rowspan", ""); // Erase span
            }
            
        }
    }
    
        
})();
