/*
 * file: RemoveDOMAction.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */
// @DEPENDS: UndoMan
bootstrap.provides("actions.RemoveDOMAction");

_registerAction("RemoveDOM",{


    /**
     * @class
     * An undoable dom action. Removes a given range from the document.
     * 
     * @author Brook Novak
     * 
     * @param {Node} startNode      The starting dom node of the fragments range.
     *
     * @param {Number} startIndex 	The inclusive start index in the start node.
     * 								Ranges from 0 to the text length for text nodes.
     *                              Where 0 indicates that the range begins at the first char, and text length
     *                              indicates that the range begins directly after the text node, but not including it.
     *                              <br>
     *                              Ranges from 0 to 1 for elements.
     *                              Where 0 indicates that the range includes the element and it's decendants,
     *                              and 1 indicates that the range excludes the element and it's decendants.
     *                              
     *
     * @param {Node} endNode        The ending dom node of the fragments range.
     *
     * @param {Number} endIndex 	The inclusive end index in the end node.
     * 								Ranges from 0 to the text length for text nodes.
     *                              Where 0 indicates that the range ends just before the text node, but not including it, 
     *                              and text length indicates that the range ends at the last charactor in the text run.
     *                              <br>
     *                              Ranges from 0 to 1 for elements.
     *                              Where 0 indicates that the range excludes the element and it's decendants,
     *                              and 1 indicates that the range includes the element and it's decendants.
     * 
     */        
    exec : function(startNode, startIndex, endNode, endIndex) {

        var fragmentRoot = _buildFragment(_getCommonAncestor(startNode, endNode, false), 
                startNode, startIndex, endNode, endIndex);

        // Convert all whitespaces to non-breaking spaces
        _convertWSToNBSP(fragmentRoot.node);
        
        // Collapse the range
        var fMigrantNode = fragmentRoot.collapse();
        
        // Normalize any NBSP entities
        _normalizeNBSP(fragmentRoot.node);
        
        if (this.flags & de.UndoMan.ExecFlag.UPDATE_SELECTION) {
            
            var newCursorPos;
        
            // If there was a migrant, get the cursor descriptor from this
            if (fMigrantNode) 
                newCursorPos = de.cursor.getNearestCursorDesc(fMigrantNode, 0, false, true);
            else {
                // Get the deepest node on the starting bounds that still resides in the document
                var fragment = fragmentRoot.getStartFragment();
                while (fragment.node != docBody && !_isAncestor(docBody, fragment.node)) {
                    fragment = fragment.parent;
                }
                var domNode = fragment.node;
                
                // Check to see if there are any nodes in the start-bounds place
                if (fragment.children.length > 0 && domNode.childNodes.length > 0 &&
                    fragment.children[0].pos <= domNode.childNodes.length) {
                
                    if (fragment.children[0].pos == 0) {
                        newCursorPos = de.cursor.getNearestCursorDesc(domNode.firstChild, 0, false, false);
                    } else {
                        var cNode = domNode.childNodes[fragment.children[0].pos - 1];
                        newCursorPos = de.cursor.getNearestCursorDesc(cNode, _nodeLength(cNode, 2) - 1, true, false);
                    }
                } else  newCursorPos = de.cursor.getNearestCursorDesc(fragment.node, _nodeLength(fragment.node, 2) - 1, true, true);
            }
            
            if (newCursorPos)
                this.selAfter = { 
                    startNode : newCursorPos.domNode, 
                    startIndex : newCursorPos.relIndex + (newCursorPos.domNode.nodeType == Node.TEXT_NODE && newCursorPos.isRightOf ? 1 : 0)
                };
        }
    }
    
});


_registerAction("RemoveNode",{
    /**
     * A simple undoable action. Removes a dom node from the document.
     * Selection will always be cleared afterwards.
     * 
     * @param {Node} node The node to remove
     * 
     */
    exec: function(node) {
        _execOp(_Operation.REMOVE_NODE, node);
    }
});
