/*
 * file: SplitContainerAction.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */

// @DEPENDS: UndoMan
bootstrap.provides("actions.SplitContainerAction");

(function() {
    
    var duplicateBlockMap = $createLookupMap("p,pre,h1,h2,h3,h4,h5,h6,li,address");
    
    _registerAction("SplitContainer", {
        /**
         * An undoable action. Splits a block level element's or body node's contents into two.
         * Used for creating new block level elements like paragraphs, headings and list items.
         * 
         * @author Brook Novak
         * 
         * @param {Node} domNode	The text or element node to split the (splitable) container owner into two.
         *                             
         * 
         * @param {Number} index 	The index within/next to the dom node to split from, all the remaining contents INCLUDING the index
         *                          will migrate to the new block-level element.
         * 							For text nodes, this ranges from zero to the length of the text. If at the length of the text, then
         *                          the split will begin after the text node.
         * 							For other nodes, this can be zero (split just before the node) or 1 (split after the node).
         * 
         */
        exec : function(domNode, index) {
        	debug.assert(index >= 0);
        	debug.assert(
        		(domNode.nodeType == Node.TEXT_NODE && index <= domNode.nodeValue.length) || 
        		(domNode.nodeType == Node.ELEMENT_NODE && index <= 1)
        	);            
            
            // Get the range from the starting point (exclusive) to the end of the splitter
            // Get the first ancestor block level element which to split at
            var splitter = _findAncestor(domNode.parentNode, docBody, _isBlockLevel, true) || docBody,
                endRangeNode, // Will be set to a block / doc body
                endNodeIndex;


            if (!duplicateBlockMap[_nodeName(splitter)]) {
            
                // In the case where new paragraphs are created at the split point due to the splitter not
                // being a container element which should be duplicated, the nodes to migrate may not be all
                // nodes to the right of the split point within the splitter... instead it should range from the split point
                // to (and excluding) the first occurance of a block level element within the splitter container
                
                // Find the first occuring block level element from the split point onwards.. within the
                // splitter container... if any...
                _visitAllNodes(splitter, domNode, true, function(node){
                    if (_isBlockLevel(node)) endRangeNode = node;
                    return endRangeNode == null;
                });
                
                // If contains no block level elements from the split point, then set as end of splitter
                if (!endRangeNode || endRangeNode == splitter) {
                    endRangeNode = splitter;
                    endNodeIndex = 1;
                } else if (endRangeNode == domNode) 
                    // Definitly nothing to migrate
                    endRangeNode = null;
                else 
                     // Make sure the range is exclusive of the block level element
                    endNodeIndex = 0;
                
            } else {
                endRangeNode = splitter;
                endNodeIndex = 1;
            }

            var migrantFragment;
            
            if (endRangeNode) {

                // Build a migration fragment
                migrantFragment = _buildFragment(
                    splitter, 
                    domNode, 
                    index, 
                    endRangeNode, 
                    endNodeIndex);
                    
                // Check to see if anything will be removed
                var isAllShared = true;
                migrantFragment.visit(function(f) {
                    if(!f.isShared) isAllShared = false;
                    return isAllShared;
                });
                
                if (isAllShared) 
                    // If nothing will actually be migrated, then get rid of the fragment
                    migrantFragment = null;
                else migrantFragment.disconnect(); // Remove the second half from the document
            }
    
            // Create a new container
            var insertAfterNode,
                container,
                subSplitter;
                
            if (duplicateBlockMap[_nodeName(splitter)]) {
                
                // Special case with list items: If the splitter's first block level ancestor is a list item,
                // then insert of splitting within the list item, split outside to a new list item
                if (_nodeName(splitter) != "li" && splitter != docBody) {
                    var superSplitter = _findAncestor(splitter.parentNode, docBody, _isBlockLevel, true);
                    if (superSplitter && _nodeName(superSplitter) == "li")  {
                        subSplitter = splitter;
                        splitter = superSplitter;
                    }
                }

                insertAfterNode = splitter;
                container = (splitter.nodeName.charAt(0).toLowerCase() == "h" && !migrantFragment) ?
                    $createElement("p") : /* If splitting at the end of a heading, split into new empty paragraph. */
                    splitter.cloneNode(false);
    
            // Default to paragraph
            } else container = $createElement("p");
         
            // Discover where to insert the new container (If haven't already)
            if (!insertAfterNode) {
                
                if (migrantFragment) { // Insert the container before the disconnected migrants
                    var startFrag = migrantFragment.getStartFragment();
                    
                    // Was the start fragment split into two?
                    if (migrantFragment.wasStartSplit()) {
                        // Insert after the first half of the divided text node
                        insertAfterNode = startFrag.getPreSplitNode();
                        
                    } else if (startFrag.isShared) { // Still remains in the document?
                        // Insert after this then..
                        insertAfterNode = startFrag.node;
                        
                    } else { // Removed from the document?
                        
                        // Get the first shared ancestor
                        var fsa = startFrag;
                        while (!fsa.isShared) {
                            fsa = fsa.parent;
                        }
                        
                        // Does the first shared ancestor have any child nodes in the document?
                        if (fsa.node.firstChild) {
                        
                            // Does the shared ancestor have remvoed child fragments?
                            if (fsa.children.length > 0) {
                            
                                // Was the first removed child fragment removed after the first child in its parent?
                                if (fsa.children[0].pos > 0) {
                                    // If so, insert after fsa's remaining child before the start bounds
                                    insertAfterNode = fsa.node.childNodes[fsa.children[0].pos - 1];
                                } else {
                                    // If not (first frag was the first child in its dom parent), insert as first child
                                    // in the shared ancestor
                                    _execOp(_Operation.INSERT_NODE, container, fsa.node, 0);
                                    insertAfterNode = null;
                                }
                                
                            } else 
                                insertAfterNode = fsa.node.lastChild;
                            
                            
                        } else if (fsa == migrantFragment) { // Is fsa the splitter?
                            // If the start bounds where fully removed, insert container as first child of the splitter
                            _execOp(_Operation.INSERT_NODE, container, splitter, 0);
                            insertAfterNode = null;

                        } else 
                            insertAfterNode = fsa.node; // if not insert after the first shared ancestor then
                    }

                } else { // Insert the container before or after the split point
                    if (index == 0) { // Before split point
                        _execOp(_Operation.INSERT_NODE, container, domNode.parentNode, _indexInParent(domNode));
                        insertAfterNode = null;
                    } else insertAfterNode = domNode; // After split point
                }

               // Now that the insertion point is discovered from the fragments start bounds, make sure the
                // insertAfterNode's parent is block level
                if (insertAfterNode && insertAfterNode != docBody && !_isBlockLevel(insertAfterNode)) {
                    while (insertAfterNode != docBody && 
                        insertAfterNode.parentNode != docBody &&
                        !_isBlockLevel(insertAfterNode) && 
                        !_isBlockLevel(insertAfterNode.parentNode)) {
                        insertAfterNode = insertAfterNode.parentNode;
                    }
                }
                    
            }
            
            // Insert the container into the document (if haven't already done so)
            if (insertAfterNode)
                _execOp(_Operation.INSERT_NODE, container, insertAfterNode.parentNode, _indexInParent(insertAfterNode) + 1);
                
           // Migrate all of the disconnected nodes in the right-side of the split to the new container
           if (migrantFragment) {
                for (var i in migrantFragment.children) {
                    var migs = buildMigrants(migrantFragment.children[i]);
                    if (migs) 
                        _execOp(_Operation.INSERT_NODE, migs, container); // TODO: Check if migrants can validly migrate to the container?
                     // _isValidRelationship should actually be ok for Descendants/ancestors... not just immediate parents/children.
                }  
           }
    
            // Add any placeholders in the containers if need be
            if (_doesNeedMNPlaceholder(container))  // Important to check container first, it could be a child of the splitter
                _execOp(_Operation.INSERT_NODE, de.doc.createMNPlaceholder(), container);
            
            if (_doesNeedMNPlaceholder(splitter)) 
                _execOp(_Operation.INSERT_NODE, de.doc.createMNPlaceholder(), splitter);
            
            if (subSplitter && _doesNeedMNPlaceholder(subSplitter))
                _execOp(_Operation.INSERT_NODE, de.doc.createMNPlaceholder(), subSplitter);

            // Discover the new cursor position
            if (this.flags & de.UndoMan.ExecFlag.UPDATE_SELECTION) {
                var cDesc = de.cursor.getNearestCursorDesc(container, 0, false, false);
                if (cDesc)
                    this.selAfter = {startNode : cDesc.domNode, startIndex : cDesc.relIndex + (cDesc.domNode.nodeType == Node.TEXT_NODE && cDesc.isRightOf ? 1 : 0)};
            }
                
            /**
             * Inner helper function
             * @param {de.dom.DOMFragment} frag
             */
            function buildMigrants(frag) {
                
                var migrantNode;
                
                if (frag.isShared) {
                    
                    // Check if all descendant fragments from this shared fragment are all shared
                    var isAllShared = true;
                    frag.visit(function(f) {
                        if(!f.isShared) isAllShared = false;
                        return isAllShared;
                    })
                    
                    // Ignore this subtree if they are all shared
                    if (isAllShared) return null;
     
                     // Clone the shared node.. since it still remains in the document...
                    migrantNode = frag.node.cloneNode(false);
                    
                } else { // disconnected
                
                    migrantNode = frag.node;
                
                    // Remove from parent and remove its children if it has any
                    if (migrantNode.parentNode && migrantNode.parentNode.nodeType != Node.DOCUMENT_FRAGMENT_NODE) 
                        _execOp(_Operation.REMOVE_NODE, migrantNode);

                    while(migrantNode.firstChild) {
                        _execOp(_Operation.REMOVE_NODE, migrantNode.firstChild);
                    }
                }
                
                // Recurse... build all children and link with the current dom node
                for (var i in frag.children) {
                    var descendants = buildMigrants(frag.children[i]);
                    if (descendants) 
                        _execOp(_Operation.INSERT_NODE, descendants, migrantNode);
                }
                
                return migrantNode;
                
            } // End inner buildMigrants            
        }
    });
    
})();
