/*
 * file: DoublyLinkedList.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */
bootstrap.provides("collections.DoublyLinkedList");

/**
 * @class
 *
 * A Doubly Linked List. Supports all data types.
 * Because this collection creates cyclic references, you should explicitely
 * call de.collections.DoublyLinkedList.clear when finished with a doubly linked
 * list to avoid memory leaks in browsers which uses reference counting garbage collections
 * (e.g. IE and FF).
 *
 * @author Brook Novak
 */
var _DoublyLinkedList = function(){

    var cls = function(){
    
        /**
         * @memberOf de.collections.DoublyLinkedList
         * Read only member that is the current length of the linked list.
         * @type Number
         */
        this.length = 0;
        
        /**
         * @memberOf de.collections.DoublyLinkedList
         * Read only member that is the current head node of the linked list.
         * @type de.collections.DLLNode
         */
        this.head = null;
        
        /**
         * @memberOf de.collections.DoublyLinkedList
         * Read only member that is the current tail node of the linked list.
         * @type de.collections.DLLNode
         */
        this.tail = null;
    };
    
    // Define the class body
    cls.prototype = {
    
        /**
         * Adds data to the end of the linked list
         *YYY
         * @param {Object} data Data to add
         */
        add: function(data){
        
            //create a new item object, place data in
            var node = {
                data: data,
                next: null,
                prev: null
            };
            
            //special case: no items in the list yet
            if (this.length == 0) {
                this.head = node;
                this.tail = node;
            } else {
            
                //attach to the tail node
                this.tail.next = node;
                node.prev = this.tail;
                this.tail = node;
            }
            
            //don't forget to update the count
            this.length++;
            
        },
        
        /**
         * Removes an item from the list.
         * @param {Object} item An item to remove.
         * @return {Boolean} True if item existed and was removed. False if the item did not exist.
         */
        remove: function(item){
            var current = this.head;
            while (current) {
                if (current.data == item) {
                    removeNode(this, current);
                    return true;
                }
                current = current.next;
            }
            return false;
        },
        
        /**
         * Removes a node at the given index.
         *
         * @param {Number} index The index to remove at.
         *
         * @return {Object} The removed data. Otherwise null if index out of bounds. Note that
         * 			if data is null, null will also be returned.
         */
        removeAtIndex: function(index){
        
            //check for out-of-bounds values
            if (index > -1 && index < this.length) {
            
                var current;
                
                // Choose faster scan direction
                if (index < (this.length / 2)) { // Head to tail
                    current = this.head;
                    for (var i = 0; i < index; i++) {
                        current = current.next;
                    }
                } else { // Tail to head
                    current = this.tail;
                    for (var i = this.length - 1; i > index; i--) {
                        current = current.prev;
                    }
                }
                
                removeNode(this, current);
                
                return current.data;
            }
            
            return null;
        },
        
        /**
         * Removes the last node in the list and returns the item. O(1) operation
         * @return {Object} The last item in the list.
         */
        pop: function(){
            return this.removeAtIndex(this.length - 1);
        },
        
        /**
         * Removes proceding items after atNode. I.E. atNode becomes the new tail.
         *
         * @param {Object} atNode The which will become the new tail.
         *
         * @return {Boolean} True if the operation succeeded (atNode is a node in this linked list),
         * 			False if the operation failed (atNode is not a node in this linked list).
         */
        chop: function(atNode){
        
            var current = this.tail,
                newLength = this.length;
            
            // Search for atNode... count how far it is from the tail
            while (current && current != atNode) {
                current = current.prev;
                newLength--;
            }
            
            // If atNode doesn't exist, return false
            if (!current)                     
                return false;
            
            // If atNode has a next-node (i.e. not the tail) then
            // chop off the proceeding nodes
            if (current.next) {
                
                var removedNodes = current.next;
                while (removedNodes) { // Kill all ref n removed nodes
                    removedNodes.prev.next = null;
                    removedNodes.prev = null;
                    removedNodes = removedNodes.next;
                }
                
                // Set the new tail
                this.tail = current;
                
                // Be sure to update the length
                this.length = newLength;
            }
            
            return true;
            
        },
        
        /**
         * Clears the linked list.. so list becomes empty.
         * Call this whenever you are finished with a Doubly linked list to avoid memory
         * leaks caused by cyclic references.
         */
        clear: function(){
            while (this.head) {
                var node = this.head;
                this.head = node.next;
                node.prev = node.next = null;
            }
            this.tail = null;
            this.length = 0;
        },
        
        /**
         * Applies a function to all items in this list from the head to the tail.
         *
         * @param {Function} func 	A function to apply to each item. Takes one argument: the item.
         * 							Return false to abort iteration.
         */
        iterate: function(func){
            var current = this.head;
            while (current) {
                var res = func(current.data);
                if (res === false) break;
                current = current.next;
            }
        }
        
    }; // End clas prototype
    
    /**
     * @function
     * @description
     * An alias for de.collections.DoublyLinkedList#add
     *
     * @see de.collections.DoublyLinkedList#add
     */
    cls.prototype.push = cls.prototype.add;
    
    
    /**
     * @private
     * Removes a node from a doubly linked list
     *
     * @param {de.collections.DoublyLinkedList} dll The DLL to remove from
     *
     * @param {de.collections.DLLNode} node The node to remove - must be a node in the DLL.
     */
    function removeNode(dll, node){
    
        if (dll.length == 1) {
            dll.clear();
            return;
        }
        
        if (node == dll.head) {
            dll.head = node.next;
            dll.head.prev = null;
            node.next = null; // Avoid cyclic reference
        } else if (node == dll.tail) {
            dll.tail = node.prev;
            dll.tail.next = null;
            node.prev = null; // Avoid cyclic reference
        } else {
            node.prev.next = node.next;
            node.next.prev = node.prev;
            node.next = node.prev = null; // Avoid cyclic reference
        }
        
        dll.length--;
    }
    
    return cls;
}();

/**
 * Exposure of _DoublyLinkedList internal
 * @see _DoublyLinkedList
 */
de.collections.DoublyLinkedList = _DoublyLinkedList;
