/*
 * file: Events.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */
bootstrap.provides("events.Events");

/**
 * Adds an event handler to a DOM Event Source.
 * 
 * If you are planning to listen for keyboard strokes, you can use "keystroke." This uses keydown and/or keypress
 * and provides an extra argument to your handler containing the normalized key identifier string.
 * 
 * @param {Object} eventSource The Element or Window to add the event to.
 * 
 * @param {String} eventName The name of the dom event. Must be lowercase and omit the "on" prefix. For example "load" or "keypress".
 * 
 * @param {Function} handler A function to handle the event being registered. When called, the dom event will
 *                           be passed as the first (and only) argument.
 */
function _addHandler(sourceEle, eventName, handlerFunc) {
    
	debug.assert(sourceEle.nodeType != Node.TEXT_NODE && sourceEle.nodeType != Node.COMMENT_NODE);
    debug.assert(eventName.indexOf('on') !== 0);
    
	// IE has trouble passing the window object around - it can be cloned for no reason
	if (sourceEle.setInterval && sourceEle != window)
        sourceEle = window;
            
    switch(eventName) {
        
        case "keystroke":
            // Keystroke is a combo of keydown and key press - normalizes 
            // key identifiers and passes them as extra arg to handler
            addEventHandler(sourceEle, "keydown", function(e) {
                var keyIndent = de.events.Keyboard.getKeyIdentifier(e, true);
                if (keyIndent) 
                    return handlerFunc(e, keyIndent);
            });
            addEventHandler(sourceEle, "keypress", function(e) {
                var keyIndent = de.events.Keyboard.getKeyIdentifier(e, false);
                if (keyIndent) 
                    return handlerFunc(e, keyIndent);
            });
            break;

        case "mousedown": // Keep mouse state updated
            addEventHandler(sourceEle, eventName, function(e) {
                return de.events.Mouse.sniffMouseDownEvent(e) ? handlerFunc(e) : true;
            });
            break;
        case "mouseup": // Keep mouse state updated
            addEventHandler(sourceEle, eventName, function(e) {
                return de.events.Mouse.sniffMouseUpEvent(e) ? handlerFunc(e) : true;
            });
            break;
        case "mousemove": // Keep mouse state updated
            addEventHandler(sourceEle, eventName, function(e) {
                return de.events.Mouse.sniffMouseMoveEvent(e) ? handlerFunc(e) : true;
            });
            break;
            
        default:
            addEventHandler(sourceEle, eventName, handlerFunc);
    }
        
    function addEventHandler(sourceEle, eventName, actualHandler) {
        
        // Keep track of wrapper handlers for removal
        
        // Create event wrapper map (Event type -> tuple[event-source, wrapper handler])
        if (!handlerFunc.evWrappers)
            handlerFunc.evWrappers = {};
        
        // Ensure the map contains an entry for this event type
        if (!handlerFunc.evWrappers[eventName])
            handlerFunc.evWrappers[eventName] = [];
            
        // Store this event tuple
        handlerFunc.evWrappers[eventName].push([sourceEle, forwardEvent]);  
        
        if (sourceEle.addEventListener)
			sourceEle.addEventListener(eventName, forwardEvent, false); // DOM Compliant
			
		else if (sourceEle.attachEvent)
			sourceEle.attachEvent("on" + eventName, forwardEvent); // IE
		
        // @DEBUG ON
        else debug.assert(false, "Unsupported browser: does not support addEventListener or attachEvent");
        // @DEBUG OFF	
         
        /**
         * Wraps an event handler so that the event handler gets a non null event
         * and if the handler returns false the event is consumed.
         * @param {Event} e provided by native event dispatcher
         */
        function forwardEvent(e){
    	    de.events.current = e = e || window.event;
            try {
                if (actualHandler(e) === false)                     
                    return de.events.consume(e);
            } finally {
                de.events.current = 0;
            }
        }
        
    }

}

/**
 * Removes an event handler from a DOM Event Source.
 * 
 * @param {Object} eventSource The Element or Window to add the event to.
 * 
 * @param {String} eventName The name of the dom event. Must be lowercase and omit the "on" prefix. For example "load" or "keypress".
 * 
 * @param {Function} handler The handler function to remove form the given event source.
 */
function _removeHandler(sourceEle, eventName, handlerFunc){
    debug.assert(eventName.indexOf('on') !== 0);
    
    // Keystoke is comprized of keydown and key press
    if (eventName == "keystroke") {
        _removeHandler(sourceEle, 'keydown', handlerFunc);
        _removeHandler(sourceEle, 'keypress', handlerFunc);
        return;
    }
    
    // Check if the handler has been wrapped for this event type / source
    if (handlerFunc.evWrappers && handlerFunc.evWrappers[eventName]) {
    
        // Look for wrapper handlers for this event type for this handler
        for (var i in handlerFunc.evWrappers[eventName]) {
            
            var tuple = handlerFunc.evWrappers[eventName][i];
            
            // Was there a wrapped handler create for this handler on this event type and source?
            if (tuple[0] == sourceEle) { 

                // Remove tuple
                if (handlerFunc.evWrappers[eventName].length == 1)
                    delete handlerFunc.evWrappers[eventName];
                else 
                    handlerFunc.evWrappers[eventName].splice(i, 1);
                    
                // Assign handlerFunc to be the wrapped handler
                handlerFunc = tuple[1];
               
                break;
            }
            
        }
    
    } 

    // Remove the handler from the event source
    if (sourceEle.removeEventListener)
		sourceEle.removeEventListener(eventName, handlerFunc, false); // DOM Compliant
		
	else if (sourceEle.detachEvent)
		sourceEle.detachEvent("on" + eventName, handlerFunc); // IE
	
    // @DEBUG ON
    else debug.assert(false, "Unsupported browser: does not support addEventListener or attachEvent");
    // @DEBUG OFF	
    
}

 
$extend(de.events, {
    
    /**
     * A pointer to the current event
     */
    current : 0,
    
    addHandler : _addHandler,
    
    removeHandler : _removeHandler,

	/**
	 * @param {Event} e An event
	 * 
	 * @return {Object} The x and y coordinates of the event relative to the window.
	 */
	getXYInWindowFromEvent : function(e){
	  
	  var targetX = 0, targetY = 0;
	  
	  if (e.clientX || e.clientX === 0) { 
            targetX = e.clientX;
            targetY = e.clientY;
	  } else if (e.pageX != null) {
	  	  var scrollPos = _getDocumentScrollPos();
	      targetX = e.pageX - scrollPos.left;
	      targetY = e.pageY - scrollPos.top;
      }
	  
	  return {x : targetX, y : targetY };	
    },
    
    /**
     * 
     * @param {Event} e A dom event
     * @return {Node} An element or text node which the event was targetted at.
     */
    getEventTarget: function(e) {
        return e.target || e.srcElement || document;
    },
    
	/**
	 * Consumes an event... stop is from propagating and prevents the default action.
	 * 
	 * @example
	 * myhandlers = function(e) {
	 * 		...
	 * 		if (somecondition) return de.events.consume(e);
	 * 		return true;
	 * }
	 * 
	 * @param {Event} e A dom event.
	 * 
	 * @return {Boolean} False always.
	 */
	consume : function(e) {
        de.events.stopPropogation(e);
		de.events.preventDefault(e);
		return false;
	},
    
    /**
     * Stops the event from propogating / bubbling.
     * @param {Event} e A dom event.
     */
    stopPropogation : function(e) {
        if (_engine == _Platform.TRIDENT)
            e.cancelBubble = true;
        else if (e.stopPropagation)
            e.stopPropagation();
    },
    
    /**
     * Prevents the default behaviour from occuring on a given event.
     * @param {Event} e An event object
     */
    preventDefault : function(e) {
        if (_engine == _Platform.TRIDENT)
            e.returnValue = false;
        else if(e.preventDefault) 
            e.preventDefault();
    }
    
});

