/*
 * file: Keyboard.js
 *
 * @BEGINLICENSE
 * Copyright 2010 Brook Novak  (email : brooknovak@seaweed-editor.com) 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * @ENDLICENSE
 */
/*
 * @see http://unixpapa.com/js/key.html
 * @see http://www.quirksmode.org/js/keys.html
 */

/* @DEPENDS: Platform */

bootstrap.provides("events.Keyboard");

(function() {

	// All the keymaps
    var keymaps = {
    
        // maps the charcodes of special printable keys to key identifiers 
        specialToCharCode: {
            8: "Backspace", // The Backspace (Back) key.
            9: "Tab", // The Horizontal Tabulation (Tab) key.
            //   Note: This key identifier is also used for the
            //   Return (Macintosh numpad) key.
            13: "Enter", // The Enter key.
            27: "Escape", // The Escape (Esc) key.
            32: "Space" // The Space (Spacebar) key.
        },
        
        // maps the keycodes of non printable keys to key identifiers
        keyCodeToId: {
            16: "Shift", // The Shift key.
            17: "Control", // The Control (Ctrl) key.
            18: "Alt", // The Alt (Menu) key.
            20: "CapsLock", // The CapsLock key
            224: "Meta", // The Meta key. (Apple Meta and Windows key)
            37: "Left", // The Left Arrow key.
            38: "Up", // The Up Arrow key.
            39: "Right", // The Right Arrow key.
            40: "Down", // The Down Arrow key.
            33: "PageUp", // The Page Up key.
            34: "PageDown", // The Page Down (Next) key.
            35: "End", // The End key.
            36: "Home", // The Home key.
            45: "Insert", // The Insert (Ins) key. (Does not fire in Opera/Win)
            46: "Delete", // The Delete (Del) Key.
            112: "F1", // The F1 key.
            113: "F2", // The F2 key.
            114: "F3", // The F3 key.
            115: "F4", // The F4 key.
            116: "F5", // The F5 key.
            117: "F6", // The F6 key.
            118: "F7", // The F7 key.
            119: "F8", // The F8 key.
            120: "F9", // The F9 key.
            121: "F10", // The F10 key.
            122: "F11", // The F11 key.
            123: "F12", // The F12 key.
            144: "NumLock", // The Num Lock key.
            44: "PrintScreen", // The Print Screen (PrintScrn, SnapShot) key.
            145: "Scroll", // The scroll lock key
            19: "Pause", // The pause/break key
            91: "Win", // The Windows Logo key
            93: "Apps" // The Application key (Windows Context Menu)
        },
        
        // maps the keycodes of the numpad keys to the right charcodes
        numpadToCharCode: {
            96: "0".charCodeAt(0),
            97: "1".charCodeAt(0),
            98: "2".charCodeAt(0),
            99: "3".charCodeAt(0),
            100: "4".charCodeAt(0),
            101: "5".charCodeAt(0),
            102: "6".charCodeAt(0),
            103: "7".charCodeAt(0),
            104: "8".charCodeAt(0),
            105: "9".charCodeAt(0),
            106: "*".charCodeAt(0),
            107: "+".charCodeAt(0),
            109: "-".charCodeAt(0),
            110: ".".charCodeAt(0),
            111: "/".charCodeAt(0)
        },
        
        // Helpers
        charCodeA: "A".charCodeAt(0),
        charCodeZ: "Z".charCodeAt(0),
        charCodea: "a".charCodeAt(0),
        charCodez: "z".charCodeAt(0),
        charCode0: "0".charCodeAt(0),
        charCode9: "9".charCodeAt(0),
    
		// Platform dependant maps
		keyCodeFix : {},
		charCodeToKeyCode : {},
        
        // Maps keycodes that cannot be distinguished in key press events, to the other keycodes that have also
        // been assigned to the same key in the key press event. ALl key codes (keys/values) in the map will be
        // simulated as key presses in the key down event.
        ambiguousKeyPressCodes : {} 
    
    };

	// Construct inverse maps
	keymaps.idToKeyCode = {};
	for (var key in keymaps.keyCodeToId) {
		keymaps.idToKeyCode[keymaps.keyCodeToId[key]] = parseInt(key, 10);
	}
	for (var key in keymaps.specialToCharCode) {
		keymaps.idToKeyCode[keymaps.specialToCharCode[key]] = parseInt(key, 10);
	}

    // Setup platform dependant key maps
    switch (_engine) {
        
        case _Platform.TRIDENT: // MSHTML
            keymaps.charCodeToKeyCode = {
                13: 13,
                27: 27
            };
            break;
        
        case _Platform.GECKO:
            keymaps.keyCodeFix = {
                12: idToKeyCode("NumLock")
            };
            break;
            
        case _Platform.WEBKIT:

            // starting with Safari 3.1 (version 525.13) Apple switched the key
            // handling to match the IE behaviour.
            if (_Platform.engineVersion && _Platform.engineVersion < 525.13) { // TODO: Check if safari?
                keymaps.charCodeToKeyCode = {
                
                    // Safari/Webkit Mappings
                    63289: idToKeyCode("NumLock"),
                    63276: idToKeyCode("PageUp"),
                    63277: idToKeyCode("PageDown"),
                    63275: idToKeyCode("End"),
                    63273: idToKeyCode("Home"),
                    63234: idToKeyCode("Left"),
                    63232: idToKeyCode("Up"),
                    63235: idToKeyCode("Right"),
                    63233: idToKeyCode("Down"),
                    63272: idToKeyCode("Delete"),
                    63302: idToKeyCode("Insert"),
                    63236: idToKeyCode("F1"),
                    63237: idToKeyCode("F2"),
                    63238: idToKeyCode("F3"),
                    63239: idToKeyCode("F4"),
                    63240: idToKeyCode("F5"),
                    63241: idToKeyCode("F6"),
                    63242: idToKeyCode("F7"),
                    63243: idToKeyCode("F8"),
                    63244: idToKeyCode("F9"),
                    63245: idToKeyCode("F10"),
                    63246: idToKeyCode("F11"),
                    63247: idToKeyCode("F12"),
                    63248: idToKeyCode("PrintScreen"),
                    3: idToKeyCode("Enter"),
                    12: idToKeyCode("NumLock"),
                    13: idToKeyCode("Enter")
                };
                
            } else { // Modern versions of webkit
                keymaps.charCodeToKeyCode = {
                    13: 13,
                    27: 27
                };
            }

            break; 
            
            case _Platform.PRESTO:
            
                keymaps.ambiguousKeyPressCodes = {
                    35: 51, // # <=> End
                    36: 52, // $ <=> Home
                    44: 188, // Comma <=> Print Screen
                    45: 109, // - <=> Insert
                    46: 190, // Period <=> Delete
                    91: 219, // [ <=> Windows
                    93: 221  // ] <=> Apps
                };

                // Add presto specific maps...
                
               // Inverse ambigious key codes to simulate a keypress for in a key down event,
               // True values indicate they only should be simulated if the shift modifier is down.
               // False values are implicit (the remainding ambiguios codes) and should only be simulated
               // if the shift modifier is not down.
                keymaps.prestoSimulateInvOnShift = {
                    51: 1, // #
                    52: 1 // $
                    // 188 : 0 // ,
                    // 109 : 0 // -
                    // 190 : 0 // period
                    // 219 : 0 // [
                    // 221 : 0 // ]
                };
                
                // The "which" codes to use (as well as alpha numeric codes) as char codes on key presses.
                keymaps.prestoUseWhichCodes = {
                   33: 1, // ! <=> Page Up
                   34: 1, // " <=> Page Down
                   40: 1, // ( <=> Down
                   39: 1, // ' <=> Right
                   38: 1, // & <=> Up
                   37: 1, // % <=> Left
                   123: 1 // { <=> F12
                };
    
            break;
    }
    
    // Create inverse maps
    keymaps.ambiguousKeyPressCodesInv = {};
    for (var key in keymaps.ambiguousKeyPressCodes) {
        keymaps.ambiguousKeyPressCodesInv[keymaps.ambiguousKeyPressCodes[key]] = parseInt(key, 10);
    }
        

	/**
	 * @class A singleton that provides cross-browser/platform keyboard-normalization facilities
 	 * @author Brook Novak
	 */
	de.events.Keyboard = {
            
            
    	/**
    	 * Returns the event's "normalizedKey" to the DOM Level 3 Spec of keyIdentifier.
    	 * @param {Event} domEvent The dom event to "normalize"
    	 * 
    	 * @param {Boolean} isKeyDown True if the dom event for keydown, false if for keypress.
    	 * 
    	 * @return {String} The key identifier for the given key.
    	 *                  Null if the key ident is unknown or should be extracted from a different event type
    	 */
    	 getKeyIdentifier : function(domEvent, isKeyDown) {
     
    		var keyCode, // non printable keys. e.g. CTRL, INSERT
                charCode, // printable symbols. e.g. A,Z,&
                useGenericMapping = false;
    		
            //debug.println((isKeyDown ? "Keydown" : "KeyPress") + ": cc=" + domEvent.charCode + "kc=" + domEvent.keyCode);
            
     		switch(_engine) {
                
    			case _Platform.TRIDENT:
        
                    if (isKeyDown && (isNonPrintableKeyCode(domEvent.keyCode) || domEvent.keyCode == 8 || domEvent.keyCode == 9)) 
                         keyCode = domEvent.keyCode;                    
                
                    // Use keydown if CTRL is down, but keyPress if CTRL is not down
                    if ((!isKeyDown && !domEvent.ctrlKey) || (isKeyDown && domEvent.ctrlKey)) {
        		        if (keymaps.charCodeToKeyCode[domEvent.keyCode]) 
        					keyCode = keymaps.charCodeToKeyCode[domEvent.keyCode];
        		        else 
                            charCode = domEvent.keyCode;
                    }
    
    				break;
    			
    			case _Platform.GECKO:
                    if (isKeyDown) {
                        // Moz doesn't get keypress events for CTRL, ALT or SHIFT,
                        // So raise them on key down
                        if (domEvent.keyCode >= 16 && domEvent.keyCode <= 18) 
                           keyCode = domEvent.keyCode;
                           
                    } else {
                        keyCode = keymaps.keyCodeFix[domEvent.keyCode] || domEvent.keyCode;
                        charCode = domEvent.charCode;
                    }
    				break;
    			
    			case _Platform.WEBKIT:
                
                    if (_browser == _Platform.SAFARI) {
                        
                        if (isKeyDown) {
                           
                           
                            if (_engineVersion && _engineVersion < 525.13)
                                keyCode = keymaps.charCodeToKeyCode[domEvent.charCode] || domEvent.keyCode;
                           
                            else keyCode = domEvent.keyCode;
                            
                            if (!isNonPrintableKeyCode(keyCode) && !this.isAcceleratorDown(domEvent))
                                keyCode = 0;
                            
                        } else { // Key Press get printable charactors
                        
                            // starting with Safari 3.1 (verion 525.13) Apple switched the key
                            // handling to match the IE behaviour.
                            if (_engineVersion && _engineVersion < 525.13) {
                            
                                if (keymaps.charCodeToKeyCode[domEvent.charCode]) 
                                    keyCode = keymaps.charCodeToKeyCode[domEvent.charCode];
                                else 
                                    charCode = domEvent.charCode;
                                
                            } else {
                            
                                if (keymaps.charCodeToKeyCode[domEvent.keyCode]) 
                                    keyCode = keymaps.charCodeToKeyCode[domEvent.keyCode];
                                else 
                                    charCode = domEvent.keyCode;
                            }                        
                            
                        }
                        
                    } else if(_browser == _Platform.CHROME) {
                      
                       // Chrome is good, it sets keycode,charcode,which and keyidentifier..
                       if (isKeyDown) {
                           
                           // Keycodes can be detected from keydowns
                           if (isNonPrintableKeyCode(domEvent.keyCode))
                               keyCode = domEvent.keyCode;
                           
                           // If the accelerator key is down while pressing printable keys the charcodes
                           // appear as key codes in the keydown event
                           else if (this.isAcceleratorDown(domEvent))
                               charCode = domEvent.keyCode;
                               
                       } else {
                           
                           // Printable keys (charcodes) occur in the keypress event (except when
                           // the accelerator is down).
                           if (domEvent.charCode && !this.isAcceleratorDown(domEvent)) 
                               charCode = domEvent.charCode;
                       
                       }
                        
                    } else useGenericMapping = true;
      
    				break;
    				
    			case _Platform.PRESTO:
                                    
                    if (isKeyDown) {
    
                        if (keymaps.ambiguousKeyPressCodesInv[domEvent.keyCode]) {
                            // Avoid simulating key codes which will have a following legit keypress event 
                            var simOnShift = keymaps.prestoSimulateInvOnShift[domEvent.keyCode];
                            if ((domEvent.shiftKey && simOnShift) || (!domEvent.shiftKey && !simOnShift))
                                charCode = keymaps.ambiguousKeyPressCodesInv[domEvent.keyCode];
                            
                        } else if (keymaps.ambiguousKeyPressCodes[domEvent.keyCode])     
                            keyCode = domEvent.keyCode;
                        
                    } else { // Key press
                    
                        if (domEvent.which && (isAlphaNumericAscii(domEvent.which) || keymaps.prestoUseWhichCodes[domEvent.which])) 
                            charCode = domEvent.which;
                        else if (keymaps.keyCodeToId[domEvent.keyCode]) 
                			keyCode = domEvent.keyCode;
        				else 
        					charCode = domEvent.keyCode;                    
                    }
    				break;
                    
               case _Platform.KHTML:
                   
                       // TODO
                    useGenericMapping = true;
                        
                    break;                
    				
    			default:
                    useGenericMapping = true;
    		} // End switch
    		
            if (useGenericMapping && isKeyDown) {
                
                if (domEvent.keyIdentifier && domEvent.keyIdentifier.length > 0) { 
                    // See http://www.w3.org/TR/DOM-Level-3-Events/keyset.html
                    
                    // Is the key identifier unicode-encoded?
                    var ucMatch = /^U\+([\dA-Fa-f]+)$/.exec(domEvent.keyIdentifier);
                    
                    if (ucMatch) {
                        
                        // Extract unicode numerical value
                        var uniNum = parseInt(ucMatch[1], 16);
                        
                        // Convert into printable symbol
                        var printable = String.fromCharCode(uniNum);
                        
                        if (domEvent.shiftKey) 
                            return printable.toUpperCase();
                        else return printable.toLowerCase();
                        
                    } else return domEvent.keyIdentifier;
                    
                } else {
                    keyCode = domEvent.keyCode || domEvent.which;
                    charCode = domEvent.charCode;
                }
                
            }
    		
    	     if (keyCode) { // Use keyCode
    	     
                 // Omit ambiguous key codes: where actual key presses cannot be distinguished since this
                 // platform assigns some single key codes to multiple keys.
                 if (!isKeyDown && keymaps.ambiguousKeyPressCodes[keyCode]) 
                     return null;
    		  	
    	         return keyCodeToId(keyCode);
    
    	      } else if (charCode)  // Use charCode
    	        return charCodeToId(charCode);
    
    		return null;
    	},
    
         /**
          * @param {Event} domEvent A dom event
          * @return {Boolean} True if the accelerator key was down for the given event.s
          */
         isAcceleratorDown : function(domEvent) {
              // Add platform specific accelerator flag
              return _os == _Platform.MAC ? domEvent.metaKey : domEvent.ctrlKey;
         }
 
    }; // End Keyboard singleton

    /**
     * @param {Number} keyCode A key code to test
     * @return {Boolean} True if the key code is non-printable. I.E. not a printable symbol.
     */
    function isNonPrintableKeyCode(keyCode) {
      return typeof keymaps.keyCodeToId[keyCode] == "string" || typeof keymaps.specialToCharCode[keyCode] == "string";
    }

    /**
     * @param {Number} keyCode a keycode to test
     * @return {Boolean} True if the given keycode is identifiable.
     */	
	function isIdentifiableKeyCode(keyCode) {
        return isAlphaNumericAscii(keyCode) ||
               keymaps.specialToCharCode[keyCode] || /* Enter, Space, Tab, Backspace */
               keymaps.numpadToCharCode[keyCode] ||  /* Numpad */
               isNonPrintableKeyCode(keyCode);       /* non printable keys */
    }
    
    /**
     * @param {Number} code A char-code (or which-code) to test
     * @return {Boolean} True if the given code is alphanumeric
     */
    function isAlphaNumericAscii(code) {
        return (code >= keymaps.charCodeA && code <= keymaps.charCodeZ) || /* Upper */
               (code >= keymaps.charCodea && code <= keymaps.charCodez) || /* Lower */
               (code >= keymaps.charCode0 && code <= keymaps.charCode9);   /* Numbers */
    }
	
    /**
     * @param {Number} charCode A charactor code
     * @return {String} A key identifier for the given charactor code. Undefined if none exists.
     */
	function charCodeToId(charCode) {
	  return keymaps.specialToCharCode[charCode] || String.fromCharCode(charCode);	
	}

    /**
     * @param {Number} keyCode A keycode
     * @return {String} A key identifier for the given key code. Undefined if none exists.
     */	
    function keyCodeToId(keyCode) {
   	
      if (isIdentifiableKeyCode(keyCode)) {
	  	
        var numPadCharCode = keymaps.numpadToCharCode[keyCode];

        if (numPadCharCode) 
          return String.fromCharCode(numPadCharCode);

        return (keymaps.keyCodeToId[keyCode] || keymaps.specialToCharCode[keyCode] || String.fromCharCode(keyCode));
      
	  } else return null;
      
    }
	
    /**
     * @param {String} keyId The key identifier string to get the keycode for
     * @return {Number} The standadized keycode of the given key identifier
     */
	function idToKeyCode(keyId) {
      return keymaps.idToKeyCode[keyId] || keyId.charCodeAt(0);
    }
	


}) ();

